/****************************************************************************
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  As a special exception, you may use this file as part of a free software
 *  library without restriction.  Specifically, if other files instantiate
 *  templates or use macros or inline functions from this file, or you compile
 *  this file and link it with other files to produce an executable, this
 *  file does not by itself cause the resulting executable to be covered by
 *  the GNU General Public License.  This exception does not however
 *  invalidate any other reasons why the executable file might be covered by
 *  the GNU General Public License.
 *
 ****************************************************************************/
/**
	Type and functions definitions for count based pane

	Author: Tiziano De Matteis <dematteis <at> di.unipi.it>
*/

#ifndef _CB_WINDOW_H
#define _CB_WINDOW_H
#include <math.h>
#include "general.h" 
#include "pane_cb.h"
using namespace std;

//Funzione per applicare una computazione con stato sulla S-Structure:
/*void compute_task(s_structure_t *structure, task_t *t) {
	//Scorro tutta la lista:
	for(int i = 0; i < structure->length; i++) {
		if((structure->s_list)[i].tag >= t->tag) {
			//(structure->s_list)[i].value++;
			//(structure->s_list)[i].value = sin((structure->s_list)[i].value + 10);
			for(int j = 0; j<5;j++) (structure->s_list)[i].value = sin((structure->s_list)[i].value + 10);
		} 
	}
}*/

class CBPane{

	int pane_size;
	
	
	public:
		int type;
	CBPane(int pane_size)
	{
		//it is already optimized: we kept window_slide elements
		this->pane_size=pane_size;
		elements=(task_t *)malloc(pane_size*sizeof(task_t));
		ins_pointer=0;
		computation_time=0;
		computation_time_empty=0;
		//this->type=type;
	}

	~CBPane()
	{
		free(elements);
	}

	int getSize()
	{
		return pane_size;
	}

	/**
		Copies the task pointed by t in the proper position inside the pane.
		Reurns
		- O (or NULL) if no new computation has occured triggered by this insertion
		- 1 if the a computation has been triggered (the result is written in res)
	*/
	inline int insert(task_t * t, GReturn_type *res)
	{
		#ifdef MONITOR_COMP_TIME
			asm volatile("":::"memory");
			ticks start=getticks();
			asm volatile("":::"memory");
			#endif
		//if(ins_pointer>0 && elements[ins_pointer-1].internal_id!=(t->internal_id-1) )
		/*if(type!=t->type)
		{
			printf("MERDA SU classe: %d -> %d (task: %Ld) \n",type,t->type,t->id);
			exit(-1);
		}*/
		memcpy(&elements[ins_pointer],t,sizeof(task_t));
		if(elements[ins_pointer].id!=t->id)
		{
			printf("CAZZO\n");
			exit(-1);
		}
		ins_pointer++;
		if(ins_pointer==pane_size)
		{		
			//printf("Computo:%d  id:%Ld ins_pointer:%d\n",t->type,t->internal_id,ins_pointer);
			*res=G();
			// *res=0.1;
			#ifdef MONITOR_COMP_TIME
			asm volatile("":::"memory");
			computation_time+=getticks()-start;
			asm volatile("":::"memory");
			//printf("Comp time: %6.3f\n",(getticks()-start)/(2000000.0));
			#endif
			ins_pointer=0;
			return 1;

		}
		else{
		#ifdef MONITOR_COMP_TIME_EMPTY
			asm volatile("":::"memory");
			computation_time_empty+=getticks()-start;
			asm volatile("":::"memory");
			#endif
		return 0;
		}
		
	}

	/**
		Print the id of all the task contained in the elements array, starting from the oldest one
	*/
	void printAll()
	{
		for(int i=0;i<pane_size;i++)
		{
			cout << i <<"-th element is: "<<elements[i].id <<endl;
		}
	}

	/**
		If defined, return the average computation time (clock cycles)
	*/
	double getAvgComputationTime()
	{
		//printf("Elementi ricevuti: %Ld, computazioni effettuate: %d. Tempo di calcolo a vuoto (usec) %6.3f\n",total_elements,(total_elements/window_slide),((double)computation_time_empty)/total_elements);

		
			return computation_time;

	}

	private:
	task_t *elements; //array that contains all the elements that have to be yet processed 
	int ins_pointer; //always points to the insertion point in elements
	ticks computation_time; //monitoring (if specified) of the computation time
	ticks computation_time_empty; //monitoring (if specified) of the computation time
	/**
		Definition of the G function: it operates on the elements array and the result will be returned
	*/
	GReturn_type G()
	{
		//by default is a double

		GReturn_type ret=0.0; 
		for(int i = 0; i < pane_size; i++) 
		{
			for(int j = 0; j<G_GRAIN*10 ;j++) 
				ret+=pow(elements[i].tag,ret);
				//ret+=sin(elements[i].tag+ret);
		} 
		return ret;
	}	

};

#endif