/****************************************************************************
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  As a special exception, you may use this file as part of a free software
 *  library without restriction.  Specifically, if other files instantiate
 *  templates or use macros or inline functions from this file, or you compile
 *  this file and link it with other files to produce an executable, this
 *  file does not by itself cause the resulting executable to be covered by
 *  the GNU General Public License.  This exception does not however
 *  invalidate any other reasons why the executable file might be covered by
 *  the GNU General Public License.
 *
 ****************************************************************************
/**
	Type and functions definitions for count based window
	USED FOR FARM
	(Unlike the case of Functional Partitioning where the window
	is triggeered by a slide, here the window have to be triggered only when it is full)

	NOTE: the slide it is passed just to have the same computation of the FP case
	Author: Tiziano De Matteis <dematteis <at> di.unipi.it>

*/

#ifndef _CB_WINDOWF_H
#define _CB_WINDOWF_H
#include <math.h>
#include <stdio.h>
#include "general.h" 
#include "functional_partitioning_cb.h"
using namespace std;
//Funzione per applicare una computazione con stato sulla S-Structure:
/*void compute_task(s_structure_t *structure, task_t *t) {
	//Scorro tutta la lista:
	for(int i = 0; i < structure->length; i++) {
		if((structure->s_list)[i].tag >= t->tag) {
			//(structure->s_list)[i].value++;
			//(structure->s_list)[i].value = sin((structure->s_list)[i].value + 10);
			for(int j = 0; j<5;j++) (structure->s_list)[i].value = sin((structure->s_list)[i].value + 10);
		} 
	}
}*/

class CBWindowFarm{

	int window_size;
	int window_slide;
	int64_t total_elements; //total elements that were contained in the window
	
	public:
	CBWindowFarm(int window_size, int window_slide, int grain=10)
	{
		//it is already optimized: we kept window_slide elements
		this->window_size=window_size;
		this->window_slide=window_slide;
		numb_pr=window_size/window_slide;
		//elements=(task_t *)malloc(window_slide*sizeof(task_t));
		partial_results=(GReturn_type *)malloc(numb_pr*sizeof(GReturn_type));
		ins_pointer=0;
		pr_ins=0;
		act_numb_pr=0;
		total_elements=0;
		computation_time=0;
		computation_time_empty=0;
		no_result=NULL;
		//this->type=type;
		iteration_multiplier=grain;

		//version with single F
		elements=(task_t *)malloc(window_size*sizeof(task_t)); //delete the prvious malloc
		eflc=0;
		//printf(" grain %d\n",grain);
	}

	~CBWindowFarm()
	{
		free(elements);
	}

	int getSize()
	{
		return window_size;
	}
	int64_t getTotalElements()
	{
		return total_elements;
	}

	/**
		Copies the task pointed by t in the proper position inside the window.
		If the window is full, the oldest element is automatically overwritten
		returns:
		- O (or NULL) if no new computation has occured triggered by this insertion
		- 1 if the a computation has been triggered (the result is written in res)
	*/
	inline int insert(task_t * t, HReturn_type *res)
	{
		#ifdef MONITOR_COMP_TIME
			asm volatile("":::"memory");
			ticks start=getticks();
			asm volatile("":::"memory");
			#endif
	
		memcpy(&elements[ins_pointer],t,sizeof(task_t));
		ins_pointer++;
		total_elements++;

	

		//version with single F
		if(ins_pointer==window_size) 
		{
			
			eflc=0;
			F(res);
			return 1;
		}
		else
			return 0;


		
	}
	void reset()
	{
		ins_pointer=0;
		total_elements=0;
		//for single F
		eflc=0;
	}
	int64_t getLastInserted()
	{
		return elements[ins_pointer-1].internal_id;
	}
	int getInsertionPointer()
	{
		return ins_pointer;
	}

	/**
		Print the id of all the task contained in the elements array, starting from the oldest one
	*/
	void printAll()
	{
		for(int i=0;i<window_slide;i++)
		{
			cout << i <<"-th element is: "<<elements[i].id <<endl;
		}
	}

	/**
		If defined, return the average computation time (clock cycles)
	*/
	double getAvgComputationTime()
	{
		//printf("Elementi ricevuti: %Ld, computazioni effettuate: %d. Tempo di calcolo a vuoto (usec) %6.3f\n",total_elements,(total_elements/window_slide),((double)computation_time_empty)/total_elements);

		if(total_elements/window_slide>=1)
		{
			/*#ifdef  MONITOR_COMP_TIME_EMPTY
			printf("%6.3f\n",(((double)computation_time_empty)/total_elements)/FREQ);
			#endif*/
			return computation_time/(total_elements/window_slide);
		}
		else
			return 0.0;
	}

	private:
	HReturn_type *no_result; //it is the neutral element
	int numb_pr; //total number of partial results
	int act_numb_pr; //actual number of partial results
	task_t *elements; //array that contains all the elements that have to be yet processed (size=slide)
	GReturn_type *partial_results; //this array contains all the partial results obtained by processing the various slide with the G function (size=window_size/window_slide)
	int ins_pointer; //always points to the insertion point in elements
	int pr_ins; //insetion point in partial_results (it is handled as a circular buffer)
	ticks computation_time; //monitoring (if specified) of the computation time
	ticks computation_time_empty; //monitoring (if specified) of the computation time
	int type;
	int eflc; //elements received from the last computation

	int iteration_multiplier;
	/**
		Definition of the G function: it operates on the elements array and the result will be put in partial_results
	*/
	GReturn_type G()
	{
		//by default is a double
		int it=iteration_multiplier*G_GRAIN;
		GReturn_type ret=0.0; 
		for(int i = 0; i < window_slide; i++) 
		{
			for(int j = 0; j<it ;j++) 
				ret+=pow(elements[i].tag,ret);
				//ret+=sin(elements[i].tag+ret);
		} 
		return ret;
	}

	/**
		Definition of the H function: it operates on the partial_results array and the result will be returned to the Worker
	*/
	void H(HReturn_type *res)
	{
		//HReturn_type* ret;
		//possibile ottimizzazione: riciclo memoria allocata
		//posix_memalign((void **)&ret,CACHE_LINE_SIZE,CACHE_LINE_SIZE);
		//ret=(HReturn_type *)malloc(sizeof(HReturn_type));
		res->res=0.0;
		//it is very similar to the G but this is only for experimentation, it has not to be a meaningful computation
		for(int i = 0; i < act_numb_pr; i++) 
		{
			for(int j = 0; j<H_GRAIN*10 ;j++) 
				res->res+=pow(partial_results[i],res->res);
				//res->res+=sin(partial_results[i]+res->res);
		} 
	}

	/**
		Definition of the F function: it should work on the whole window and mimic the composition of
		G+H.
		Since it is just sinthetic workload, we will NOT consider the right ordering of the internal element
		of the window
	*/
	void F(HReturn_type *res)
	{
		int it=iteration_multiplier*G_GRAIN;
		res->res=0.0;
		memset(partial_results,0,sizeof(double)*numb_pr);
		for(int i = 0; i < numb_pr; i++) 
		{

			for(int j = i*window_slide; j<(i+1)*window_slide ;j++) 
				for(int l = 0; l<it ;l++) 
					partial_results[i]+=pow(elements[j].tag,partial_results[i]);
				//ret+=sin(elements[i].tag+ret);
		} 

		for(int i = 0; i < numb_pr; i++) 
		{
			for(int j = 0; j<H_GRAIN*10 ;j++) 
				res->res+=pow(partial_results[i],res->res);
		}

	}


};

#endif