/*
	Header containing general declaration.
	It contains also all the definitions for inlineable functions

*/

#ifndef _GENERAL
#define _GENERAL
//Include:
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <malloc.h>

#include <math.h>
#include "zipf.h"
//#include "queue.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/param.h>
#include "cycle.h"
#include "iaperf.h"
#include <assert.h>
#include <ff/buffer.hpp>
#include <ff/allocator.hpp>	


//#include <libmemcached/memcached.h>

//Define:
//#define DEBUG
//#define ANDROMEDA //Define per il mapping non lineare dei thread sui core.
#define STD_DEVIATION 0.05 //Standard deviation nel caso di generazione task con distribuzione Normale.
//#define FREQ 2268 //Frequenza di Andromeda.
#define MAX_OCCURRENCE 100 //Massimo numero di occorrenze per classe (generazione probabilità zipf).
#define CACHE_LINE_SIZE 64 //Dimensione dei blocchi di cache.
#define QUEUE_SIZE 12000 //Define per la dimensione della code di ingresso dei task (non valida nel caso di on-demand).
#define MAX_STEP_NUMBER 100000 //Numero massimo di step di controllo di una esecuzione.
#define MAX_MSG_THRESHOLD 100 //Numero massimo di messaggi inviabili senza riscontro (Generatore Client-Server).
#define SERVICETIME_FILENAME "../results/servicetime.txt" //Nome del file dei risultati del tempo di servizio per step.
#define THROUGHPUT_FILENAME "../results/throughput.txt" //Nome del file dei risultati del throughput per step.
#define QUEUE_FILE_HDR "../results/queuelength-" //Nome del file dei risultati della lunghezza della coda del Farm per step.
#define LATENCY_FILENAME "../results/latency.txt" //Nome del file dei risultati della latenza per step.
#define CONTROL_STEP 120 //Lunghezza del control step usato dai manager per l'Adattività (espresso in secondi).
#define MAX_SEND_RETRIES 1000 //maximal number of send retries before declaring it unsuccesfull
//errors definition
#define BOTTLENECK_ERR -100
#define INTERNAL_ERR -200
#define RECONF_ERR -300
#define MAX_LAT 1000000
#define NO_MORE_INIT 60000000
//Names of output files
#define QLENGTH_FILE "qlength"
#define FREQ_FILE "frequencies"
#define CLASS_FILE "classes"
#define LAT_FILE "latencies"
#define COMP_H_FILE "comph"
#define DEBUG(X) X;

//Monitoring definitions
#define NMONITORING 10 //number of monitoring data structure for each entity
#define MONITORING_STEP 1000 //time interval between two monitoring phases of monitoring (milliseconds)
#define QUEUE_SIZE_MON 5 //size of queues used for sent monitoring data

//Computational grains
//coarse grain is 15-8
//fine grain i 5-3
//finer 3-2
//grain sustainable to pianosa (at max 1.5MT/s)
#define G_GRAIN 18
#define H_GRAIN 12


#define ANSI_COLOR_RED     "\x1b[31m"
#define ANSI_COLOR_GREEN   "\x1b[32;1m"
#define ANSI_COLOR_RED_BOLD "\x1b[31;1m"
#define ANSI_COLOR_YELLOW  "\x1b[33m"
#define ANSI_COLOR_BLUE    "\x1b[34m"
#define ANSI_COLOR_MAGENTA "\x1b[35m"
#define ANSI_COLOR_CYAN    "\x1b[36m"
#define ANSI_COLOR_RESET   "\x1b[0m"
#define ANSI_COLOR_BLUE_REVERSE		 "\x1b[34;7m"


/*
//Struct con gli estremi (ip_address e porta) per collegarsi ad una componente Farm:
typedef struct {
	int port;
	char *ip_address;
} farm_addrport_t;

//Struct con la lista dei Farm a cui una componente Dispatcher deve collegarsi:
typedef struct {
	int num;
	farm_addrport_t *list;
} farm_list_t;
*/

//Utility macros
#define ERRNULL(X) if( (X)==NULL)  return -1;
#define REPEAT_25( x ) { x x x x x x x x x x x x x x x x x x x x x x x x x }
typedef enum punctation_t{
	NO=0, //in order to be the default value
	MOVING_IN=1,
	MOVING_OUT=2
}punctation_t;

//typedef int type_t;
//Generic task definition
typedef struct {
	union{	
		struct
		{
			int64_t id; //Unique ID of the task: its meaning may depends from the particular implemenetation
			int type; //Class of the task (with -1  it representsEOS)
			
			double tag; //Numerical Data of the task.
			int tag2; // Data of the task

			long long timestamp; //Timestamp
			
			int64_t window_id; //unique id of the window for that particular class. It will be meaningful only in the FARM version

			int64_t internal_id; //id assigned to the task while being computed in the program (e.g. incremental for each class). Not necessary for all implementations
			char worker; //tmp for debug
			punctation_t punctuation;

			// ticks ts_em;
		};
		char padding[CACHE_LINE_SIZE];
	};
} task_t;


// For the moment they are global just for testing

#ifdef USE_ZMQ
void* connect_to(void *zmq_context, const char*address, int port);
void * receive_connection(void *zmq_context,int port) ;
int  socket_bsend(void* socket, void *msg, size_t len);
int socket_send(void* socket, void *msg, size_t len);
int socket_receive(void *socket, void *msg, size_t len);
int close(void * socket);
#else
//Funzione per creare una connessione TCP verso un IP address e una porta (e restituire il socket):
int connect_to(const char *ip_address, int port);

//Funzione per ricevere num_conn connessioni TCP su una porta (ritorna un array di socket):
int * receive_connection(int num_conn, int port);

//Funzione per trasmettere un messaggio su un socket (SEND):
inline size_t socket_send(int s, void *msg, size_t len) __attribute__((always_inline));

//Funzione per ricevere un messaggio su un socket (RECEIVE):
inline size_t socket_receive(int s, void *vtg, size_t len) __attribute__((always_inline));
inline size_t socket_send(int s, void *msg, size_t len) {
	size_t sentBytes = 0;
	size_t sent = 0;
	char *msg_char = (char *) msg;
	//Faccio le send sul socket fintanto che non ho mandato len bytes:
	while(sentBytes < len) {
		//sent = send(s, msg_char, len, (int) MSG_DONTWAIT); 
		sent = send(s, msg_char, len, 0); 
		if(sent==-1 && errno==EWOULDBLOCK)
		{
			perror("Error send(): buffer full");
			return BOTTLENECK_ERR;
		}
		if (sent == -1) { //per renderlo bloccante, si lascia questo ultimo if e non si setta il flag sul socket
			perror("Error send() call");
			return 0;
		}
		else if(sent <= len) {
			//printf("send parziale");
			//exit(-1);
			//Altrimenti si tratta di una send parziale che devo completare:
			sentBytes += sent;
			msg_char += (sent);
		}
	}
	return sentBytes;
}

//Funzione per ricevere un messaggio su un socket (RECEIVE):
inline size_t socket_receive(int s, void *vtg, size_t len) {
	char *vtg_char = (char *) vtg;
	//Effettuo la receive sul socket:
	size_t received = recv(s, vtg_char, len, MSG_WAITALL);
	if(received != len){
		perror("Errore recv() call");
		return 0;
	}
	return received;
}
inline int send ( void *t, ff::SWSR_Ptr_Buffer *outqueue) __attribute__((always_inline))/* __attribute__((noinline))*/; 
inline int bsend ( void *t, ff::SWSR_Ptr_Buffer *outqueue) __attribute__((always_inline))/* __attribute__((noinline))*/; 
inline void receive ( void **t, ff::SWSR_Ptr_Buffer *inqueue) __attribute__((always_inline))/* __attribute__((noinline))*/;
static bool blocking_send=true;
inline void set_blocking(bool value)
{
	blocking_send=value;
	printf("New value of blocking send: %d\n",blocking_send);
}
//this difference (with blocking send) is only for testing purposes
inline int send ( void *t, ff::SWSR_Ptr_Buffer *outqueue)
{
	int attempt=0;
	while(!outqueue->push((void*)t))
	{
		REPEAT_25(asm volatile("PAUSE" ::: "memory");)

		#ifdef NON_BLOCKING_FF
		// if(!blocking_send)
		// {
		attempt++;
		if(attempt>MAX_SEND_RETRIES)
			return 0; //send failure
		// }
		#endif

	}
	return 1;
	
}

inline int bsend ( void *t, ff::SWSR_Ptr_Buffer *outqueue)
{
	int attempt=0;
	while(!outqueue->push((void*)t))
	{
		REPEAT_25(asm volatile("PAUSE" ::: "memory");)
	}
	return 1;
	
}


inline void receive ( void **t, ff::SWSR_Ptr_Buffer *inqueue)
{
	while (!inqueue->pop( t)) {
		// Small sleep
		REPEAT_25(asm volatile("PAUSE" ::: "memory");)
	}
}


//Funzione per selezionare non-deterministicamente un Socket Descriptor da cui ricevere:
int selectConnection(fd_set *fds, int *s_array, int nConnections, int lastSocket, int max);

bool SetSocketBlockingEnabled(int fd, bool blocking);
size_t socket_bsend(int s, void *msg, size_t len);
#endif

//Funzione per la chiusura di un socket
int close(int socket);


#endif