/****************************************************************************
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  As a special exception, you may use this file as part of a free software
 *  library without restriction.  Specifically, if other files instantiate
 *  templates or use macros or inline functions from this file, or you compile
 *  this file and link it with other files to produce an executable, this
 *  file does not by itself cause the resulting executable to be covered by
 *  the GNU General Public License.  This exception does not however
 *  invalidate any other reasons why the executable file might be covered by
 *  the GNU General Public License.
 *
 ****************************************************************************
/**
	Headers containing the various definitions for adaptiveness and reconfigurations
	in the pane based version with count based windows

	Author: Tiziano De Matteis <dematteis <at> di.unipi.it>
*/
#ifndef _FP_CB_RECONF
#define _FP_CB_RECONF
#include <atomic>
#include <unordered_map>
#include <ff/buffer.hpp>
#include <ff/allocator.hpp>

#include "general.h"
#include "pane_cb.h"
using namespace std;
using namespace ff;
//Data structure containing the information that will drive the reconfiguration (sent from controller to emitter)
typedef struct{
	union{	
		struct
		{
			int par_degree_changes; //identifies how many worker have been added (if >0) or removed (if <0). It is equal to zero if there are no changes
			SWSR_Ptr_Buffer **wqueues; //queues to the newly spawned workers (if any)
			char *scheduling_table;
		};
		char padding[CACHE_LINE_SIZE];
	};
}reconf_data_em_t;

//Data structure for reconfigurations from worker to collector
typedef struct{
	union{	
		struct
		{
			int par_degree_changes; //identifies how many worker have been added (if >0) or removed (if <0). It is equal to zero if there are no changes
			SWSR_Ptr_Buffer **wqueues; //queues to the newly spawned workers (if any)
		};
		char padding[CACHE_LINE_SIZE];
	};
}reconf_data_c_t;


#endif