/**
	Version for count based windows, parallelized with the farm paradigim.

	In this case whole windows are scheduled to workers. Tuple distribution is performed
	on the fly as soon as the tuple arrive, by creating multiple copies of the tuple (one for each window)
	
	FOR THE MOMENT IT IS ASSUMED THE EXISTENCE OF A SLIDE: I think that can be easily adapted to a case in which there is no slide

	!!!NOTE: since this version has been implemented using a formulation F=G+H (in order to be comparable with the Pane version),
	in the window will take place only the elements of the last slide.
	EDIT: This can be changed by properly comment/uncomment code in cbwindow.hpp
	

*/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <ff/buffer.hpp>
#include <iostream>
#include <unordered_map>
#include <vector>
#include <sched.h>
#include "../Includes/RandomGenerator.h"
#include "../Includes/cycle.h"
#include "../Includes/iaperf.h"
#include "../Includes/general.h"
#include "../Includes/farm_cb.h"
#include "../Includes/farm_cb_reconf.h"
#include "../Includes/affinities.h"
#include "../Includes/cbwindow_farm.hpp"
#include <ff/allocator.hpp>	
#if defined(USE_ZMQ)
#include <zmq.h>
#endif
using namespace ff;
using namespace std;


#if defined(ADAPTIVITY)
	#ifndef MONITORING 
	#define MONITORING //monitoring is required
	#endif
#endif


//vector<unsigned long> queue_lengths_mon(100)[64];
//number of monitoring step perfomed by the emitter
int mon_step=0;




int main(int argc, char *argv[])
{
	
	PERFINITMHZ(FREQ);

	int num_workers;
	int num_classes;
	int port;
	int i;
	int window_size;
	int window_slide;
	int input_rate=0, input_length=0;
	cpu_set_t cpuset;
	SWSR_Ptr_Buffer **quEW; //queues emitter->worker
	SWSR_Ptr_Buffer **quWC; //queues worker->collector
	#if defined(MONITORING)
		SWSR_Ptr_Buffer *quECN; //queue emitter->controller
		SWSR_Ptr_Buffer *quCCN; //queue collector->controller
		SWSR_Ptr_Buffer **quWCN; //queues workers->controller
		SWSR_Ptr_Buffer *quCNE; //queue controller->emitter
		SWSR_Ptr_Buffer *quCNC; //queue controller->collector
		reconf_repository_t *repo; //repository that will contain all the structures needed for reconfigurations
	#endif
	//arguments for threads
	worker_data_t *worker_data;
	emitter_data_t emitter_data;
	collector_data_t collector_data;
	controller_data_t controller_data;
	//threads ids
	pthread_t *wtids;
	pthread_t ctid,cntid;
	/**
		check line arguments...
	*/
	if(argc<5)
	{
		fprintf(stderr, "Usage: %s num_classes num_workers port window_size [window_slide] [input_rate] [input_length]\n",argv[0] );
		return EXIT_FAILURE;
	}
	printf("Sizeof :%d\n",sizeof(task_t));
	num_classes=atoi(argv[1]);
	num_workers=atoi(argv[2]);
	assert(num_workers<=max_workers);
	port=atoi(argv[3]);
	window_size=atoi(argv[4]);
	if (argc>=6)
		window_slide=atoi(argv[5]);
	else
		window_slide=1;
	assert(window_size%window_slide==0);
	//the last two are used only for testing purposes
	if(argc>=7)
		input_rate=(int)atof(argv[6]);
	if(argc>=8)
		input_length=(int)atof(argv[7]);

	/**
		Initialize data structures
	*/
	pthread_barrier_t barrier;
	pthread_barrier_init (&barrier, NULL, num_workers + 2);
	//queues (we allocate space for having max_workers queues in case of reconfigurations that involve changes in par degree)
	quEW = (SWSR_Ptr_Buffer**) malloc(max_workers * sizeof(SWSR_Ptr_Buffer*));
	ERRNULL(quEW);
	quWC = (SWSR_Ptr_Buffer**) malloc(max_workers * sizeof(SWSR_Ptr_Buffer*));
	ERRNULL(quWC);
	#if defined(MONITORING)
		quWCN= (SWSR_Ptr_Buffer**) malloc(max_workers * sizeof(SWSR_Ptr_Buffer*));
		ERRNULL(quWCN);
		quECN=new SWSR_Ptr_Buffer(QUEUE_SIZE_MON);
		quECN->init();
		quCCN=new SWSR_Ptr_Buffer(QUEUE_SIZE_MON);
		quCCN->init();
		quCNE=new SWSR_Ptr_Buffer(QUEUE_SIZE_MON);
		quCNE->init();
		quCNC=new SWSR_Ptr_Buffer(QUEUE_SIZE_MON);
		quCNC->init();
		posix_memalign((void **)&repo,CACHE_LINE_SIZE,sizeof(reconf_repository_t));
		//repo->moving_windows=new unordered_map<int,CBWindow*>();
		repo->moving_windows=(CBWindowFarm **)malloc(sizeof(void *)*num_classes);
		posix_memalign((void**)&repo->reconfiguration_finished,CACHE_LINE_SIZE,sizeof(atomic<int64_t>)*max_workers);
		for(int i=0;i<max_workers;i++)
			repo->reconfiguration_finished[i].store(1);
	#endif

	#ifdef QSPLITTED
	int qsize=QUEUE_SIZE/num_workers;
	#else
	int qsize=QUEUE_SIZE;
	#endif	
	printf("Queue size: %d\n",qsize);
	//We initialize up to the actual par degree (not the maximum one)
	for (i = 0; i < num_workers; i++) {
		quEW[i] = new SWSR_Ptr_Buffer(qsize);
		quEW[i]->init();
		quWC[i] = new SWSR_Ptr_Buffer(qsize); //technically we could reduce this
		quWC[i]->init();
		#if defined(MONITORING)
			quWCN[i]=new SWSR_Ptr_Buffer(QUEUE_SIZE_MON);
			quWCN[i]->init();
		#endif
	}
	
	//frequency counters
	//int64_t* classes_freq=(int64_t*)calloc(num_classes,sizeof(int64_t));
	//computation times
	//double *comp_time=(double*)calloc(num_classes,sizeof(double));
	//scheduling table
	//char *scheduling_table=(char *)calloc(num_classes,sizeof(char));
	//shared start tick
	ticks *start_global_ticks=(ticks*)calloc(1,sizeof(ticks));
	//suffix for output files
	char suffix[80]; //(it will exist until all activities are not finished)
	//suffix for output files
	sprintf(suffix,"%d_%d_%d_%d_%d_%d",num_classes,num_workers,window_size,window_slide,input_rate,input_length);
	#if !defined(DONT_USE_FFALLOC) && !defined(TASK_BUFF)
            ff_allocator* ffalloc=new ff_allocator;
            assert(ffalloc);
            //qui indicano il numero di segmenti per dimensione 32,64,128,...
            int nslabs[N_SLABBUFFER]={0,QUEUE_SIZE*num_workers,0,0,0,0,0,0,0 };
            if (ffalloc->init(nslabs)<0) {
                error("FATAL ERROR: allocator init failed\n");
                abort();
            }
	#endif    
	
	/**
		Threads creations
	*/
	//Workers creation
	wtids = (pthread_t*) malloc(num_workers * sizeof(pthread_t));
	worker_data =(worker_data_t*) malloc(num_workers * sizeof(worker_data_t));
	for (i = 0; i < num_workers; i++) {		
		pthread_t tid;
		worker_data[i].workerId = i;
		worker_data[i].inqueue = quEW[i];
		worker_data[i].outqueue = quWC[i];

		worker_data[i].barrier=&barrier;
		worker_data[i].window_size=window_size;
		worker_data[i].window_slide=window_slide;
		worker_data[i].start_global_ticks=start_global_ticks;
		//worker_data[i].comp_time=comp_time;
		worker_data[i].freq=FREQ;
		worker_data[i].num_classes=num_classes;
		#if !defined(DONT_USE_FFALLOC) && !defined(TASK_BUFF)
			worker_data[i].ffalloc=ffalloc;
		#endif
		#if defined(MONITORING)
			worker_data[i].cn_outqueue=quWCN[i];
			worker_data[i].repo=repo;
		#endif
		pthread_create(&tid, NULL, worker, &(worker_data[i]));
		//set CPU affinity of worker threads
		CPU_ZERO(&cpuset);
		CPU_SET(affinities[i], &cpuset);

		if (pthread_setaffinity_np(tid, sizeof(cpu_set_t), &cpuset)) {
			cerr << "Cannot set thread to CPU " << i << endl;
		}
		//worker_data[i].num_workers = num_workers;
		wtids[i] = tid;
	}

	//collector creation
	collector_data.num_workers=num_workers;
	collector_data.num_classes=num_classes;
	collector_data.inqueue=quWC;
	collector_data.wqueue=quEW;
	collector_data.barrier=&barrier;
	collector_data.freq=FREQ;
	collector_data.start_global_ticks=start_global_ticks;
	collector_data.suffix=suffix;
	collector_data.window_slide=window_slide;
	collector_data.window_size=window_size;
	collector_data.max_workers=max_workers;
	#if defined(MONITORING)
		collector_data.cn_outqueue=quCCN;
		collector_data.cn_inqueue=quCNC;
	#endif
	pthread_create(&ctid, NULL, collector, &collector_data);
	//set CPU affinity of collector thread
	CPU_ZERO(&cpuset);
	CPU_SET(collector_affinity, &cpuset);

	if (pthread_setaffinity_np(ctid, sizeof(cpu_set_t), &cpuset)) {
		cerr << "Cannot set thread to CPU " << i << endl;
	}
	
	#if defined(MONITORING)
	//controller creation
	controller_data.num_workers=num_workers;
	controller_data.num_classes=num_classes;
	controller_data.e_inqueue=quECN;
	controller_data.w_inqueue=quWCN;
	controller_data.c_inqueue=quCCN;
	controller_data.e_outqueue=quCNE;
	controller_data.c_outqueue=quCNC;
	controller_data.suffix=suffix;
	controller_data.repo=repo;
	controller_data.affinities=affinities;
	controller_data.window_size=window_size;
	controller_data.window_slide=window_slide;
	controller_data.max_workers=max_workers;
	controller_data.freq=FREQ;
	controller_data.start_global_ticks=start_global_ticks;
	#if !defined(DONT_USE_FFALLOC) && !defined(TASK_BUFF)
		controller_data.ffalloc=ffalloc;
	#endif

	pthread_create(&cntid, NULL, controller, &controller_data);
	//set CPU affinity of collector thread
	CPU_ZERO(&cpuset);
	CPU_SET(controller_affinity, &cpuset);
	if (pthread_setaffinity_np(cntid, sizeof(cpu_set_t), &cpuset)) {
		cerr << "Cannot set thread to CPU " << i << endl;
	}

	#endif
	// Start working as emitter
	emitter_data.num_workers=num_workers;
	emitter_data.num_classes=num_classes;
	emitter_data.port=port;
	emitter_data.outqueue=quEW;
	emitter_data.barrier=&barrier;
	emitter_data.window_size=window_size;
	emitter_data.window_slide=window_slide;
	// emitter_data.classes_freq=classes_freq;
	//emitter_data.scheduling_table=scheduling_table;
	emitter_data.start_global_ticks=start_global_ticks;
	emitter_data.freq=FREQ;
	#if !defined(DONT_USE_FFALLOC) && !defined(TASK_BUFF)
		emitter_data.ffalloc=ffalloc;
	#endif
	#if defined(MONITORING)
		emitter_data.cn_outqueue=quECN;
		emitter_data.cn_inqueue=quCNE;
		emitter_data.repo=repo;
	#endif
	// Set my affinity
	CPU_ZERO(&cpuset);
	CPU_SET(emitter_affinity, &cpuset);
	pthread_t etid = pthread_self();

	if (pthread_setaffinity_np(etid, sizeof(cpu_set_t), &cpuset)) {
		cerr << "Cannot set thread to CPU " << i << endl;
	}
	emitter(&emitter_data);

	// Wait for the completion of the threads
	void *retval ;
	for (i = 0; i < num_workers; i++) {
		pthread_join(wtids[i], &retval);
	}
	pthread_join(ctid, &retval);
	#if defined(MONITORING)
	pthread_join(cntid, &retval);
	#endif
	//print to file classes statistics
	char out_file[100];
	sprintf(out_file,"%s_%s.dat",CLASS_FILE,suffix);
	/*FILE* fclass=fopen(out_file,"w");
	fprintf(fclass,"#Number of classes: %d\n#",num_classes);
	fprintf(fclass, "#CLASS_ID FREQ ASSIGNED_WORKER COMP_TIME\n");

	for(int i=0;i<num_classes;i++)
		fprintf(fclass,"%d\t%Ld\t%d\t%6.3f\n",i,classes_freq[i],scheduling_table[i],comp_time[i]);
	fclose(fclass);
*/
	exit(EXIT_SUCCESS);
}
