/**
	Collector for Functional Partitioning Implementation
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <ff/buffer.hpp>
#include <iostream>
#include <unordered_map>
#include <vector>
#include <sched.h>
#include <algorithm> 
#include "../Includes/RandomGenerator.h"
#include "../Includes/cycle.h"
#include "../Includes/iaperf.h"
#include "../Includes/general.h"
#include "../Includes/farm_cb.h"
#include "../Includes/cbwindow.hpp"
#include "../Includes/farm_cb_reconf.h"

#include <ff/allocator.hpp>	
#if defined(USE_ZMQ)
#include <zmq.h>
#endif
using namespace ff;
using namespace std;

void * collector(void *args) {

	HReturn_type *rcvd=(HReturn_type*)malloc(sizeof(HReturn_type));
	collector_data_t *data = (collector_data_t *) args;
	int received_EOS=0; //number of workers from which the collector has received the EOS
	int index=0;
	double dlat=0.0; //latency in usec
	int64_t rcvd_results=0;
	int64_t partial_rcvd=0;
	long int freq=data->freq;
	char *suffix=data->suffix;
	ticks *start_global_ticks=data->start_global_ticks;
	ticks last_print, start_ticks;
	ticks  print_rate=(unsigned long long)freq*1000000;
	ticks no_more_init=(unsigned long long)freq*NO_MORE_INIT; 
	int window_slide=data->window_slide;
	int window_size=data->window_size;
	int num_workers=data->num_workers;
	int num_classes=data->num_classes;
	int max_workers=data->max_workers;

	
	//vector for computing response time percentile (maybe we can find something more efficient)
	vector<double> percentile_lat;
	//Ordering variables
	//expected internal id for the received results (it will be used for correctness)
	int64_t *expected_iid;
	//number of results received disordered partial results for each class
	int *disordered;
	//buffer for results received out of order
	vector<vector<HReturn_type>> buffer_disordered;

	//FOR TESTING ONLY
	//latencies, percentiles, actual par degree... all these metrics are printed in a file
	//at the end of the program (on a second basis)
	double latencies[800];
	double percentiles[800];
	int64_t lat_recv[800];
	double seconds[800];
	char par_degrees[800];
	int num_latencies=0;
	double *lat_per_class=(double*)calloc(num_classes,sizeof(double));
	int *recvd_per_class=(int *)calloc(num_classes,sizeof(int));
	int *recvd_per_worker=(int*)calloc(max_workers,sizeof(int));
	vector<double> all_lat;
	vector<ticks> all_lat_times;
	vector<char> all_lat_worker; //who computes the task


	//Init
	pthread_barrier_t *barrier = data->barrier;
	SWSR_Ptr_Buffer **inqueue=data->inqueue;
	
	

	#if defined(MONITORING)
		//the queue for sending monitoring data to the controller
		SWSR_Ptr_Buffer *cn_outqueue=data->cn_outqueue;
		//define the data structure for monitoring
		volatile ticks monitoring_timer;
		volatile ticks monitoring_step=MONITORING_STEP*1000*freq; //after how many cycles we have to monitor
		collector_monitoring_t *monitoring;
		char mon_ptr=0; //identifies the next monitoring structure in which write data
		posix_memalign((void **)&monitoring,CACHE_LINE_SIZE,NMONITORING*sizeof(collector_monitoring_t));
		for(int i=0;i<NMONITORING;i++)	
		{
			//init the internals
			monitoring[i].mon_id=i;
			monitoring[i].results=0;
			posix_memalign((void **)&(monitoring[i].latency_per_class),CACHE_LINE_SIZE,num_classes*sizeof(double));
			memset(monitoring[i].latency_per_class,0,num_classes*sizeof(double));
			posix_memalign((void **)&(monitoring[i].results_per_class),CACHE_LINE_SIZE,num_classes*sizeof(int));
			memset(monitoring[i].results_per_class,0,num_classes*sizeof(int));
		}

		#if defined(ADAPTIVITY)
			//in the case of the adaptive version we have also incoming data (the reconfiguration commands) from the controller
			SWSR_Ptr_Buffer *cn_inqueue=data->cn_inqueue;
			//this particular monitoring data is used to notify that the reconfiguration is finished
			collector_monitoring_t reconf_finished;
			reconf_finished.tag=RECONF_FINISHED;
			reconf_data_c_t *reconf_data;
			bool reconf_phase_pard_down=false; //if it is true, it means that we are in a reconfiguration phase in which we have to terminate some worker
			char work_down_degree;
		#endif
	#endif

	expected_iid=(int64_t *)malloc(num_classes*sizeof(int64_t));
	for(int i=0;i<num_classes;i++)
		expected_iid[i]=window_size-1; //since the ids start from 0. The first window will be triggered only when full
	disordered=(int *)calloc(num_classes,sizeof(int));
	for(int i=0;i<num_classes;i++)
		buffer_disordered.emplace_back(5);

	//queues warmup
	for(int i=0;i<num_workers;i++)
		receive((void**)&rcvd,inqueue[i]);
	//synchronization barrier
	DEBUG(printf("Collector ready, running on %d\n",sched_getcpu()));
	pthread_barrier_wait(barrier);
	
	/**
		Start receiving from Workers
		NOTA: per ora polla i vari Workers, poi chi lo sa...
	*/
	
	while(*start_global_ticks==0)
		REPEAT_25(asm volatile("PAUSE" ::: "memory");) //wait for the correct time
	ticks start_gticks=*start_global_ticks;
	#if defined(MONITORING)
		monitoring_timer=*start_global_ticks+monitoring_step; 
	#endif
	last_print=getticks();
	start_ticks=last_print;
	printf("Collector, discrepancy: %Ld\n",getticks()-start_gticks);
	int mon_step=0;
	while(received_EOS<num_workers) //the collector has to receive the EOS from all the workers
	{
		//Round-robin polling from the various workers
		if(inqueue[index]->pop((void **)&rcvd)) 
		{
			//something has been received from a worker
			if(rcvd->isEOS) //it is an EOS
			{				
				#if defined(ADAPTIVITY)
				if(!reconf_phase_pard_down && cn_inqueue->pop((void **)(&reconf_data)))
				{
					//This is a particular case in which an EOS has arrived before the message from the controller
					if(reconf_data->par_degree_changes<0)					
					{
						//we should expect some EOS from them
						reconf_phase_pard_down=true;
						work_down_degree=reconf_data->par_degree_changes;
					}
				}
				//If we are in a reconfiguration phase in which we downgrade the parallelism degree
				//we should check if the EOS is arriving from a worker that has to be deleted				
				if(reconf_phase_pard_down && index>=num_workers+reconf_data->par_degree_changes)
				{

					work_down_degree++;
					//free the result buffer of the worker
					//printf("Collector libero per W: %d  res_buff: %p. Mancano %d EOS\n",index,rcvd->res_buff,work_down_degree);
					free(rcvd->res_buff);
					if(work_down_degree==0) //reconf finished
					{
						
						for(int i=0;i>reconf_data->par_degree_changes;i--) //(att: par_degree_cahnges<0)
						{
							//destroy the queues
							delete(inqueue[num_workers+i-1]); 
							//set to NULL outqueues (they will destructed by workers)
							inqueue[num_workers+i-1]=NULL;
						}
						
						//set the new num_workers
						num_workers+=reconf_data->par_degree_changes;
						reconf_phase_pard_down=false;
						//send ack to the controller
						bsend((void*)&reconf_finished,cn_outqueue);
						printf(ANSI_COLOR_YELLOW "[COLLECTOR] reconf finished\n" ANSI_COLOR_RESET);
					}

				}
				else //else received_EOS++
				#endif
				received_EOS++;
				//printf("Ricevuto EOS da %d\n",index);
			}
			else
			{
				//standard result coming from a Worker
				rcvd_results++;
				partial_rcvd++;
				recvd_per_worker[rcvd->wid]++;

				if((rcvd->id)>expected_iid[rcvd->type])
				{
					//we have received a partial result out of order
					//we have to buffer it and reuse when necessary
					//CURRENTLY IT IS IMPLEMENTED, BUT NOT INTENSIVELY TESTED

					//buffer it
					//fprintf(stderr,"Received partial result disordered for class: %d from Worker %d expected: %Ld received: %Ld \n",rcvd->type,index, expected_iid[rcvd->type],rcvd->id);
					//insert it into the right position (in order)
					buffer_disordered[rcvd->type].push_back(*rcvd);
					disordered[rcvd->type]++;
				}
				else
				{
					//IN REAL IMPLEMENTATIONS, the result will be sent afterward to the next module
					//...

					//compute latency statistics
					double lat=((double)(getticks()-rcvd->timestamp-start_gticks))/freq;
					//printf("TImestamp %Ld, computed: %Ld start: %Ld\n",rcvd->ts,rcvd->timestamp,start_gticks);
					// double lat=((double)((rcvd->ts-rcvd->timestamp)-start_gticks))/freq;

					//METRICS FOR RESPONSE TIME (just for testing)
					//all_lat.push_back(lat);
					//all_lat_times.push_back(rcvd->timestamp);
					//all_lat_worker.push_back(rcvd->wid);
					percentile_lat.push_back(lat);
					lat_per_class[rcvd->type]+=lat;
					recvd_per_class[rcvd->type]++;

					dlat+=lat;
					#if defined (MONITORING)
						monitoring[mon_ptr].results++;
						monitoring[mon_ptr].results_per_class[rcvd->type]++;	
						monitoring[mon_ptr].latency_per_class[rcvd->type]+=lat; //usec
						
					#endif
					//notice that here we will expect a result only for completed window (not just some slide)
					expected_iid[rcvd->type]+=window_slide;

					if(disordered[rcvd->type]>0) //previously we received disordered results: the idea is that we sent all the results
					{
						//look if there is the next in the vector
						for(int i=0;i<buffer_disordered[rcvd->type].size();i++)
						{
							if(buffer_disordered[rcvd->type][i].id==expected_iid[rcvd->type])
							{
								//ok it is the right one
								//send afterwards
								//....

								double lat=((double)(getticks()-buffer_disordered[rcvd->type][i].timestamp-start_gticks))/freq;
								dlat+=lat;
								#if defined (MONITORING)
									monitoring[mon_ptr].results++;
									monitoring[mon_ptr].results_per_class[rcvd->type]++;
									monitoring[mon_ptr].latency_per_class[rcvd->type]+=lat; //usec
								#endif
								expected_iid[rcvd->type]+=window_slide;
								//erase the element
								buffer_disordered[rcvd->type].erase(buffer_disordered[rcvd->type].begin()+i);
								disordered[rcvd->type]--;
								//reset i to zero, we have to start from the beginning
								i=0;
								//fprintf(stderr,"Received partial result out of order for class: %d, problem resolved!\n",rcvd->type);
							}
						}
					}
				}
			}
		}
		
		//check if we have to print
		if(getticks()-last_print>print_rate)
		{
			double avglat=dlat/partial_rcvd;
			int indexp;
			//sort percentiles
			if(percentile_lat.size()>0)
			{
				sort(percentile_lat.begin(), percentile_lat.end()); 
				indexp=percentile_lat.size()*0.95;

				printf(ANSI_COLOR_GREEN "Time: %6.3f, recvd results: %Ld, avg latency (usec): %6.3f 95-percentile: %6.3f (top: %6.3f)\n" ANSI_COLOR_RESET,((double)(getticks()-start_ticks)/(freq*1000000)),partial_rcvd,avglat,percentile_lat[indexp],percentile_lat[percentile_lat.size()-1]);
			}
			else
				printf("Time: %6.3f, recvd results: %Ld, avg latency (usec): %6.3f\n",((double)(getticks()-start_ticks)/(freq*1000000)),partial_rcvd,avglat);
			

			//throughput per worker
			printf(ANSI_COLOR_GREEN "Th x W: ");
			for(int i=0;i<num_workers;i++)
				printf("%d\t",recvd_per_worker[i]);
			printf(ANSI_COLOR_RESET "\n");

			if(getticks()-start_ticks>no_more_init)
			{
				//check latency for termination if bottleneck
				if(avglat>MAX_LAT)
				{	//non basta dobbiamo mettere qualcos'altro
					//si tengono le ultime 5 e si fa la media da 5 a 5?
					fprintf(stderr, ANSI_COLOR_RED "Collector found that the latency is too much high\n" ANSI_COLOR_RESET);
					//print to file the latencies
					char out_file[100];
					sprintf(out_file,"%s_%s.dat",LAT_FILE,suffix);
					FILE* flat=fopen(out_file,"w");
					fprintf(flat,"#Second Num_results Latency\n");
					for(int i=0;i<num_latencies;i++)
					{
						fprintf(flat,"%3.3f\t%Ld\t%6.3f\n",seconds[i],lat_recv[i],latencies[i]);
					}
					fclose(flat);
					exit(BOTTLENECK_ERR);
				}
			}
			last_print=getticks();
			//for print to file
			latencies[num_latencies]=avglat;
			par_degrees[num_latencies]=num_workers;
			if(percentile_lat.size()>0)
				percentiles[num_latencies]=percentile_lat[indexp];
			else
				percentiles[num_latencies]=0.0;
			lat_recv[num_latencies]=partial_rcvd;
			seconds[num_latencies]=((double)(getticks()-start_ticks)/(freq*1000000));
			num_latencies++;
			partial_rcvd=0;
			dlat=0.0;
			percentile_lat.clear();
			memset(recvd_per_worker,0,num_workers*sizeof(int));
		}

		#if defined(MONITORING)
			//check if there are reconfiguration messages from the controller
			#if defined(ADAPTIVITY)
				if(cn_inqueue->pop((void **)(&reconf_data)))
				{
					
					if(reconf_data->par_degree_changes>0)
					{
						//if there are newly spanned threads, copy their queues
						printf(ANSI_COLOR_YELLOW "[COLLECTOR] ricevuto messaggio, incremento par degree di %d worker\n" ANSI_COLOR_RESET,reconf_data->par_degree_changes);
						//take the new queues
						for(int i=0;i<reconf_data->par_degree_changes;i++)
							inqueue[num_workers+i]=reconf_data->wqueues[i];
						//increments the par degree
						num_workers+=reconf_data->par_degree_changes;
						//ok, notify to the collector that the reconfiguration started (we don't need a specific message, since it is a well defined req-reply protocol)
						bsend((void*)&reconf_finished,cn_outqueue);
						/*for(int i=0;i<NMONITORING;i++)
						{
							printf("%d %p\n",i,monitoring[i].latency_per_class);
						}*/
					}
					else
						if(reconf_data->par_degree_changes<0)
						{
							printf(ANSI_COLOR_YELLOW "[COLLECTOR] ricevuto messaggio, decremento par degree di %d worker\n" ANSI_COLOR_RESET,reconf_data->par_degree_changes);
							//we should expect some EOS from them
							reconf_phase_pard_down=true;
							work_down_degree=reconf_data->par_degree_changes;
						}
				}
			#endif

			//check if we can send monitoring data to controller
			if(getticks()>monitoring_timer)	
			{ 	
				//MAYBE SOME OF THESE ARE USELESS
				asm volatile("":::"memory");
				// printf("Monitoring data %d are valid. Elements received: %Ld\n",mon_ptr,monitoring[mon_ptr].elements);
				//compute average latencies
				for(int i=0;i<num_classes;i++)
					monitoring[mon_ptr].latency_per_class[i]=((monitoring[mon_ptr].latency_per_class[i])/(monitoring[mon_ptr].results_per_class[i]));
				monitoring[mon_ptr].monitoring_time=getticks();
				//blocking send towards the collector
				bsend(&monitoring[mon_ptr],cn_outqueue);
				//printf("Collector sent step: %d, lqueue:%d\n",mon_step++,cn_outqueue->length());
				monitoring_timer=getticks();
				mon_ptr=(mon_ptr+1)%NMONITORING;
				//reset the monitoring data
				monitoring[mon_ptr].results=0;
				memset(monitoring[mon_ptr].latency_per_class,0,num_classes*sizeof(double));
				memset(monitoring[mon_ptr].results_per_class,0,num_classes*sizeof(int));
				monitoring_timer+=monitoring_step; 
			}
		#endif
		index=(index+1)%num_workers;
		asm volatile("PAUSE" ::: "memory");
	}
	
	//last print
	double avglat=dlat/partial_rcvd;
	//order percentile
	if(percentile_lat.size()>0)
	{
		sort(percentile_lat.begin(), percentile_lat.end()); 
		int indexp=percentile_lat.size()*0.95;

		printf(ANSI_COLOR_GREEN "Time: %6.3f, recvd results: %Ld, avg latency (usec): %6.3f 95-percentile: %6.3f (top: %6.3f)\n" ANSI_COLOR_RESET,((double)(getticks()-start_ticks)/(freq*1000000)),partial_rcvd,avglat,percentile_lat[indexp],percentile_lat[percentile_lat.size()-1]);
		percentiles[num_latencies]=percentile_lat[indexp];
	}
	else
		printf("Time: %6.3f, recvd results: %Ld, avg latency (usec): %6.3f\n",((double)(getticks()-start_ticks)/(freq*1000000)),partial_rcvd,avglat);
					
	
	last_print=getticks();
	//for print to file
	latencies[num_latencies]=avglat;
	lat_recv[num_latencies]=partial_rcvd;
	
	seconds[num_latencies]=((double)(getticks()-start_ticks)/(freq*1000000));
	num_latencies++;
	#if defined(MONITORING)
	//send the last monitoring, just for terminating the controller
	bsend(&monitoring[mon_ptr],cn_outqueue);
	#endif
	//print to file the latencies
	char out_file[100];
	sprintf(out_file,"%s_%s.dat",LAT_FILE,suffix);
	FILE* flat=fopen(out_file,"w");
	fprintf(flat,"#Second Num_results Latency Percentile NW\n");
	for(int i=0;i<num_latencies;i++)
	{
		fprintf(flat,"%3.3f\t%Ld\t%6.3f\t%6.3f\t%d\n",seconds[i],lat_recv[i],latencies[i],percentiles[i],par_degrees[i]);
	}
	fclose(flat);
	/*flat=fopen("all_lat.dat","w");
	 fprintf(flat,"#Second Latency WID\n");
	for(int i=0;i<all_lat.size();i++)
	{
		fprintf(flat,"%6.4f %6.4f %d\n",all_lat_times[i]/(freq*1000.0),all_lat[i],all_lat_worker[i]);
	}
	fclose(flat);*/
	//for(int i=0;i<num_classes;i++)
	//	printf("%d %Ld\n",i,recvd_per_class[i]);
	 /*flat=fopen("lat_per_class","w");
	}
	fprintf(flat,"Class Rcvd Latency(avg)\n");
	for(int i=0;i<num_classes;i++)
	{
		fprintf(flat,"%d\t%d\t%6.3f\n",i,recvd_per_class[i],lat_per_class[i]/recvd_per_class[i]);
	}
	fclose(flat);
	*/
	/*FILE* fcomp=fopen("tmp","w");
	fprintf(fcomp,"#Class Num_computation AvgComputationTime (usec)\n");
	for(int i=0;i<num_classes;i++)
	{
		fprintf(fcomp,"%d\t%Ld\t\n",i,class_rcvd[i]);
	}
	fclose(fcomp);*/
	DEBUG(printf("Collector terminated, received results: %Ld\n",rcvd_results));
	//pthread_barrier_wait(barrier);

	return NULL;
}
