/**
	Controller for Functional Partitioning implementation
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <ff/buffer.hpp>
#include <ff/allocator.hpp>	
#include <math.h>

#include <iostream>
#include <unordered_map>
#include <vector>
#include <sched.h>
#include "../Includes/RandomGenerator.h"
#include "../Includes/cycle.h"
#include "../Includes/iaperf.h"
#include "../Includes/general.h"
#include "../Includes/farm_cb.h"
#include "../Includes/cbwindow.hpp"
#include "../Includes/farm_cb_reconf.h"

#define MAX_RHO 1.1 //maximum rho sustainable --->problema sulle fasi di settling time
#define MAX_RHO_MODULE 0.9 //the maximum rho that the module can have before being considered bottleneck
using namespace ff;
using namespace std;
int max_workers;
/**
	Receive on FF queues. If there is more than one element, returns the last inserted
*/
inline void receiveLast ( void **t, ff::SWSR_Ptr_Buffer *inqueue) __attribute__((always_inline))/* __attribute__((noinline))*/; 
inline void receiveLast ( void **t, ff::SWSR_Ptr_Buffer *inqueue)
{
	do
	{
		while (!inqueue->pop( t)) {
			// Small sleep
			REPEAT_25(asm volatile("PAUSE" ::: "memory");)
		}
	}while(!inqueue->empty());
}

double compute_st(long double qj,long double lambda, long double tcalc)
{
	if(1/tcalc<lambda)
	{
		printf(ANSI_COLOR_RED_BOLD "Settling time is not computable!!\n" ANSI_COLOR_RESET);
		return 0;
	}
	long double t=0;
	printf("Start computing st: %Lf %Lf %Lf\n",qj,lambda,tcalc);
	while(qj>=1)
	{
		//printf("S: %f T:%f\n",qj*tcalc*lambda,qj*tcalc);
		t+=qj*tcalc;
		qj=qj*tcalc*lambda;
	}
	return t;
}

/**
	Compute fully balanced scheduling table
	some argument are modified since they are helpful
	PARAM:
		- num_workes, num_classes;
		- load_w: it will be modified in order to contains the load per worker (intended as the sum of the product class_frequency*tcalc_class for all the class assigned to the worker)
		- wtcalc_per_class: it contains for each class the product  class_frequency*tcalc_class, computed using the monitored data
		- scheduling_table: the past scheduling table that will be modified
*/
void compute_fb_st(int num_workers,int num_classes,double *load_w,double* wtcalc_per_class, char *scheduling_table)
{
	char w=0;
		//true if a certain class it has been assigned
	bool* assigned=(bool *)calloc(num_classes,sizeof(bool));
	memset(load_w,0,sizeof(double)*max_workers);
	
	int ass_work[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	
	for(int i=0;i<num_classes;i++)
	{
		//DA OTTIMIZZARE
		//find max class not already assigned
		double max_p=0;
		int class_index=0;
		for(int j=0;j<num_classes;j++)
		{
			if(wtcalc_per_class[j]>=max_p && !assigned[j])
			{
				// printf("new max: %6.3f\n",wtcalc_per_class[j]);
				class_index=j;
				max_p=wtcalc_per_class[j];
			}
		}
		// printf("max: %d\n",class_index);
		assigned[class_index]=true;

		//assign to the less load worker (ASSUMPTION: NONE OF THEM WILL BE BOTTLENECK)
		double min_w=load_w[0];
		int work_index=0;
		for(int j=0;j<num_workers;j++)
		{
			if(load_w[j]<min_w)
			{
				work_index=j;
				min_w=load_w[j];
			}
		}
		// printf("class: %d assigned to: %d\n",class_index,work_index);
		scheduling_table[class_index]=work_index+1;
		load_w[work_index]+=wtcalc_per_class[class_index];

		ass_work[work_index]++;

	}


}

void *controller (void *args)
{
	//init
	controller_data_t *data = (controller_data_t *) args;
	int num_workers=data->num_workers;
	int *affinities=data->affinities;
	int num_classes=data->num_classes;
	SWSR_Ptr_Buffer *e_inqueue=data->e_inqueue;
	SWSR_Ptr_Buffer **w_inqueue=data->w_inqueue;
	SWSR_Ptr_Buffer *c_inqueue=data->c_inqueue;
	SWSR_Ptr_Buffer *e_outqueue=data->e_outqueue;
	SWSR_Ptr_Buffer *c_outqueue=data->c_outqueue;
	char *suffix=data->suffix;
	long int freq=data->freq;
	int window_size=data->window_size;
	int window_slide=data->window_slide;
	max_workers=data->max_workers;
	ticks *start_global_ticks=data->start_global_ticks;
	#if !defined(DONT_USE_FFALLOC) && !defined(TASK_BUFF)
		ff_allocator *ffalloc=data->ffalloc;
	#endif
	cpu_set_t cpuset;

	emitter_monitoring_t *em;
	worker_monitoring_t **wm=(worker_monitoring_t**)malloc(sizeof(worker_monitoring_t*)*max_workers);
	collector_monitoring_t *cm; 
	int64_t monitoring_step=0;
	reconf_repository_t *repo=(reconf_repository_t *)data->repo;
	DEBUG(printf("Controller ready, running on: %d, freq: %Ld\n",sched_getcpu(),freq))
	//The controller will be stopped by the emitter
	bool stop=false;
	#if defined(QSPLITTED)
		printf("Occhio, nel controllore non gestiamo qsplitted\n");
	#endif
	//NOTA: da capire come gestire questi messaggi di configurazione (quanto ci basta le assunzioni sul ciclo chiuso tra emettitore  e controllore)
	//NOTE: for the moment being we will use only one variable (per entity) for reconfigurations, under the assumption that reconfigurations will be concluded in 
	//finite time
	reconf_data_em_t *reconf_data_em;
	posix_memalign((void **)&reconf_data_em,CACHE_LINE_SIZE,sizeof(reconf_data_em_t));
	posix_memalign((void **)&reconf_data_em->scheduling_table,CACHE_LINE_SIZE,sizeof(char)*num_classes);
	reconf_data_em->par_degree_changes=0;
	reconf_data_c_t *reconf_data_c;
	posix_memalign((void **)&reconf_data_c,CACHE_LINE_SIZE,sizeof(reconf_data_c_t));
	reconf_data_c->par_degree_changes=0;
	worker_data_t *worker_data; //for thread spawning
	pthread_t *wtids=(pthread_t *)malloc(sizeof(pthread_t)*(max_workers));
	char tid_idx=0;
	int threads_spawned=0; //number of thread spawned (that have to be checked on)
	//TMP: per il momento facciamo una serie di stampe su file
	int *classes_freq=(int *)calloc(num_classes,sizeof(int));
	double *comp_time=(double *)calloc(num_classes,sizeof(double));
	int *results=(int *)calloc(num_classes,sizeof(int));
	int attempt;
	bool exiting=false;
	int n_opt=0;
	//DATA STRUCTURES USED FOR COMPUTING THE SCHEDULING TABLE
	//this will contain the calculation time for each class, updated with the values coming from workers
	double* tcalc_per_class=(double *)calloc(num_classes,sizeof(double));
	// weighted tcalc_per_class it will containt the product freq(i)*tcalc_per_class(i) (where i is a class)
	double* wtcalc_per_class=(double *)calloc(num_classes,sizeof(double));
	//load per worker (it depends on the class that will be assigned to it)
	double* load_w=(double *)calloc(max_workers,sizeof(double));

	//current load factor for the various workers (it will consider only the triggering tuples)
	double *rho=(double*)calloc(max_workers,sizeof(double));
	//average tcalc per worker (considering only the slide computations)
	double *wtcalc_per_worker=(double *)calloc(max_workers,sizeof(double));

	//expect rho: it will be useful for the strategy and to compute the settling time 
	double *erho=(double*)calloc(max_workers,sizeof(double));
	//expected tcalc per worker
	double *etcalc=(double*)calloc(max_workers,sizeof(double));
	//frequencies of the classes assigned to a particular worker
	double *freq_to_worker=(double*)calloc(max_workers,sizeof(double));

	//the various performance metrics computed upon monitored data
	double trigger_per_second;//number of tuple that trigger a slide per second
	double additional_elements; //the number of additional elements received by the emitter and not yet counted by it
	double tta_msec; //trigger interarrival time (in msec)
	double module_tcalc; //the tcalc of the whole parallel module (in seconds)
	
	//variables for computing the settling time
	long double q,qj,lambda,tcalc,st;			


	//for metrics info
	vector<char> par_degrees; //vector that contains the par degrees
	vector<ticks> time_par_degrees; //and the relative times
	vector<double> reconf_par_degrees; //reconfiguration times

	//prendiamo il tempo, utile per alcuni test
	struct timeval tmp_t;
	
	long int start_t;
	long int reconf_start_t;
	bool mess_sent_degree=true;
	bool mess_sent_scheduling=true;
	bool start_time_taken=false;
	while(!stop)
	{
		//PER IL MOMENTO FACCIAMO UNA LETTURA ROUND ROBIN
		//receive from emitter
		//NOTA PER ORA SONO BLOCCANTI: FORSE DOVREMO POI RENDENDERLE NON BLOCCANTI
		receiveLast((void **)&em,e_inqueue);
		// printf("Controller ricevuto da emitter: %d\n",em->mon_id);
		if(!start_time_taken)
		{
			start_time_taken=true;
			gettimeofday(&tmp_t,NULL);
			start_t=tmp_t.tv_sec*1000+(tmp_t.tv_usec)/1000;
		}

		//printf("Contr: inizio ciclo %Ld\n",monitoring_step);
		if(em->stop)
		{
			stop=true;
			//printf("Controller, ricevuto STOP da Emitter\n");
		}
		else //facciamo attenzione (non vorrei che ci fosse un deadlock)
		{
			//receive from workers
			for(int i=0;i<num_workers;i++)
			{
				receiveLast((void **)&wm[i],w_inqueue[i]);	
				// printf("Controller ricevuto da worker: %d\n",i);
				//if some of them is equal to 1, it means that we are exiting from the program
				//this is the only occasion in which this event can occur
				if(wm[i]==(void*)(1)) //EOS ARRIVED
				{
					exiting=true;
					break;
				}
			
			}
			if(exiting)
			{
				printf("Controller EXITING!");
				break; //come back to emitter data reading (we are exiting...)
			}
			
			receiveLast((void **)&cm,c_inqueue);
			//TMP: contiamo le varie frequenze
			//NOTA QUESTO CONTEGGIO PUO' non essere preciso perche' saltiamo qualche messaggio
			//STATISTICS
			for(int i=0;i<num_classes;i++)
			{
				classes_freq[i]+=em->elements_per_class[i];
					
				for(int j=0;j<num_workers;j++)
				{
					if(wm[j]!=(void *)1) //it is valid
					{
						//just for info
						comp_time[i]+=wm[j]->tcalc_per_class[i];
						results[i]+=wm[j]->computations_per_class[i];
					}
					
				}				
				
			}

			#if defined(ADAPTIVITY)

			//compute the inter arrival time (on a per-second basis) of trigger tuples
			//A trigger tuple is the tuple that closes a slide and start the computation
			//for precision problem we will refer to KILO TUPLE PER SECOND (while referring to rate)
			//and therefore to MSEC (when referring to interarrival)
			//NOTE: many metrics are reported on a per second basis, just multiplying for the number of monitoring step contained in a second 
			//(that is 1000/MONITORING_STEP)

			
			if(em->grow_rate_per_step==0) //emitter found that socket buffer is stable
				additional_elements=(double)em->buffer_elements;
			else//the socket buffer is growing
				additional_elements=em->grow_rate_per_step;
			//printf("Controller additional elements: %g, elements monitored by emitter: %d\n",additional_elements,em->elements);
			trigger_per_second=((em->elements+additional_elements)*1000.0/MONITORING_STEP)/window_slide;
			
			tta_msec=1000.0/trigger_per_second;
			//compute the average tcalc (considering the class frequencies)
			module_tcalc=0;
			for(int i=0;i<num_workers;i++)
			{
				rho[i]=0;
				freq_to_worker[i]=0;
				wtcalc_per_worker[i]=0;
			}
			for(int i=0;i<num_classes;i++)
			{
				//prendere il worker a cui era stata assegnata
				char assw=em->scheduling_table[i]-1;
				if(wm[assw]->tcalc_per_class[i]>0)
				{
					//average tcalc
					tcalc_per_class[i]=(wm[assw]->tcalc_per_class[i]/wm[assw]->computations_per_class[i])/1000;
					//weigth it: we are considered the frequencies measured by the emitter (that is up to the max load sustainable with this configuration)
					wtcalc_per_class[i]=(((double)em->elements_per_class[i])/em->elements)*tcalc_per_class[i];
					// wtcalc_per_class[i]=(((double)em->elements_per_class[i])/em->elements)*(1000/MONITORING_STEP)*tcalc_per_class[i];
					freq_to_worker[assw]+=(((double)em->elements_per_class[i])/em->elements);
					//compute the averaged tcalc considering only the elements received by the worker
					wtcalc_per_worker[assw]+=tcalc_per_class[i]*((double)wm[assw]->elements_per_class[i]/wm[assw]->elements_rcvd); //in realta dovremmo dividere entrambi per 500 ma siamo li
					//comp_per_class+=wm[assw]->computations_per_class[i];
					//count it in the whole weighted tcalc
					module_tcalc+=wtcalc_per_class[i];
				}
			}
			//compute the load of the various workers
			for(int i=0;i<num_workers;i++)
			{
				rho[i]=wtcalc_per_worker[i]/1000; //in second
				rho[i]/=1/((1000.0/MONITORING_STEP)*(wm[i]->computations+(additional_elements*freq_to_worker[i])/window_slide));//normalize it to a second basis
				//printf("Worker %d rho %f, tt ricevute %d , tcalc %f\n",i,rho[i],wm[i]->elements_rcvd/window_slide,wtcalc_per_worker[i]);
			}

		

			//compute optimal parallelism degree
			if(module_tcalc>0 && tta_msec>0) //data are valid
				n_opt=ceil(module_tcalc/(tta_msec*MAX_RHO_MODULE));
			else
				n_opt=num_workers;

			//some checks
			if(n_opt<=0) //(something is going wrong... n_opt must be at least 1)
			{
				printf(ANSI_COLOR_RED_BOLD "Humh...something goes wrong: n_opt must be at least 1\n",ANSI_COLOR_RESET);
				n_opt=1;
			}
			if(n_opt>max_workers) //we have not sufficient resource
				n_opt=max_workers;

			printf(ANSI_COLOR_BLUE "[Controller] Global TA(msec): %6.3f Rate (KTT/s):%.3f, Tcalc: %6.3f, Nopt: %d\n" ANSI_COLOR_RESET,tta_msec,1000/tta_msec,module_tcalc,n_opt);
			
			gettimeofday(&tmp_t,NULL);
			long int end_t=tmp_t.tv_sec*1000+(tmp_t.tv_usec)/1000-start_t;
		
			//after some initial step of monitoring we evaluate the strategy
			if(monitoring_step>1)
			{	
				//qui dovresti checkare lo sbilanciamento sui worker
				int changes=n_opt-num_workers;
				int prev_num_workers=num_workers;
				threads_spawned+=changes;
				if(changes!=0)
					printf(ANSI_COLOR_YELLOW "[CONTROLLER] spawning %d workers\n" ANSI_COLOR_RESET,changes);

				if(changes==0) 
				{
					//SE SIAMO IN SETTLING QUI NON DOVREMMO ENTRARE

					//siamo ok in teoria col numero di worker, ma potrebbe sussistere una situazione di sbilanciamento
					//che porta ad essere collo di bottiglia
					bool reconfigure=false;
					for(int i=0;i<num_workers;i++)
					{
						if(rho[i]>MAX_RHO)
							reconfigure=true;
					}
					if(reconfigure)
						printf("Mi dovrei riconfigurare per sbilanciamento\n");
					/*if(reconfigure)
					{
						
							gettimeofday(&tmp_t,NULL);
							reconf_data_em->par_degree_changes=0;
							memcpy(reconf_data_em->scheduling_table,em->scheduling_table,sizeof(char)*num_classes);
							printf(ANSI_COLOR_CYAN "[Controller] Rebalancing classes... \n" ANSI_COLOR_RESET);
							compute_fb_st(num_workers, num_classes,load_w,wtcalc_per_class,reconf_data_em->scheduling_table);
							
							reconf_start_t=tmp_t.tv_sec*1000000+(tmp_t.tv_usec);
							bsend(reconf_data_em,e_outqueue);
							

							//wait for the reply from the emitter
							//ATTENZIONE QUESTA INTERAZIONE E' UN PO' TRABALLANTE
							do
							{
								receive((void **)&em,e_inqueue);
								// printf("Controller received: %p, with tag:%d\n",em,em->tag);
								//fprintf(stderr,"Reconfiguration error\n");
								//exit(RECONF_ERR);
							}while((em->tag!=RECONF_FINISHED));
							//QUI DOVREMMO GARANTIRE CHE UNA NUOVA RICONFIGURAZIONI NON INIZI PRIMA CHE LA PRECEDENTE TERMINI
							for(int i=0;i<num_workers;i++)
							{
								while(repo->reconfiguration_finished[i]==0)
								{
									asm volatile("PAUSE" ::: "memory");
								}
							}
							gettimeofday(&tmp_t,NULL);
							printf(ANSI_COLOR_CYAN "[Controller] reconfiguration finished in %Ld usecs\n" ANSI_COLOR_RESET,tmp_t.tv_sec*1000000+(tmp_t.tv_usec)-reconf_start_t);
					}*/
						

				}

				
				if(changes>0) //increase the number of workers
				{
					//changes=max_workers-num_workers;

					gettimeofday(&tmp_t,NULL);
					reconf_start_t=tmp_t.tv_sec*1000000+(tmp_t.tv_usec);
					memcpy(reconf_data_em->scheduling_table,em->scheduling_table,sizeof(char)*num_classes);
					//queues creation (E->W,W->C, W->Contr)
					reconf_data_em->wqueues=(SWSR_Ptr_Buffer **)malloc(sizeof(SWSR_Ptr_Buffer *)*changes);
					reconf_data_c->wqueues=(SWSR_Ptr_Buffer **)malloc(sizeof(SWSR_Ptr_Buffer *)*changes);
					worker_data=(worker_data_t *)malloc(sizeof(worker_data_t)*changes);
					reconf_data_em->par_degree_changes=changes;
					reconf_data_c->par_degree_changes=changes;
					for(int i=0;i<changes;i++)	
					{
						reconf_data_em->wqueues[i]=new SWSR_Ptr_Buffer(QUEUE_SIZE);
						reconf_data_em->wqueues[i]->init();
						reconf_data_c->wqueues[i]=new SWSR_Ptr_Buffer(QUEUE_SIZE);
						reconf_data_c->wqueues[i]->init();
						//assuming that there is enough space, define queue worker->controller
						w_inqueue[num_workers+i]=new SWSR_Ptr_Buffer(QUEUE_SIZE_MON);
						w_inqueue[num_workers+i]->init();
					}
					//thread spawning
					for(int i=0;i<changes;i++)
					{
						pthread_t tid;
						worker_data[i].workerId = i+num_workers;
						worker_data[i].inqueue = reconf_data_em->wqueues[i];
						worker_data[i].outqueue = reconf_data_c->wqueues[i];
						worker_data[i].barrier=NULL; //in this way, newly spawned threads will not perform wait on barrier 
						worker_data[i].window_size=window_size;
						worker_data[i].window_slide=window_slide;
						worker_data[i].freq=freq;
						worker_data[i].num_classes=num_classes;
						worker_data[i].start_global_ticks=start_global_ticks;
						#if !defined(DONT_USE_FFALLOC) && !defined(TASK_BUFF)
							worker_data[i].ffalloc=ffalloc;
						#endif
					
						worker_data[i].cn_outqueue=w_inqueue[num_workers+i];
						worker_data[i].repo=repo;						
						pthread_create(&tid, NULL, worker, &(worker_data[i]));
						//set CPU affinity of worker threads
						CPU_ZERO(&cpuset);
						CPU_SET(affinities[i+num_workers], &cpuset);

						if (pthread_setaffinity_np(tid, sizeof(cpu_set_t), &cpuset)) {
							cerr << "Cannot set thread to CPU " << i << endl;
						}
						//worker_data[i].num_workers = num_workers;
						wtids[tid_idx] = tid;
						tid_idx++;
					}
					
					
					num_workers+=changes;
					//compute the optimal scheduling table and the expected load
					compute_fb_st(num_workers, num_classes,load_w,wtcalc_per_class,reconf_data_em->scheduling_table);
					memset(freq_to_worker,0,sizeof(double)*num_workers);
					//compute the new frequency of arrival to each worker
					for(int i=0;i<num_classes;i++)
						freq_to_worker[reconf_data_em->scheduling_table[i]-1]+=((double)em->elements_per_class[i])/em->elements;


					//COMPUTE THE EXPECTED SETTLING TIME
					//NOTE: DEVE ESSERE FATTO QUALCOSA PER RICONFIGURAZIONI SUCCESSIVE <-----------------

					for(int i=0;i<num_workers;i++)
					{
						//compute the expected tcalc and expected rho
						etcalc[i]=(load_w[i])/freq_to_worker[i];
						erho[i]=etcalc[i]/((tta_msec)/freq_to_worker[i]);//ATTENZIONE ALLE FORMULE
						//printf("rho: %d %f (loadw: %f etcalc: %f tta_msec: %f freq2w:%f)\n",i,erho[i],load_w[i],etcalc[i],tta_msec,freq_to_worker[i]);
					}


					//find the worker that now will be the most overloaded
					char max_w=0;
					for(int i=1;i<num_workers;i++)
					{
						if(erho[max_w]<erho[i])
							max_w=i;
					}
					//di questo conosciamo la somma dei fi (Sta in freq_to_work)
					
					//considerando solo l'ultimo monitoring step...
					//quante tuple si sono incodate?
					//devo conoscere il worker che prima era il piu' carico....
					char prev_max=0;
					for(int i=1;i<prev_num_workers;i++){
						if(rho[prev_max]<rho[i])
							prev_max=i;
					}
					//ora devo capire quanto traffico non e' stato gestito per sapere quante tuple sono incodate
					//lo faccio considerando il tcalc medio del worker con rho maggiore
					//gia' consideriamo anche il tempo della riconfigurazione
					//POTREMMO MONITORARE ANCHE I TEMPI DI RICONFIGURAZIONE PER AVERE UN IDEA PIU' PRECISA
					q=(MONITORING_STEP/1000+0.15)*(1/(tta_msec/1000)-prev_num_workers/(wtcalc_per_worker[prev_max]/1000)); // anche il wtcalc andrebbe aggiornato con  il tempo necessario a gestire le tuple non trigger
					
					printf("Tuple incodate nell'ultimo monitoring step: %Lf (rate effettivo: %f, rate gestito (%d): %f)\n",q,1/(tta_msec/1000),prev_num_workers,prev_num_workers/(wtcalc_per_worker[prev_max]/1000));
					//ora calcoliamo il settling time
					qj=q*freq_to_worker[max_w]; //triggering tuple incodate e afferenti al worker che ora sara' piu' carico
					lambda=(1000.0/tta_msec)*freq_to_worker[max_w]; //rate di interarrivo al worker
				
					tcalc=etcalc[max_w]/1000;
					//tcalc+=0.00008;
					st=compute_st(qj,lambda,tcalc);
					printf(ANSI_COLOR_BLUE_REVERSE "[Controller] Settling time: %Lf (max: %d fre max worker: %f)\n" ANSI_COLOR_RESET,st,max_w,freq_to_worker[max_w]);

					//send the data toward the emitter
					bsend(reconf_data_em,e_outqueue);

					//communication toward the collector
					bsend(reconf_data_c,c_outqueue);


					//wait for the reply from the emitter
					do
					{
						receive((void **)&em,e_inqueue);
					}while((em->tag!=RECONF_FINISHED));
					// printf("Controller: emitter ha finito\n");
					do
					{
						receive((void **)&cm,c_inqueue);
						
					}while((cm->tag!=RECONF_FINISHED));
					// printf("Controller: collector ha finito\n");
					//QUI DOVREMMO GARANTIRE CHE UNA NUOVA RICONFIGURAZIONI NON INIZI PRIMA CHE LA PRECEDENTE TERMINI
					for(int i=0;i<num_workers;i++) //take into account also the newly spawned threds
					{
						while(repo->reconfiguration_finished[i]==0)
						{
							asm volatile("PAUSE" ::: "memory");
						}
					}
					gettimeofday(&tmp_t,NULL);
					printf(ANSI_COLOR_YELLOW "[CONTROLLER] increased par degree in %Ld usecs (current global time is: %.3f) \n" ANSI_COLOR_RESET,tmp_t.tv_sec*1000000+(tmp_t.tv_usec)-reconf_start_t,((double)(getticks()-*start_global_ticks)/(freq*1000000)));
					//STATISTICS 
					par_degrees.push_back(num_workers);
					time_par_degrees.push_back(getticks());
					reconf_par_degrees.push_back(tmp_t.tv_sec*1000000+(tmp_t.tv_usec)-reconf_start_t);
				}
				if(changes<0) //we have to erase some worker
				{
					gettimeofday(&tmp_t,NULL);
					reconf_start_t=tmp_t.tv_sec*1000000+(tmp_t.tv_usec);
					memcpy(reconf_data_em->scheduling_table,em->scheduling_table,sizeof(char)*num_classes);
					
					reconf_data_em->par_degree_changes=changes;
					reconf_data_c->par_degree_changes=changes;
					num_workers+=changes;
					
					//compute the new scheduling table
					compute_fb_st(num_workers, num_classes,load_w,wtcalc_per_class,reconf_data_em->scheduling_table);	
					
					//send the data toward the emitter
					bsend(reconf_data_em,e_outqueue);

					//communication toward the collector
					bsend(reconf_data_c,c_outqueue);

					//wait for the reply from the emitter
					//ATTENZIONE QUESTA INTERAZIONE E' UN PO' TRABALLANTE
					do
					{
						receive((void **)&em,e_inqueue);
						//fprintf(stderr,"Reconfiguration error\n");
						//exit(RECONF_ERR);
					}while((em->tag!=RECONF_FINISHED));
					//printf("Controller: emitter ha finito\n");
					do
					{
						receive((void **)&cm,c_inqueue);
					}while((cm->tag!=RECONF_FINISHED));
					//printf("Controller: collector ha finito\n");
					//QUI DOVREMMO GARANTIRE CHE UNA NUOVA RICONFIGURAZIONI NON INIZI PRIMA CHE LA PRECEDENTE TERMINI
					for(int i=0;i<num_workers;i++) //take into account only the remained worker
					{
						while(repo->reconfiguration_finished[i]==0)
						{
							asm volatile("PAUSE" ::: "memory");
						}
						
					}

					for(int i=num_workers;i<num_workers-changes;i++)
					{
						//receive the last message (while exiting) by the worker
						do
						{
							receiveLast((void **)&wm[i],w_inqueue[i]);	
						}while(wm[i]!=(void*)(1));
						//free queue from workers
						delete(w_inqueue[i]);
						w_inqueue[i]=NULL;
						if(tid_idx>0)
							tid_idx--;
					}
					

					
					gettimeofday(&tmp_t,NULL);
					printf(ANSI_COLOR_YELLOW "[CONTROLLER] decreased par degree in %Ld usecs\n" ANSI_COLOR_RESET,tmp_t.tv_sec*1000000+(tmp_t.tv_usec)-reconf_start_t);

					par_degrees.push_back(num_workers);
					time_par_degrees.push_back(getticks());
					reconf_par_degrees.push_back(tmp_t.tv_sec*1000000+(tmp_t.tv_usec)-reconf_start_t);
				}
				
			}
			#endif

		}
		monitoring_step++;		
	}
	//wait for the threads that have been spawned
	// Wait for the completion of the threads
	//void *retval ;
	//for (int i = 0; i < tid_idx; i++) {
	//	pthread_join(wtids[i], &retval);
	//}
	//print to file classes statistics
	char out_file[100];
	sprintf(out_file,"m%s_%s.dat",CLASS_FILE,suffix);
	FILE* fclass=fopen(out_file,"w");
	fprintf(fclass,"#Number of classes: %d\n#",num_classes);
	fprintf(fclass, "#CLASS_ID FREQ ASSIGNED_WORKER COMP_TIME\n");
	
	for(int i=0;i<num_classes;i++)
		fprintf(fclass,"%d\t%d\t%d\t%6.3f\n",i,classes_freq[i],results[i],comp_time[i]/results[i]);
	 
	fclose(fclass);
	char out_file2[100];
	sprintf(out_file2,"%s_%s.dat","par_degree",suffix);
	FILE *fpar=fopen(out_file2,"w");
	fprintf(fpar,"#Time Par_degree Reconf_time\n");
	for(int i=0;i<par_degrees.size();i++)
	{
		
		fprintf(fpar,"%6.3f\t%d\t%6.3f\n",(double)(time_par_degrees[i]-*start_global_ticks)/(freq*1000000),par_degrees[i],reconf_par_degrees[i]);	
	}
	fclose(fpar);
		printf("Controllor finito\n");

	return NULL;
}
