/**
	Emitter for Functional Partitioning Implementation
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <ff/buffer.hpp>
#include <iostream>
#include <unordered_map>
#include <vector>
#include <list>
#include <sched.h>
#include "../Includes/cycle.h"
#include "../Includes/iaperf.h"
#include "../Includes/general.h"
#include "../Includes/farm_cb.h"
#include "../Includes/cbwindow.hpp"
#include "../Includes/farm_cb_reconf.h"
#include <sys/ioctl.h>
 #include <linux/sockios.h>

#if defined(USE_ZMQ)
#include <zmq.h>
#endif
using namespace ff;
using namespace std;


/**
	Determines the id of the Worker to which send the task
	according to a round robin allocation strategy of the logical streams
	to the Workers
*/
//In the scheduling table we will have for each class the current list of worker to which send
//the tuples of that class
char next_schedulingRR=0;
list<char> * scheduling_table; 
int64_t *window_ids; //unique id of the windows (they will start from zero)
/*inline char schedulingRR (task_t *t, int num_workers, int window_size, int window_slide) 
{
	//The scheduling table is a simple array with numb_classes positions
	//if the i-th element is zero then for that logical stream we don't have 
	//an assigned Worker. Otherwise it contains the ID_Worker+1 to which send
	//the tasks of that class
	if(scheduling_table[t->type]==0)
	{
		//DEBUG(printf("***New logical stream, assigned to Worker: %d\n",next_schedulingRR));
		scheduling_table[t->type]=next_schedulingRR	+1;
		next_schedulingRR=(next_schedulingRR+1)%num_workers;
	}
	return scheduling_table[t->type]-1;

}*/

void *emitter(void *args)
{

	struct timeval tmp_t;
	long start_t=0, end_t=0;
	char to_send_to;
	task_t tmp;
	int ret;
	//Init
	emitter_data_t *data = (emitter_data_t *) args;
	pthread_barrier_t *barrier = data->barrier;
	int num_workers=data->num_workers;
	int port=data->port;
	int window_size=data->window_size;
	int window_slide=data->window_slide;
	int64_t msg=0;
	SWSR_Ptr_Buffer **outqueue=data->outqueue;
	ticks *start_global_ticks=data->start_global_ticks;
	int num_classes=data->num_classes;
	long int freq=data->freq;
	//this is the number of live windows, i.e. the lenght of the lists in the scheduling table
	int num_live_windows=window_size/window_slide; 
	task_t eos_t; //task for terminating workers
	eos_t.type=-1;
	scheduling_table=new list<char>[num_classes];
	window_ids=(int64_t *)malloc(num_classes*sizeof(int64_t));
	memset(window_ids,-1,num_classes*sizeof(int64_t));

	printf("FARM EMITTER: ROUND ROBIN WINDOWING\n");
	
	//frequency counter for the various classes
	int64_t *classes_freq=(int64_t*)calloc(num_classes,sizeof(int64_t));
	#if defined(MONITORING)
		
	#endif

	#if defined(TASK_BUFF)
		printf(ANSI_COLOR_RED "Pay attention to this version with task_buff!\n" ANSI_COLOR_RESET);
		//create the buffer for tasks
		task_t * task_buff;
		//We do not know how large should be this buffer in order to guarantee correctness
		int buff_size=num_workers*(QUEUE_SIZE+2)*num_workers;
		posix_memalign((void **)&task_buff,CACHE_LINE_SIZE, buff_size*sizeof(task_t));
		int bi=0; //index pointer for the task buffer
	#else
		#if !defined(DONT_USE_FFALLOC)
			ff_allocator *ffalloc=data->ffalloc;
		#endif
	#endif
	task_t *tb;

	#if defined(MONITORING)
		//define the data structures for monitoring
		//queue towards the controller
		SWSR_Ptr_Buffer *cn_outqueue=data->cn_outqueue;		
		volatile ticks monitoring_timer;
		volatile ticks monitoring_step=MONITORING_STEP*1000*freq; //after how many cycles we have to monitor
		volatile ticks sub_monitoring_step=monitoring_step/10;
		//reconfiguration message
		reconf_data_em_t *reconf_data;
		emitter_monitoring_t *monitoring;
		char mon_ptr=0; //identifies the next monitoring structure in which write data
		posix_memalign((void **)&monitoring,CACHE_LINE_SIZE,NMONITORING*sizeof(emitter_monitoring_t));
		for(int i=0;i<NMONITORING;i++)	
		{
			//init the internals
			monitoring[i].mon_id=i;
			monitoring[i].tag=MON_TAG;
			monitoring[i].elements=0;
			monitoring[i].stop=false;
			monitoring[i].scheduling_table=NULL;
			posix_memalign((void **)&(monitoring[i].elements_per_class),CACHE_LINE_SIZE,num_classes*sizeof(int));
			memset(monitoring[i].elements_per_class,0,num_classes*sizeof(int));
		}
		reconf_repository_t *repo=(reconf_repository_t *)data->repo;

		#if defined(ADAPTIVITY)
			//queue from the controller for incoming reconfiguration commands
			SWSR_Ptr_Buffer *cn_inqueue=data->cn_inqueue;
			//this particular monitoring data is used to notify that the reconfiguration is finished
			emitter_monitoring_t reconf_finished;
			reconf_finished.tag=RECONF_FINISHED;
		#endif
	#endif
	
	//queues warmup (probably useless)
	for(int i=0;i<num_workers;i++)
		send((void *)1,outqueue[i]);
	DEBUG(printf("Emitter ready, running on: %d\n",sched_getcpu()))
	//synchronization barrier
	pthread_barrier_wait(barrier);
	

	//accept connection from generator
	#if defined (USE_ZMQ)
		void *context = zmq_ctx_new ();
		void *socket=receive_connection(context,port);
	#else
		int socket;
		socket=*(int *)receive_connection(1,port);		
	#endif

	/**
		The first receive is for synchronizing the global start time
	*/
	if((ret=socket_receive(socket, &tmp, sizeof(task_t)))!=sizeof(task_t))
	{
		fprintf(stderr,"The program is a bottleneck\n");
		exit(BOTTLENECK_ERR);
	}
	
	*start_global_ticks=tmp.timestamp; //ticks at generator side; this will be a sort of synchronized wall clock time
	gettimeofday(&tmp_t,NULL);

	#if defined(MONITORING)
		//start monitoring
        monitoring_timer=*start_global_ticks+monitoring_step; //start_global_ticks act as a shared wall clock time, we need this concept for synchronizing monitoring
        ticks sub_monitoring_timer=monitoring_timer;
        //per monitorare il socket
        int prev_numb_tuples=0;
        int max_diff=0;
        int max_diff2=0;
        int grow_rate=0;
        int *numb_tuples=(int *)calloc(20,sizeof(int)); //potevamo alle volte uscire fuori da questo array, nonostante il sotto intervallo sia 1/10. Da sistemare meglio
        int sum_numb_tuples=0;
        int sub_mon_idx=0;
    	volatile ticks start_ticks=getticks();
    #endif


	/**
		Start receiving real elements
	*/

	#if defined(TASK_BUFF)
		tb=&task_buff[bi];
	#else
		#if defined(DONT_USE_FFALLOC)
			posix_memalign((void **)&tb,CACHE_LINE_SIZE,sizeof(task_t));
		#else
			tb=(task_t *)ffalloc->malloc(sizeof(task_t)); //posix memalign does not work for ff allocator (bugged)
		#endif
	#endif
	if((ret=socket_receive(socket, tb, sizeof(task_t)))!=sizeof(task_t))
	{
		fprintf(stderr,"The program is a bottleneck\n");
		exit(BOTTLENECK_ERR);
	}
	//NOTA: usiamo la seconda soluzione per il tempo di partenza: viene considerato il tempo di arrivo della prima tupla
	// DA COMMENTARE SE SI VOGLIONO I TEMPI "SINCRONIZZATI" TRA EMETTITORE E DISTRIBUTORE 	

	//*start_global_ticks=getticks();
	asm volatile("":::"memory");
    start_t=tmp_t.tv_sec*1000+(tmp_t.tv_usec)/1000;

   	int mon_step=0; //tmp

   	//start receiving and distribute task
	while(tb->type!=-1)
	{
		//printf("Ricevuto task con id: %Ld\n",tb->type);
		msg++;
		
		#if defined (MONITORING)
			monitoring[mon_ptr].elements++;
			monitoring[mon_ptr].elements_per_class[tb->type]++;
		#endif
		tb->internal_id=classes_freq[tb->type]++;
		tb->punctuation=NO;
		/**
			SCHEDULING
			determine to which workers send the current tuple
			this will be done in a Round Robin fashion (each new window is allocated to a different worker)
		*/
		if(tb->internal_id%window_slide==0)//the tuple starts a new window
		{
			//remove the oldest worker to which the tuples of this class were sent and add a new one
			//NOTICE: in this case we can set the real worker id (not id+1 as in the other cases)
			if(scheduling_table[tb->type].size()==num_live_windows) //there are the right number of live windows
				scheduling_table[tb->type].pop_front(); //otherwise we have to keep it
			scheduling_table[tb->type].push_back(next_schedulingRR);
			next_schedulingRR=(next_schedulingRR+1)%num_workers;
			window_ids[tb->type]++; //increment it
		}
		list<char>::iterator it=scheduling_table[tb->type].begin(); 
		//we have to create a certain number of copies of the original task
		task_t* copy;
		for(int i=0;i<scheduling_table[tb->type].size()-1;i++)
		{
			//crea la copia
			to_send_to=*it;
			// printf("Class: %d to_send_to: %d int_id: %Ld (Number of current window: %d)\n",tb->type,*it,tb->internal_id,scheduling_table[tb->type].size());
			//take the pointer to a task
			#if defined(TASK_BUFF)
				//advance buffer index
				bi=(bi+1)%(buff_size);
				copy=&task_buff[bi];
			#else
				#if defined(DONT_USE_FFALLOC)
					posix_memalign((void **)&copy,CACHE_LINE_SIZE,sizeof(task_t));
				#else
					copy=(task_t *)ffalloc->malloc(sizeof(task_t));
			#endif
			#endif
			//copy it
			memcpy(copy,tb,sizeof(task_t));
			//set the CORRECT window_id
			copy->window_id=window_ids[tb->type]-scheduling_table[tb->type].size()+i+1;
			//send to the worker
			if(!send(copy,outqueue[to_send_to]))
			{
				fprintf(stderr, "Worker %d is a bottleneck\n",to_send_to);
				exit(BOTTLENECK_ERR);
			}
			it++;
		}
		// printf("-Class: %d to_send_to: %d\n",tb->type,*it);
		to_send_to=*it;
		//set the window_id (of the last window)
		tb->window_id=window_ids[tb->type];
		//send to the worker
		if(!send(tb,outqueue[to_send_to]))
		{
			fprintf(stderr, "Worker %d is a bottleneck\n",to_send_to);
			exit(BOTTLENECK_ERR);
		}
		//take the pointer to the next task that will be received
		#if defined(TASK_BUFF)
		//advance buffer index
			bi=(bi+1)%(buff_size);
			tb=&task_buff[bi];
		#else
			#if defined(DONT_USE_FFALLOC)
				posix_memalign((void **)&tb,CACHE_LINE_SIZE,sizeof(task_t));
			#else
				tb=(task_t *)ffalloc->malloc(sizeof(task_t));
			#endif
		#endif

		#if defined(MONITORING)
			#if defined(ADAPTIVITY) //mettere prima o dopo il monitoraggio?
				NOT IMPLEMENTED
			#endif
			//check if we can send monitoring data to controller
			//if(getticks()-monitoring_timer>monitoring_step)
			volatile ticks act_time=getticks();
			if(act_time>sub_monitoring_timer)
			{
				//per ora approssimato...vediamo quante tuple ci sono nel socket
				int value;
	            ioctl(socket, SIOCINQ, &value);
	            int diff=value/sizeof(task_t)-prev_numb_tuples; //se abbiamo un solo gradino questo non funziona
	            if(diff>max_diff)
	            {
	            	max_diff2=max_diff;
	            	max_diff=diff;
	            	// printf("Max diff: %d\n",max_diff);
	            }
	            //devi vedere se alla fine e' significativo
	            prev_numb_tuples=value/sizeof(task_t);
	            numb_tuples[sub_mon_idx++]=value/sizeof(task_t);
	            sum_numb_tuples+=value/sizeof(task_t);
	            sub_monitoring_timer+=sub_monitoring_step;
	        }
			if(act_time>monitoring_timer)
			{ 	
				
				asm volatile("":::"memory");
				//consider also the elements in the buffer:
				int value;
            	ioctl(socket, SIOCINQ, &value);
				
           		
            	monitoring[mon_ptr].buffer_elements=value/sizeof(task_t);
            	//compute slope in socket buffer
            	double sx=0,sxy=0;
            	double xavg=sub_mon_idx*(sub_mon_idx-1)/2; //we start from zero
            	for(int i=0;i<sub_mon_idx;i++)
            	{
            		sx+=(i-xavg)*(i-xavg);
            		sxy+=(i-xavg)*(numb_tuples[i]-sum_numb_tuples/sub_mon_idx);
            		//printf("%d ",numb_tuples[i]);
            	}
            	//printf("\n");
            	double slope=sxy/sx;
            	printf(ANSI_COLOR_MAGENTA "-----------Emitter: buffered tasks: %d, grow rate: %g slope:%f\n" ANSI_COLOR_RESET,value/sizeof(task_t),max_diff*9.5,slope);
            	if(slope>0)
            		monitoring[mon_ptr].grow_rate_per_step=max_diff*9.5;
            	else
            		monitoring[mon_ptr].grow_rate_per_step=0;
		
				monitoring[mon_ptr].scheduling_table=scheduling_table; //NOTA: per il momento passiamo il puntatore (o fare la copia? dipende dal controllore che fa)
				
				//printf("Emitter sending...%d\n",cn_outqueue->length());
				bsend(&monitoring[mon_ptr],cn_outqueue); //pay attention, if queue is full it is blocked

				//printf("Emitter sent step: %d\n",mon_step++);
				monitoring_timer=getticks();
				mon_ptr=(mon_ptr+1)%NMONITORING;
				//reset the monitoring data
				monitoring[mon_ptr].elements=0;
				monitoring[mon_ptr].buffer_elements=0;
				grow_rate=0;
				max_diff=0;
				max_diff2=0;
				sum_numb_tuples=0;
				sub_mon_idx=0;
				memset(monitoring[mon_ptr].elements_per_class,0,num_classes*sizeof(int));
				monitoring_timer+=monitoring_step;

			}
			


			
		#endif
		if((ret=socket_receive(socket, tb, sizeof(task_t)))!=sizeof(task_t))
		{
			fprintf(stderr,"The program is a bottleneck?\n");
			exit(BOTTLENECK_ERR);
		}
		
	}
	printf("Emitter, send termination----------\n");

	//EOS: send terminating task to all the workers
	for(int i=0;i<num_workers;i++)
	{
		//printf("Emitter, send EOS to:%d\n",i);
		send(tb,outqueue[i]);
	}
	
	#if defined(MONITORING)
		//send termination to controller
		monitoring[mon_ptr].stop=true;
		bsend(&monitoring[mon_ptr],cn_outqueue);
	#endif
	//compute statistics
	gettimeofday(&tmp_t,NULL);
	end_t=tmp_t.tv_sec*1000+(tmp_t.tv_usec)/1000-start_t;
	//pthread_barrier_wait(barrier);
	printf("***EMITTER***\n");
	printf("Tempo passato (msec): %Ld, msg per second: %6.3f\n",end_t,((double)msg)/((double)end_t/1000));
		
	#if defined (USE_ZMQ)
	close(socket);
	zmq_ctx_destroy (context);
	#endif


	return NULL;
}
