/**
	Worker for FARM implementation
	The main difference (wrt the Functional Partioning) is that a Worker
	can have multiple windows of the same class active at the same time.

	For this reason the map is this time ORDERED and it is index with a pair <class_id,window_id>.
	(this was the fastes way to achieve it)
	NOTICE: for this reason, currently the ADAPTIVITY version is not working
	(Probably we will have a completely different adaptation strategy)
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <ff/buffer.hpp>
#include <iostream>
#include <map>
#include <vector>
#include <sched.h>
#include <set>
#include "../Includes/cycle.h"
#include "../Includes/iaperf.h"
#include "../Includes/general.h"
#include "../Includes/farm_cb.h"
#include "../Includes/cbwindow_farm.hpp"
#include "../Includes/farm_cb_reconf.h"
#include <ff/allocator.hpp>	
#if defined(USE_ZMQ)
#include <zmq.h>
#endif
using namespace ff;
using namespace std;

void * worker(void *args) {

	task_t *tmp;
	CBWindowFarm *window;
	HReturn_type *res_buff; //result buffer in order to reuse memory
	int buff_size;
	int bi=0;
	//Init
	worker_data_t *data = (worker_data_t *) args;
	pthread_barrier_t *barrier = data->barrier;
	int id = data->workerId;
	SWSR_Ptr_Buffer *inqueue=data->inqueue;
	SWSR_Ptr_Buffer *outqueue=data->outqueue;
	int window_slide=data->window_slide;
	int window_size=data->window_size;
	long int freq=data->freq;
	int num_classes=data->num_classes;
	ticks *start_global_ticks=data->start_global_ticks;
	#if !defined(DONT_USE_FFALLOC) && !defined(TASK_BUFF)
			ff_allocator *ffalloc=data->ffalloc;
	#endif
	#if defined(MONITORING)
		SWSR_Ptr_Buffer *cn_outqueue=data->cn_outqueue;	
	#endif

	map<pair<int,int64_t>,CBWindowFarm*> map; //map that contains the various windows
	//memory allocation
	buff_size=QUEUE_SIZE+10;//is it correct?
	posix_memalign((void **)&res_buff,CACHE_LINE_SIZE, buff_size*sizeof(HReturn_type));
	
	//int64_t triggers[100000];
	//int64_t computed=0;
	#if defined(MONITORING)
		//define the data structures for monitoring
		
		ticks start;

		volatile ticks monitoring_timer;
		volatile ticks monitoring_step=MONITORING_STEP*1000*freq; //after how many cycles we have to monitor
		worker_monitoring_t *monitoring;
		char mon_ptr=0; //the current monitoring structure in which write data
		posix_memalign((void **)&monitoring, CACHE_LINE_SIZE,NMONITORING*sizeof(worker_monitoring_t));
		for(int i=0;i<NMONITORING;i++)
		{
			//init the internals
			posix_memalign((void**)&(monitoring[i].tcalc_per_class),CACHE_LINE_SIZE,num_classes*sizeof(double));
			memset(monitoring[i].tcalc_per_class,0,num_classes*sizeof(double));
			posix_memalign((void **)&(monitoring[i].elements_per_class),CACHE_LINE_SIZE,num_classes*sizeof(int));
			memset(monitoring[i].elements_per_class,0,num_classes*sizeof(int));
			posix_memalign((void **)&(monitoring[i].computations_per_class),CACHE_LINE_SIZE,num_classes*sizeof(int));
			memset(monitoring[i].computations_per_class,0,num_classes*sizeof(int));
			monitoring[i].elements_rcvd=0;
			monitoring[i].computations=0;

		}
		//adaptivity data structures
		reconf_repository_t *repo=(reconf_repository_t *)data->repo;
		CBWindowFarm** moving_windows=(repo->moving_windows);
		bool reconfiguration_phase_in=false; //indicates if the worker is currently in the reconfiguration phase (moving_in)
		bool reconfiguration_phase_out=false; //indicates if the worker is currently in the reconfiguration phase (moving_out)
		bool reconfiguration_phase=false; //indicate the whole process
		set<int> classes_moving_in; //it will contains all the id of the classes that are currently moving in
		vector<task_t *> task_moving_in(20000); //it will contain all the task belonging to moving in classes (not yet arrived to the worker) (to see why i choose vector look at the explnation below, when i use it)
		task_moving_in.clear();
	#endif



	if(barrier)	//it is a thread spawned at the beginning
	{
		//queues warmup
		receive((void **)&tmp,inqueue);
		send((void *)1,outqueue);
		//synchronization barrier
		// DEBUG(printf("Worker %d ready, running on %d mon_step: %Ld\n",id,sched_getcpu(),monitoring_step));
		pthread_barrier_wait(barrier);
	}

	while(*start_global_ticks==0)
		REPEAT_25(asm volatile("PAUSE" ::: "memory");) //wait for the correct time
	

	#if defined(MONITORING)
	if(barrier)
		monitoring_timer=*start_global_ticks+monitoring_step; //we skip the first step due to starting phase
	else //it a newly thread
	{
		monitoring_timer=*start_global_ticks+((getticks()-*start_global_ticks)/monitoring_step+1)*monitoring_step;
		 //printf("%d Prima del monitoring ho %Ld\n",id,monitoring_timer-getticks());
	}
	#endif
	//buffer of results
	
	/**
		Start receiving elements
	*/

	receive((void **)&tmp,inqueue);

	int ret;
	int mon_step=0;
	int count=0;
	ticks no_trigger=0;
	int no_trigger_tuple=0;
	while(tmp->type!=-1)
	{
		/*
			copy the task into the window and if necessary, perform computation
		*/
		//printf("Ricevuto elemento %Ld con tipo: %d\n",tmp->id,tmp->type);
		#if defined(MONITORING)
			asm volatile("":::"memory");
			start=getticks();
			monitoring[mon_ptr].elements_per_class[tmp->type]++;
			monitoring[mon_ptr].elements_rcvd++;
			asm volatile("":::"memory");
		#endif
		#if defined(ADAPTIVITY)
		NOT IMPLEMENTED
			
		#endif
				
				//check if there is already an entry for that window and that class
				window=map[make_pair(tmp->type,tmp->window_id)];
				
				if(window==NULL) //it's a new logical stream/window that has arrived
				{
					//DEBUG(printf("--------------------[%d] Ricevuto un elemento di un nuovo stream logico: %d\n",id,tmp->type));
					//qui possiamo ricavarci una grana: ad esempio facciamo 7+class_id%4 *2

					//window=new CBWindowFarm(window_size, window_slide,(tmp->type%4)*2+7);
					window=new CBWindowFarm(window_size, window_slide);
					map[make_pair(tmp->type,tmp->window_id)]=window;
				}
				//insert the element in window
				//res_buff[bi].ts=getticks();
				
				ret=window->insert(tmp,&res_buff[bi]);
				 // printf("[%d] Arrivato elemento: %Ld per classe %d (window: %Ld)\n",id,tmp->internal_id,tmp->type,tmp->window_id);
				if (ret!=0) //we have a new result (NOTE: actually, it will be better to handle this situation in a different way)
				{
					 // printf("[%d] Calcolat! Classe: %d, id %Ld window_id %Ld\n",id,tmp->type,tmp->internal_id,tmp->window_id);
					// if(tmp->type==68)
					// printf("Computato per task: %Ld, classe: %d\n",tmp->id, tmp->type);
					//set the timestamp to the timestamp of the one of the task that has triggered the computation
					res_buff[bi].timestamp=tmp->timestamp;

					//the same for the id
					res_buff[bi].id=tmp->internal_id; //for the moment just for ordering
					res_buff[bi].type=tmp->type;
					res_buff[bi].isEOS=false;
					res_buff[bi].wid=(char)id;

					//triggers[computed]=tmp->id;
					//computed++;
					/*
					if((res_buff[bi].id+1)%500!=0)
					{
						printf("********orker %d error, computed on %Ld, class:%d\n",id, tmp->internal_id,tmp->type );
						exit(-1);
					}*/
					//to be sent to the collector
					//send result to the collector
					send(&res_buff[bi],outqueue);
					//advance the buffer index
					bi=(bi+1)%buff_size;
					#if defined(MONITORING)
					asm volatile("":::"memory");
					monitoring[mon_ptr].tcalc_per_class[tmp->type]+=(double)(getticks()-start)/freq; //usec
					monitoring[mon_ptr].computations_per_class[tmp->type]++;
					monitoring[mon_ptr].computations++;
					asm volatile("":::"memory");
					#endif

					//delete the window and the entry in the map
					delete window;
					map.erase(make_pair(tmp->type,tmp->window_id));
				}
				// else
				// {
				// 	no_trigger+=getticks()-start;
				// 	no_trigger_tuple++;
				// }
				
				
				#if defined(MONITORING)
					//check if we can send monitoring data to controller
					

					//if(getticks()-monitoring_timer>monitoring_step)
					if(getticks()>monitoring_timer)
					{ 	
						asm volatile("":::"memory");


						// printf("Monitoring data %d are valid. Elements received: %Ld\n",mon_ptr,monitoring[mon_ptr].elements);
						
						//blocking send towards the collector
						// printf("Worker %d sends %d step: %d\n",id,cn_outqueue->length(),mon_step++);
						bsend(&monitoring[mon_ptr],cn_outqueue);
						
						// printf("Worker %d sent\n",id);
						monitoring_timer=getticks();
						mon_ptr=(mon_ptr+1)%NMONITORING;
						//reset the monitoring data
						memset(monitoring[mon_ptr].elements_per_class,0,num_classes*sizeof(int));
						memset(monitoring[mon_ptr].tcalc_per_class,0,num_classes*sizeof(double));
						memset(monitoring[mon_ptr].computations_per_class,0,num_classes*sizeof(int));
						monitoring[mon_ptr].elements_rcvd=0;
						monitoring[mon_ptr].computations=0;
						monitoring_timer+=monitoring_step;
						// printf("Empty average tcalc: %Lf usec\n",(((long double)no_trigger)/no_trigger_tuple)/freq);
						// no_trigger=0;
						// no_trigger_tuple=0;
					}
				#endif

				#if !defined(TASK_BUFF)
					#if defined(DONT_USE_FFALLOC)
						free(tmp);
					#else
					ffalloc->free(tmp);
					#endif
				#endif


		#if defined(ADAPTIVITY)
			NOT IMPLEMENTED
		#endif 
		
		//receive the next element
		receive((void **)&tmp,inqueue);
	}
	#if defined(ADAPTIVITY)
	NOT IMPLEMENTED
	#endif
	#if defined(MONITORING) //BOHHHHHHHHHHHHHHHHHHHHHHHHHHHHH??????????????????????????
	//send the last monitoring, in order to avoid stall on the controller
	//bsend(&monitoring[mon_ptr],cn_outqueue);
	bsend((void *)(1),cn_outqueue);
	#endif
	//send EOS to collector
	
	res_buff[bi].isEOS=true;
	res_buff[bi].res_buff=res_buff;
	
	send(&res_buff[bi],outqueue);
	//printf("Worker %d ricevuto EOS and sended to collector, rcvd buff: %p\n",id,res_buff[bi].res_buff);
	//MEMORY FREEING
	//destroy queues from emitter and toward collector
	delete(inqueue);
	//delete(cn_outqueue); this has to be done by controller
	//there is no need to free tmp(it can be statically allocated on the emitter )
	//delete structure for monitoring, except the last one sent?????
	for(int i=0;i<NMONITORING;i++)
	{
		//init the internals
		free(monitoring[i].tcalc_per_class);
		monitoring[i].tcalc_per_class=NULL;
		free(monitoring[i].elements_per_class);
		monitoring[i].elements_per_class=NULL;
		free(monitoring[i].computations_per_class);
		monitoring[i].computations_per_class=NULL;
		
	}
	free(monitoring);
	monitoring=NULL;
	//since it is still used, res_buff will be destroyed by the collector
	
	//printf("Worker %d finito\n",id);

	//pthread_barrier_wait(barrier);
	//sleep(id+1);

	/*for(int i=0;i<240;i++)
		if(counts[i]>0)
		{
			comp_time[i]=((double)computation_time[i])/counts[i]/FREQ;
		}
*/
	/*for ( unsigned i = 0; i < map.bucket_count(); ++i) {
   		for ( auto local_it = map.begin(i); local_it!= map.end(i); ++local_it ){
    		//printf("Worker %d terminated. Class: %d Average computation time empy (usec): \n",id,(local_it->first));
    		comp_time[local_it->first]=(local_it->second->getAvgComputationTime())/freq;
    	}
  	}*/
   /* char fname[20];
    sprintf(fname,"worker_%d.dat",id);
    FILE *fout=fopen(fname,"w");
    for(int i=0;i<computed;i++)
    	fprintf(fout,"%Ld\n",triggers[i]);
    fclose(fout);*/
    return NULL;
}
