/**
	Emitter for Functional Partitioning Implementation
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <ff/buffer.hpp>
#include <iostream>
#include <unordered_map>
#include <vector>
#include <sched.h>
#include "../Includes/cycle.h"
#include "../Includes/iaperf.h"
#include "../Includes/general.h"
#include "../Includes/functional_partitioning_cb.h"
#include "../Includes/cbwindow.hpp"
#include "../Includes/fp_cb_reconf.h"
#include <sys/ioctl.h>
 #include <linux/sockios.h>

#if defined(USE_ZMQ)
#include <zmq.h>
#endif
using namespace ff;
using namespace std;

int readDistributions(char *file, int **times, double ***distributions)
{
	FILE *fdistr=fopen(file,"r");
	srand(10);
	int nclass;
	int ndistributions;
	int i,j;
	fscanf(fdistr,"%d %d\n",&nclass,&ndistributions);
	//allocate the space
	int *t=(int *)malloc(sizeof(int)*nclass);
	double **d=(double **)malloc(sizeof(double *)*ndistributions);
	for(i=0;i<ndistributions;i++)
		d[i]=(double *)malloc(sizeof(double)*nclass);

	//start reading
	for(i=0;i<ndistributions;i++)
	{
		fscanf(fdistr, "%d",&t[i]);
		for(j=0;j<nclass;j++)
		{
			fscanf(fdistr,"%lf",&d[i][j]);
			// printf("%3.5f ",d[i][j]);
		}
		
	}
	// printf("\n");
	*times=t;
	*distributions=d;
	return ndistributions;

}

/**
	Determines the id of the Worker to which send the task
	according to a round robin allocation strategy of the logical streams
	to the Workers
*/
	//next assigned worker scheduling RR
char next_schedulingRR=0;
char * scheduling_table;
inline char schedulingRR (task_t *t, int num_workers) 
{
	//The scheduling table is a simple array with numb_classes positions
	//if the i-th element is zero then for that logical stream we don't have 
	//an assigned Worker. Otherwise it contains the ID_Worker+1 to which send
	//the tasks of that class
	if(scheduling_table[t->type]==0)
	{
		//DEBUG(printf("***New logical stream, assigned to Worker: %d\n",next_schedulingRR));
		scheduling_table[t->type]=next_schedulingRR	+1;
		next_schedulingRR=(next_schedulingRR+1)%num_workers;
	}
	return scheduling_table[t->type]-1;

}

void *emitter(void *args)
{

	struct timeval tmp_t;
	long start_t=0, end_t=0;
	
	task_t tmp;
	char to_send_to;
	int ret;
	//Init
	emitter_data_t *data = (emitter_data_t *) args;
	pthread_barrier_t *barrier = data->barrier;
	int num_workers=data->num_workers;
	int port=data->port;
	int64_t msg=0;
	SWSR_Ptr_Buffer **outqueue=data->outqueue;
	ticks *start_global_ticks=data->start_global_ticks;
	int num_classes=data->num_classes;
	long int freq=data->freq;
	task_t eos_t; //task for terminating workers
	eos_t.type=-1;
	scheduling_table=(char *)calloc(num_classes,sizeof(char));

	//static scheduling table (actually it should be only at the beginning....)
	int *times;
	double **distributions;
	int ndistributions=readDistributions("distr1000_zipf_non_risolv",&times, &distributions);
	printf("---> ATTENTION: Emitter: static scheduling <---\n");
	char w=0;
	//tcalc variabile, scheduling equi
	//double class_tcalcs[]={2.65,3.38,4.12,4.84};
	//double class_tcalcs[]={3.75,3.75,3.75,3.75};
	double class_tcalcs[]={7.7,7.7,7.7,7.7};
	/*scheduling_table[0]=1;
	scheduling_table[1]=2;
	scheduling_table[2]=2;
	scheduling_table[3]=1;
	scheduling_table[4]=3;
	scheduling_table[5]=4;
	scheduling_table[6]=4;
	scheduling_table[7]=3;
	scheduling_table[8]=5;
	scheduling_table[9]=6;
	scheduling_table[10]=6;
	scheduling_table[11]=5;
	scheduling_table[12]=7;
	scheduling_table[13]=8;
	scheduling_table[14]=8;
	scheduling_table[15]=7;
	for(int i=0;i<num_workers;i++)
		for(int j=0;j<28;j++)
			scheduling_table[16+i*28+j]=i+1;*/
	
	/*
		ALLOCAZIONE ALLA BISACCIA
		UTILE NEL CASO ZIPF. TCALC UGUALI
		ATT: INDICARE CORRETTAMENTE IL FILE DELLA DISTRIBUZIONE
		ATTENZIONE ANCHE AL PARAMETRO RHO SE TROPPO ALTO TI TROVI 
		CON WORKER SENZA CLASSI


	*/
	/*
	double load_w[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double pi_w[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double rho=0.8;
	double tmp_load,tmp_pi;
	bool assigned[1000];
	for(int i=0;i<1000;i++)
		assigned[i]=false;
	//fatto molto alla cazzo di cane
	for(int i=0;i<1000;i++)
	{
		//dobbiamo caricare il tcalc medio sulla base delle classi assegnate e della loro probabilita'
		//li pesiamo e poi divediamo per la probabilita' fino ad allora accumulata
		
		//find max class not already assigned. They have all the same tcalc
		double max_p=0;
		int class_index=0;
		for(int j=0;j<1000;j++)
		{
			if(distributions[0][j]>max_p && !assigned[j])
			{
				//printf("new max: %6.3f\n",distributions[0][j]);
				class_index=j;
				max_p=distributions[0][j];
			}
		}
		//printf("max: %f\n",max_p);
		assigned[class_index]=true;
		
		//int class_index=i;

		//trova worker in grado da poter sostenere l'assegnazione della classe
		//ossia quello con load minore
		double min_w=load_w[0];
		int work_index=0;
		for(int j=0;j<num_workers;j++)
		{
			if(load_w[j]<min_w)
			{
				work_index=j;
				min_w=load_w[j];
			}
		}
		scheduling_table[class_index]=work_index+1;
		load_w[work_index]+=class_tcalcs[0]*distributions[0][class_index]; //they have all the same tcalc
		//printf("max: %f assegnata a: %d che ora ha load_w: %f\n",max_p,work_index,load_w[work_index]);

		/*
		OLD
		w=0;
		do{
			tmp_load=load_w[w]+class_tcalcs[class_index%4]*distributions[0][class_index];
			tmp_pi=pi_w[w]+distributions[0][class_index];
			//printf("%d %d tmp_load %f tmp pi %f\n",i,w,tmp_load,tmp_pi);
			w++;
		}while(w<num_workers && tmp_load/rho/tmp_pi>1000.0/(tmp_pi*1600));
		if(w>num_workers)
		{
			printf("Errore\n");
			exit(-1);
		}
		w--;
		load_w[w]+=class_tcalcs[class_index%4]*distributions[0][class_index];
		pi_w[w]+=distributions[0][class_index];
		//printf("wid:%d i:%d %f %f %f\n",w,i,load_w[w]/pi_w[w],distributions[0][class_index],1000.0/(pi_w[w]*1600));
		scheduling_table[class_index]=w+1;*/
/*	}

	/*double load_w=0.0;
	double max_load=135;
	for(int i=0;i<240;i+=4)
	{
		scheduling_table[i]=w+1;
		load_w+=class_tcalcs[0];
		if(load_w>max_load)
		{
			load_w=0.0;
			w++;
		}
	}
	for(int i=1;i<240;i+=4)
	{
		scheduling_table[i]=w+1;
		load_w+=class_tcalcs[1];
		if(load_w>max_load)
		{
			load_w=0.0;
			w++;
		}
	}
	for(int i=2;i<240;i+=4)
	{
		scheduling_table[i]=w+1;
		load_w+=class_tcalcs[2];
		if(load_w>max_load)
		{
			load_w=0.0;
			w++;
		}
	}
	for(int i=3;i<240;i+=4)
	{
		scheduling_table[i]=w+1;
		load_w+=class_tcalcs[3];
		if(load_w>max_load)
		{
			load_w=0.0;
			w++;
		}
	}
*/
	
	// ALLOCAZIONE RR
	for(int i=0;i<num_classes;i++)
	{
		scheduling_table[i]=w+1;
		w++;
		w%=num_workers;
	}
	/*scheduling_table[0]=1;
	scheduling_table[1]=1;
	scheduling_table[2]=1;
	scheduling_table[3]=1;
	scheduling_table[4]=1;
	scheduling_table[5]=1;
	scheduling_table[6]=1;
	scheduling_table[7]=1;
	for(int i=8;i<num_classes;i++)
	{
		scheduling_table[i]=w+2;
		w++;
		w%=(num_workers-1);
	}*/

	//distr 64
	/*scheduling_table[0]=1;
	scheduling_table[1]=1;
	scheduling_table[2]=1;
	scheduling_table[3]=1;
	for(int i=4;i<num_classes;i++)
	{
		scheduling_table[i]=w+2;
		w++;
		w%=(num_workers-1);
	}*/
	/*scheduling_table[0]=1;
	scheduling_table[1]=1;
	scheduling_table[2]=1;
	scheduling_table[3]=1;
	scheduling_table[4]=1;
	scheduling_table[5]=2;
	scheduling_table[6]=2;
	scheduling_table[7]=2;
	scheduling_table[8]=2;
	scheduling_table[9]=2;
	scheduling_table[10]=3;
	scheduling_table[11]=3;
	scheduling_table[12]=3;
	scheduling_table[13]=3;
	scheduling_table[14]=3;
	scheduling_table[15]=4;


	for(int i=16;i<num_classes;i++)
	{
		scheduling_table[i]=w+1;
		w++;
		w%=(num_workers);
	}
*/
		//distr_48
	/*scheduling_table[0]=1;
	scheduling_table[1]=1;
	scheduling_table[2]=1;
	scheduling_table[3]=1;
	scheduling_table[4]=1;

	for(int i=5;i<num_classes;i++)
	{
		scheduling_table[i]=w+2;
		w++;
		w%=(num_workers-1);
	}*/
	/*scheduling_table[0]=1;
	scheduling_table[1]=1;
	scheduling_table[2]=1;
	scheduling_table[3]=1;
	scheduling_table[4]=1;
	scheduling_table[5]=2;
	scheduling_table[6]=2;
	scheduling_table[7]=2;
	scheduling_table[8]=2;
	scheduling_table[9]=2;
	scheduling_table[10]=3;
	scheduling_table[11]=3;
	scheduling_table[12]=3;
	scheduling_table[13]=3;
	scheduling_table[14]=3;


	for(int i=15;i<num_classes;i++)
	{
		scheduling_table[i]=w+4;
		w++;
		w%=(num_workers-3);
	}*/
		//compute NC delay
	
	/*double nc_delay[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double probs[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double weighted_tcalc[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	for(int i=0;i<num_classes;i++)
	{
		printf("%d Assegnata a: %d, tcalc:%6.3f\n",i, scheduling_table[i]-1,class_tcalcs[i%4]);
	//	nc_delay[scheduling_table[i]-1]+=class_tcalcs[i%4];
	//	probs[scheduling_table[i]-1]+=distributions[0][i];
	//	weighted_tcalc[scheduling_table[i]-1]+=class_tcalcs[i%4]*distributions[0][i];
	}
	printf("NC delay - probs - wtcalc:\n");
	for(int i=0;i<num_workers;i++)
			printf("%d %6.3f %6.3f %6.3f rho: %f\n",i,nc_delay[i],probs[i],weighted_tcalc[i]/probs[i],(weighted_tcalc[i]/probs[i])/(1000/(1600*probs[i])));
	*/

	//number of classes assigned to each Worker
	int ncw[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	for(int i=0;i<num_classes;i++)
		ncw[scheduling_table[i]-1]++;
	printf("Class per worker\n");
	for(int i=0;i<num_workers;i++)
		printf("%d: %d\n",i,ncw[i]);

	//frequency counter for the various classes
	int64_t *classes_freq=(int64_t*)calloc(num_classes,sizeof(int64_t));
	#if defined(MONITORING)
		
	#endif

	#if defined(TASK_BUFF)
		printf(ANSI_COLOR_RED "Pay attention to this version with task_buff!\n" ANSI_COLOR_RESET);
		//create the buffer for tasks
		task_t * task_buff;
		//We do not know how large should be this buffer in order to guarantee correctness
		int buff_size=num_workers*(QUEUE_SIZE+2)*num_workers;
		posix_memalign((void **)&task_buff,CACHE_LINE_SIZE, buff_size*sizeof(task_t));
		int bi=0; //index pointer for the task buffer
	#else
		#if !defined(DONT_USE_FFALLOC)
			ff_allocator *ffalloc=data->ffalloc;
		#endif
	#endif
	task_t *tb;

	#if defined(MONITORING)
		//define the data structures for monitoring
		//queue towards the controller
		SWSR_Ptr_Buffer *cn_outqueue=data->cn_outqueue;		
		volatile ticks monitoring_timer;
		volatile ticks monitoring_step=MONITORING_STEP*1000*freq; //after how many cycles we have to monitor
		volatile ticks sub_monitoring_step=monitoring_step/10;
		//reconfiguration message
		reconf_data_em_t *reconf_data;
		emitter_monitoring_t *monitoring;
		char mon_ptr=0; //identifies the next monitoring structure in which write data
		posix_memalign((void **)&monitoring,CACHE_LINE_SIZE,NMONITORING*sizeof(emitter_monitoring_t));
		for(int i=0;i<NMONITORING;i++)	
		{
			//init the internals
			monitoring[i].mon_id=i;
			monitoring[i].tag=MON_TAG;
			monitoring[i].elements=0;
			monitoring[i].stop=false;
			monitoring[i].scheduling_table=NULL;
			posix_memalign((void **)&(monitoring[i].elements_per_class),CACHE_LINE_SIZE,num_classes*sizeof(int));
			memset(monitoring[i].elements_per_class,0,num_classes*sizeof(int));
		}
		reconf_repository_t *repo=(reconf_repository_t *)data->repo;

		#if defined(ADAPTIVITY)
			//queue from the controller for incoming reconfiguration commands
			SWSR_Ptr_Buffer *cn_inqueue=data->cn_inqueue;
			//this particular monitoring data is used to notify that the reconfiguration is finished
			emitter_monitoring_t reconf_finished;
			reconf_finished.tag=RECONF_FINISHED;
		#endif
	#endif
	
	//queues warmup (probably useless)
	for(int i=0;i<num_workers;i++)
		send((void *)1,outqueue[i]);
	DEBUG(printf("Emitter ready, running on: %d\n",sched_getcpu()))
	//synchronization barrier
	pthread_barrier_wait(barrier);
	

	//accept connection from generator
	#if defined (USE_ZMQ)
		void *context = zmq_ctx_new ();
		void *socket=receive_connection(context,port);
	#else
		int socket;
		socket=*(int *)receive_connection(1,port);		
	#endif

	/**
		The first receive is for synchronizing the global start time
	*/
	if((ret=socket_receive(socket, &tmp, sizeof(task_t)))!=sizeof(task_t))
	{
		fprintf(stderr,"The program is a bottleneck\n");
		exit(BOTTLENECK_ERR);
	}
	
	*start_global_ticks=tmp.timestamp; //ticks at generator side; this will be a sort of synchronized wall clock time
	gettimeofday(&tmp_t,NULL);

	#if defined(MONITORING)
		//start monitoring
        monitoring_timer=*start_global_ticks+monitoring_step; //start_global_ticks act as a shared wall clock time, we need this concept for synchronizing monitoring
        ticks sub_monitoring_timer=monitoring_timer;
        //per monitorare il socket
        int prev_numb_tuples=0;
        int max_diff=0;
        int max_diff2=0;
        int grow_rate=0;
        int *numb_tuples=(int *)calloc(20,sizeof(int)); //potevamo alle volte uscire fuori da questo array, nonostante il sotto intervallo sia 1/10. Da sistemare meglio
        int sum_numb_tuples=0;
        int sub_mon_idx=0;
    	volatile ticks start_ticks=getticks();
    #endif


	/**
		Start receiving real elements
	*/

	#if defined(TASK_BUFF)
		tb=&task_buff[bi];
	#else
		#if defined(DONT_USE_FFALLOC)
			posix_memalign((void **)&tb,CACHE_LINE_SIZE,sizeof(task_t));
		#else
			tb=(task_t *)ffalloc->malloc(sizeof(task_t)); //posix memalign does not work for ff allocator (bugged)
		#endif
	#endif
	if((ret=socket_receive(socket, tb, sizeof(task_t)))!=sizeof(task_t))
	{
		fprintf(stderr,"The program is a bottleneck\n");
		exit(BOTTLENECK_ERR);
	}
	//NOTA: usiamo la seconda soluzione per il tempo di partenza: viene considerato il tempo di arrivo della prima tupla
	// DA COMMENTARE SE SI VOGLIONO I TEMPI "SINCRONIZZATI" TRA EMETTITORE E DISTRIBUTORE 	

	//*start_global_ticks=getticks();
	asm volatile("":::"memory");
    start_t=tmp_t.tv_sec*1000+(tmp_t.tv_usec)/1000;

   	int mon_step=0; //tmp

   	//start receiving and distribute task
	while(tb->type!=-1)
	{
		//printf("Ricevuto task con id: %Ld\n",tb->type);
		msg++;
		
		#if defined (MONITORING)
			monitoring[mon_ptr].elements++;
			monitoring[mon_ptr].elements_per_class[tb->type]++;
		#endif
		tb->internal_id=classes_freq[tb->type]++;
		tb->punctuation=NO;
		to_send_to=schedulingRR(tb,num_workers); //e qui
		
		if(!send(tb,outqueue[to_send_to]))
		{
			fprintf(stderr, "Worker %d is a bottleneck\n",schedulingRR(tb,num_workers));
			exit(BOTTLENECK_ERR);
		}
		
		//take the pointer to the next task that will be received
		#if defined(TASK_BUFF)
		//advance buffer index
			bi=(bi+1)%(buff_size);
			tb=&task_buff[bi];
		#else
			#if defined(DONT_USE_FFALLOC)
				posix_memalign((void **)&tb,CACHE_LINE_SIZE,sizeof(task_t));
			#else
				tb=(task_t *)ffalloc->malloc(sizeof(task_t));
			#endif
		#endif

		#if defined(MONITORING)
			#if defined(ADAPTIVITY) //mettere prima o dopo il monitoraggio?

			//check if there are reconfiguration messages from the controller
			if(cn_inqueue->pop((void **)(&reconf_data)))
			{
				//printf("Emitter: ricevuto messaggio riconfigurazione\n");
				//reset repository
				for(int i=0;i<num_classes;i++)
					repo->moving_windows[i]=NULL;

				//check if there are newly spawned threads
				if(reconf_data->par_degree_changes>0)
				{
					printf(ANSI_COLOR_YELLOW "[EMITTER] ricevuto messaggio, incremento par degree di %d worker\n" ANSI_COLOR_RESET,reconf_data->par_degree_changes);
					//take the new queues
					for(int i=0;i<reconf_data->par_degree_changes;i++)
						outqueue[num_workers+i]=reconf_data->wqueues[i];
					//increments the par degree
					num_workers+=reconf_data->par_degree_changes;
				}
				//par degree decrease
				if(reconf_data->par_degree_changes<0)
					printf(ANSI_COLOR_YELLOW "[EMITTER] ricevuto messaggio, decremento par degree di %d worker\n" ANSI_COLOR_RESET,reconf_data->par_degree_changes);
				
				//Find and handle differences between current and previous scheduling table
				for(int i=0;i<num_classes;i++)
				{
					if(scheduling_table[i]!=reconf_data->scheduling_table[i])
					{
						// printf("Da riconfigurare classe %d %d->%d\n",i,scheduling_table[i]-1,reconf_data->scheduling_table[i]-1);
						//send the proper signal to the worker that until now has mantained the class
						//NOTA: QUESTI ANDREBBERO POI FREE
						task_t *signalt=(task_t*)malloc(sizeof(task_t));
						signalt->type=i; //signal the class that has to be moved
						signalt->punctuation=MOVING_OUT;
						if(!send(signalt,outqueue[scheduling_table[i]-1]))
						{
							fprintf(stderr, ANSI_COLOR_RED "[EMITTER] Worker %d is a bottleneck\n" ANSI_COLOR_RESET,scheduling_table[i]-1);
							exit(BOTTLENECK_ERR);
						}

						//set the atomic value to zero (it will be used to understand when the reconfiguration has finished)
						repo->reconfiguration_finished[scheduling_table[i]-1].store(0);	

						//send the signal to worker that from now on will receive that class
						signalt=(task_t*)malloc(sizeof(task_t));
						signalt->type=i; //signal task
						signalt->punctuation=MOVING_IN;
						if(!send(signalt,outqueue[reconf_data->scheduling_table[i]-1]))
						{
							fprintf(stderr,ANSI_COLOR_RED "[EMITTER] Worker %d is a bottleneck\n" ANSI_COLOR_RESET,reconf_data->scheduling_table[i]-1);
							exit(BOTTLENECK_ERR);
						}	
						repo->reconfiguration_finished[reconf_data->scheduling_table[i]-1].store(0);
					}

					//copy the entry in the scheduling table
					scheduling_table[i]=reconf_data->scheduling_table[i];
				}
				//if this is a reconfiguration that requires the termination of some worker,
				//send the EOS to them (it is assumed that scheduling table was correctly derived from the strategy)
				if(reconf_data->par_degree_changes<0)
				{
					//we will terminate the workers with highest id
					for(int i=0;i>reconf_data->par_degree_changes;i--)
					{
						bsend(&eos_t,outqueue[num_workers+i-1]);
						//set to NULL outqueues (they will destructed by workers)
						outqueue[num_workers+i-1]=NULL;
					}
					//set the new num_workers
					num_workers+=reconf_data->par_degree_changes;

				}

				//ok, notify to the collector that the reconfiguration started (we don't need a specific message, since it is a well defined req-reply protocol)
				// printf("Emitter: send notification to controller\n");
				//bsend((void **)1,cn_outqueue); //nota potremmo avere comunque dei problemi di sincronizzazione (inviano contemporaneamente)
				bsend((void*)&reconf_finished,cn_outqueue);
			}
			#endif
			//check if we can send monitoring data to controller
			//if(getticks()-monitoring_timer>monitoring_step)
			volatile ticks act_time=getticks();
			if(act_time>sub_monitoring_timer)
			{
				//per ora approssimato...vediamo quante tuple ci sono nel socket
				int value;
	            ioctl(socket, SIOCINQ, &value);
	            int diff=value/sizeof(task_t)-prev_numb_tuples; //se abbiamo un solo gradino questo non funziona
	            if(diff>max_diff)
	            {
	            	max_diff2=max_diff;
	            	max_diff=diff;
	            	// printf("Max diff: %d\n",max_diff);
	            }
	            //devi vedere se alla fine e' significativo
	            prev_numb_tuples=value/sizeof(task_t);
	            numb_tuples[sub_mon_idx++]=value/sizeof(task_t);
	            sum_numb_tuples+=value/sizeof(task_t);
	            sub_monitoring_timer+=sub_monitoring_step;
	        }
			if(act_time>monitoring_timer)
			{ 	
				
				asm volatile("":::"memory");
				//consider also the elements in the buffer:
				int value;
            	ioctl(socket, SIOCINQ, &value);
				
           		
            	monitoring[mon_ptr].buffer_elements=value/sizeof(task_t);
            	//compute slope in socket buffer
            	double sx=0,sxy=0;
            	double xavg=sub_mon_idx*(sub_mon_idx-1)/2; //we start from zero
            	for(int i=0;i<sub_mon_idx;i++)
            	{
            		sx+=(i-xavg)*(i-xavg);
            		sxy+=(i-xavg)*(numb_tuples[i]-sum_numb_tuples/sub_mon_idx);
            		//printf("%d ",numb_tuples[i]);
            	}
            	//printf("\n");
            	double slope=sxy/sx;
            	printf(ANSI_COLOR_MAGENTA "-----------Emitter: buffered tasks: %d, grow rate: %g slope:%f\n" ANSI_COLOR_RESET,value/sizeof(task_t),max_diff*9.5,slope);
            	if(slope>0)
            		monitoring[mon_ptr].grow_rate_per_step=max_diff*9.5;
            	else
            		monitoring[mon_ptr].grow_rate_per_step=0;
		
				monitoring[mon_ptr].scheduling_table=scheduling_table; //NOTA: per il momento passiamo il puntatore (o fare la copia? dipende dal controllore che fa)
				
				//printf("Emitter sending...%d\n",cn_outqueue->length());
				bsend(&monitoring[mon_ptr],cn_outqueue); //pay attention, if queue is full it is blocked

				//printf("Emitter sent step: %d\n",mon_step++);
				monitoring_timer=getticks();
				mon_ptr=(mon_ptr+1)%NMONITORING;
				//reset the monitoring data
				monitoring[mon_ptr].elements=0;
				monitoring[mon_ptr].buffer_elements=0;
				grow_rate=0;
				max_diff=0;
				max_diff2=0;
				sum_numb_tuples=0;
				sub_mon_idx=0;
				memset(monitoring[mon_ptr].elements_per_class,0,num_classes*sizeof(int));
				monitoring_timer+=monitoring_step;

			}
			


			
		#endif
		if((ret=socket_receive(socket, tb, sizeof(task_t)))!=sizeof(task_t))
		{
			fprintf(stderr,"The program is a bottleneck?\n");
			exit(BOTTLENECK_ERR);
		}
		
	}
	printf("Emitter, send termination----------\n");

	//EOS: send terminating task to all the workers
	for(int i=0;i<num_workers;i++)
	{
		//printf("Emitter, send EOS to:%d\n",i);
		send(tb,outqueue[i]);
	}
	
	#if defined(MONITORING)
		//send termination to controller
		monitoring[mon_ptr].stop=true;
		bsend(&monitoring[mon_ptr],cn_outqueue);
	#endif
	//compute statistics
	gettimeofday(&tmp_t,NULL);
	end_t=tmp_t.tv_sec*1000+(tmp_t.tv_usec)/1000-start_t;
	//pthread_barrier_wait(barrier);
	printf("***EMITTER***\n");
	printf("Tempo passato (msec): %Ld, msg per second: %6.3f\n",end_t,((double)msg)/((double)end_t/1000));
		
	#if defined (USE_ZMQ)
	close(socket);
	zmq_ctx_destroy (context);
	#endif


	return NULL;
}
