/**
	Worker for Functional Partitioning Implementation
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <ff/buffer.hpp>
#include <iostream>
#include <unordered_map>
#include <vector>
#include <sched.h>
#include <set>
#include "../Includes/cycle.h"
#include "../Includes/iaperf.h"
#include "../Includes/general.h"
#include "../Includes/functional_partitioning_cb.h"
#include "../Includes/cbwindow.hpp"
#include "../Includes/fp_cb_reconf.h"
#include <ff/allocator.hpp>	
#if defined(USE_ZMQ)
#include <zmq.h>
#endif
using namespace ff;
using namespace std;

void * worker(void *args) {

	task_t *tmp;
	CBWindow *window;
	HReturn_type *res_buff; //result buffer in order to reuse memory
	int buff_size;
	int bi=0;
	//Init
	worker_data_t *data = (worker_data_t *) args;
	pthread_barrier_t *barrier = data->barrier;
	int id = data->workerId;
	SWSR_Ptr_Buffer *inqueue=data->inqueue;
	SWSR_Ptr_Buffer *outqueue=data->outqueue;
	int window_slide=data->window_slide;
	int window_size=data->window_size;
	long int freq=data->freq;
	int num_classes=data->num_classes;
	ticks *start_global_ticks=data->start_global_ticks;
	#if !defined(DONT_USE_FFALLOC) && !defined(TASK_BUFF)
			ff_allocator *ffalloc=data->ffalloc;
	#endif
	#if defined(MONITORING)
		SWSR_Ptr_Buffer *cn_outqueue=data->cn_outqueue;	
	#endif
	unordered_map<int,CBWindow*> map; //map that contains the various windows
	//memory allocation
	buff_size=QUEUE_SIZE+10;//is it correct?
	posix_memalign((void **)&res_buff,CACHE_LINE_SIZE, buff_size*sizeof(HReturn_type));
	
	//int64_t triggers[100000];
	//int64_t computed=0;
	#if defined(MONITORING)
		//define the data structures for monitoring
		
		ticks start;

		volatile ticks monitoring_timer;
		volatile ticks monitoring_step=MONITORING_STEP*1000*freq; //after how many cycles we have to monitor
		worker_monitoring_t *monitoring;
		char mon_ptr=0; //the current monitoring structure in which write data
		posix_memalign((void **)&monitoring, CACHE_LINE_SIZE,NMONITORING*sizeof(worker_monitoring_t));
		for(int i=0;i<NMONITORING;i++)
		{
			//init the internals
			posix_memalign((void**)&(monitoring[i].tcalc_per_class),CACHE_LINE_SIZE,num_classes*sizeof(double));
			memset(monitoring[i].tcalc_per_class,0,num_classes*sizeof(double));
			posix_memalign((void **)&(monitoring[i].elements_per_class),CACHE_LINE_SIZE,num_classes*sizeof(int));
			memset(monitoring[i].elements_per_class,0,num_classes*sizeof(int));
			posix_memalign((void **)&(monitoring[i].computations_per_class),CACHE_LINE_SIZE,num_classes*sizeof(int));
			memset(monitoring[i].computations_per_class,0,num_classes*sizeof(int));
			monitoring[i].elements_rcvd=0;
			monitoring[i].computations=0;

		}
		//adaptivity data structures
		reconf_repository_t *repo=(reconf_repository_t *)data->repo;
		// unordered_map<int,CBWindow*> moving_windows=*(repo->moving_windows);
		CBWindow** moving_windows=(repo->moving_windows);
		bool reconfiguration_phase_in=false; //indicates if the worker is currently in the reconfiguration phase (moving_in)
		bool reconfiguration_phase_out=false; //indicates if the worker is currently in the reconfiguration phase (moving_out)
		bool reconfiguration_phase=false; //indicate the whole process
		set<int> classes_moving_in; //it will contains all the id of the classes that are currently moving in
		vector<task_t *> task_moving_in(20000); //it will contain all the task belonging to moving in classes (not yet arrived to the worker) (to see why i choose vector look at the explnation below, when i use it)
		task_moving_in.clear();
	#endif



	if(barrier)	//it is a thread spawned at the beginning
	{
		//queues warmup
		receive((void **)&tmp,inqueue);
		send((void *)1,outqueue);
		//synchronization barrier
		// DEBUG(printf("Worker %d ready, running on %d mon_step: %Ld\n",id,sched_getcpu(),monitoring_step));
		pthread_barrier_wait(barrier);
	}

	while(*start_global_ticks==0)
		REPEAT_25(asm volatile("PAUSE" ::: "memory");) //wait for the correct time
	

	#if defined(MONITORING)
	if(barrier)
		monitoring_timer=*start_global_ticks+monitoring_step; //we skip the first step due to starting phase
	else //it a newly thread
	{
		monitoring_timer=*start_global_ticks+((getticks()-*start_global_ticks)/monitoring_step+1)*monitoring_step;
		 //printf("%d Prima del monitoring ho %Ld\n",id,monitoring_timer-getticks());
	}
	#endif
	//buffer of results
	
	/**
		Start receiving elements
	*/

	receive((void **)&tmp,inqueue);

	int ret;
	int mon_step=0;
	int count=0;
	ticks no_trigger=0;
	int no_trigger_tuple=0;
	while(tmp->type!=-1)
	{
		/*
			copy the task into the window and if necessary, perform computation
		*/
		//printf("Ricevuto elemento %Ld con tipo: %d\n",tmp->id,tmp->type);
		#if defined(MONITORING)
			asm volatile("":::"memory");
			start=getticks();
			monitoring[mon_ptr].elements_per_class[tmp->type]++;
			monitoring[mon_ptr].elements_rcvd++;
			asm volatile("":::"memory");
		#endif
		#if defined(ADAPTIVITY)
		if(tmp->punctuation==NO) //standard task
		{
			//NOTE: probably this series of if can be simplified but i don't know how.
			//The point is that we have to be able to ingest elements that do not belong to classes involved 
			//in reconfiguration and, at the same time, guarantee that a reconfiguration take place in finite time

			reconfiguration_phase_out=false; //this for sure (reconfiguration messages are sent all togheter from the emitter before new standard task)
			//if we are in a reconfiguration phase look if the task arrived refers to a moving in class
			if(reconfiguration_phase_in && classes_moving_in.size()>0)
			{
				set<int>::iterator it = classes_moving_in.begin();
				for (it = classes_moving_in.begin(); it != classes_moving_in.end(); )  //check if moving windows are ready 
				{
					int moving_class=*it;
					//let's see if in the repository there is the class's window for it, in case acquire it
					// printf("Worker: %d provo a portare dentro classe: %d\n",id,moving_class);
					// printf("%d %p\n",id,(*(repo->moving_windows))[moving_class]);
					//if((*(repo->moving_windows))[moving_class])
					if(moving_windows[moving_class])
					{
						//take it, add to the map of class's windows hold by the worker
						//add to my map
						window=moving_windows[moving_class];
						map[moving_class]=window;
						//remove from repository (but not delete the window)
						moving_windows[moving_class]=NULL; 
						//erase from the set of class thare are currently ''come'' toward this worker
						classes_moving_in.erase(it++);
						//printf("MOVED_IN class: %d Total elements:%Ld Ins pointer: %d\n", moving_class,window->getTotalElements(),window->getInsertionPointer());
						//look in the vector of task arrived during the coming_in phase: if some of them refer to this class add to the window
						//we don't erase them from the vector: this will be an efficient operation
						//on the other hand, use a structure such as a list if efficient in term of insertion operations (each of them will require a memory allocation)
						//for this reason, elements are kept in the vector and when the reconfiguration phase phinishes the whole vector is cleared
						for(int i=0;i<task_moving_in.size();i++)
						{ 
							if(task_moving_in[i]->type==moving_class)
							{
								//printf("Inserisco task con id: %Ld\n",task_moving_in[i]->internal_id);
								ret=window->insert(task_moving_in[i],&res_buff[bi]);
								if (ret!=0) //we have a new result 
								{
									//set the timestamp to the timestamp of the one of the task that has triggered the computation
									res_buff[bi].timestamp=task_moving_in[i]->timestamp;

									//the same for the id
									res_buff[bi].id=task_moving_in[i]->internal_id; //for the moment just for ordering
									if((res_buff[bi].id+1)%500!=0)
									{
										printf("-----------Worker %d error, computed on %Ld, class: %d\n",id, task_moving_in[i]->internal_id,task_moving_in[i]->type );
										exit(-1);
									}
									res_buff[bi].type=task_moving_in[i]->type;
									res_buff[bi].isEOS=false;
									res_buff[bi].wid=id;
									//send result to the collector
									send(&res_buff[bi],outqueue);
									//advance the buffer index
									bi=(bi+1)%buff_size;

									//it is not meaningful to perform the monitoring in this situation
									/*
									#if defined(MONITORING) 
									asm volatile("":::"memory");
									monitoring[mon_ptr].tcalc_per_class[task_moving_in[i]->type]+=(double)(getticks()-start)/freq; //usec
									monitoring[mon_ptr].computations_per_class[task_moving_in[i]->type]++;
									asm volatile("":::"memory");
									#endif*/
						
								}
								//later on we will free the space occupied by this tasks
								
							}
						}
					}
					else
						++it;
				}

				if(classes_moving_in.size()==0)
				{
					//we finished the reconfiguration phase
					reconfiguration_phase_in=false;
					//free the elements in the vector (we cannot free them previously otherwise new task will take their place in vector)
					#if !defined(TASK_BUFF)
					for(int i=0;i<task_moving_in.size();i++)
					{
								#if defined(DONT_USE_FFALLOC)
									free(task_moving_in[i]);
								#else
								ffalloc->free(task_moving_in[i]);
								#endif
					}
					#endif
					task_moving_in.clear(); 

				}
				
			}
			//if we were in a reconfiguration phase, check if all is finished
			if(reconfiguration_phase)
			{//(necessary if we have only a reconf out)
				if (!reconfiguration_phase_in && !reconfiguration_phase_out)
				{
					//FINISHED
					//printf(ANSI_COLOR_YELLOW "[WORKER %d] reconf finished\n" ANSI_COLOR_RESET,id);
					reconfiguration_phase=false;
					repo->reconfiguration_finished[id].store(1);
				}
			}
			if(reconfiguration_phase_in && classes_moving_in.count(tmp->type)==1) //the new task refers to a still moving in class
			{
				//we don't see the class in the repository
				//add the task to task_coming_in (respecting the order)
					
				task_moving_in.push_back(tmp);
			}
			else
			{
							
				//it is a task that refer to a class currently held by the worker (or that it has just moved in)
			
		#endif
				
				/*count++;

				if(count==500)
				{
					//volatile ticks st=getticks();
					//while(getticks()-st<200000);
					//usleep(3050);
					count=0;
					res_buff[bi].timestamp=tmp->timestamp;
					send(&res_buff[bi],outqueue);
					bi=(bi+1)%buff_size;
					#if defined(MONITORING)
					asm volatile("":::"memory");
					monitoring[mon_ptr].tcalc_per_class[tmp->type]+=(double)(getticks()-start)/freq; //usec
					monitoring[mon_ptr].computations_per_class[tmp->type]++;
					asm volatile("":::"memory");
					#endif
				}*/
				window=map[tmp->type];
				
				if(window==NULL) //it's a new logical stream that has arrived
				{
					//DEBUG(printf("--------------------[%d] Ricevuto un elemento di un nuovo stream logico: %d\n",id,tmp->type));
					//qui possiamo ricavarci una grana: ad esempio facciamo 7+class_id%4 *2

					//window=new CBWindow(window_size, window_slide,(tmp->type%4)*2+7);
					window=new CBWindow(window_size, window_slide);
					map[tmp->type]=window;
				}
				//insert the element in window
				//res_buff[bi].ts=getticks();
				
				ret=window->insert(tmp,&res_buff[bi]);
				if (ret!=0) //we have a new result (NOTE: actually, it will be better to handle this situation in a different way)
				{
					// if(tmp->type==68)
					// printf("Computato per task: %Ld, classe: %d\n",tmp->id, tmp->type);
					//set the timestamp to the timestamp of the one of the task that has triggered the computation
					res_buff[bi].timestamp=tmp->timestamp;

					//the same for the id
					res_buff[bi].id=tmp->internal_id; //for the moment just for ordering
					res_buff[bi].type=tmp->type;
					res_buff[bi].isEOS=false;
					res_buff[bi].wid=(char)id;

					//triggers[computed]=tmp->id;
					//computed++;
					/*
					if((res_buff[bi].id+1)%500!=0)
					{
						printf("********orker %d error, computed on %Ld, class:%d\n",id, tmp->internal_id,tmp->type );
						exit(-1);
					}*/
					//to be sent to the collector
					//send result to the collector
					send(&res_buff[bi],outqueue);
					//advance the buffer index
					bi=(bi+1)%buff_size;
					#if defined(MONITORING)
					asm volatile("":::"memory");
					monitoring[mon_ptr].tcalc_per_class[tmp->type]+=(double)(getticks()-start)/freq; //usec
					monitoring[mon_ptr].computations_per_class[tmp->type]++;
					monitoring[mon_ptr].computations++;
					asm volatile("":::"memory");
					#endif
				}
				// else
				// {
				// 	no_trigger+=getticks()-start;
				// 	no_trigger_tuple++;
				// }
				
				
				#if defined(MONITORING)
					//check if we can send monitoring data to controller
					

					//if(getticks()-monitoring_timer>monitoring_step)
					if(getticks()>monitoring_timer)
					{ 	
						asm volatile("":::"memory");


						// printf("Monitoring data %d are valid. Elements received: %Ld\n",mon_ptr,monitoring[mon_ptr].elements);
						
						//blocking send towards the collector
						// printf("Worker %d sends %d step: %d\n",id,cn_outqueue->length(),mon_step++);
						bsend(&monitoring[mon_ptr],cn_outqueue);
						
						// printf("Worker %d sent\n",id);
						monitoring_timer=getticks();
						mon_ptr=(mon_ptr+1)%NMONITORING;
						//reset the monitoring data
						memset(monitoring[mon_ptr].elements_per_class,0,num_classes*sizeof(int));
						memset(monitoring[mon_ptr].tcalc_per_class,0,num_classes*sizeof(double));
						memset(monitoring[mon_ptr].computations_per_class,0,num_classes*sizeof(int));
						monitoring[mon_ptr].elements_rcvd=0;
						monitoring[mon_ptr].computations=0;
						monitoring_timer+=monitoring_step;
						// printf("Empty average tcalc: %Lf usec\n",(((long double)no_trigger)/no_trigger_tuple)/freq);
						// no_trigger=0;
						// no_trigger_tuple=0;
					}
				#endif

				#if !defined(TASK_BUFF)
					#if defined(DONT_USE_FFALLOC)
						free(tmp);
					#else
					ffalloc->free(tmp);
					#endif
				#endif


		#if defined(ADAPTIVITY)
			}
		}
		else
			if(tmp->punctuation==MOVING_OUT)
			{
				if(!reconfiguration_phase_out)
				{
					reconfiguration_phase_out=true; //we are entering in the reconfiguration phase
					reconfiguration_phase=true;
				}
				//printf("Worker %d: moving OUT class %d, last inserted:%Ld total elements: %Ld ins pointer: %d\n",id,tmp->type,map[tmp->type]->getLastInserted(),map[tmp->type]->getTotalElements(),map[tmp->type]->getInsertionPointer());
				//write it into the repository
				if(!(window=map[tmp->type]))
				{
					//the worker does not have this class. Probably it is not yet arrived. We will create a new window
					//and insert it into the repository
						
					window=new CBWindow(window_size, window_slide);
					moving_windows[tmp->type]=window;
					map.erase(tmp->type); 
				}
				else
				{
					//printf("Worker %d: moving OUT class %d, last inserted:%Ld total elements: %Ld\n",id,tmp->type,map[tmp->type]->getLastInserted(),map[tmp->type]->getTotalElements());
					moving_windows[tmp->type]=map[tmp->type];
					//printf("Total elements:%Ld\n",(*(repo->moving_windows))[tmp->type]->getTotalElements());
					//remove from the worker's map
					map.erase(tmp->type); //NOTA: se ci fossero problemi (in teoria dovrebbe richiamare il distruttore), semplicemente mettere a NULL
				}
				free(tmp); //allocated by the emitter
			}
			else
				if(tmp->punctuation==MOVING_IN)
				{
					if(!reconfiguration_phase_in)
					{
						reconfiguration_phase_in=true; //we are entering in the reconf phase for incoming classes
						reconfiguration_phase=true;
					}
					
					//printf("Worker %d: moving IN class %d\n",id,tmp->type);
					classes_moving_in.insert(tmp->type);
					
					free(tmp); //allocated by the emitter
				}
		#endif 
		
		//receive the next element
		receive((void **)&tmp,inqueue);
	}
	#if defined(ADAPTIVITY)
	//at this point no other valid tuple can arrive, we have only to wait some moving in class if any
	while(reconfiguration_phase_in && classes_moving_in.size()>0)
	{
					
		
		set<int>::iterator it = classes_moving_in.begin();

		// iterate through the set and erase 
		for (it = classes_moving_in.begin(); it != classes_moving_in.end(); ) 
		{
			//let's see if in the repository there is the class's window for it, in case acquire it
			int moving_class=*it;
			if(moving_windows[moving_class])
			{
				
				//take it, add to the map of class's windows hold by the worker
				window=moving_windows[moving_class];
				map[moving_class]=window;

				//remove from repository (but not delete the window)
				moving_windows[moving_class]=NULL; 
				//erase from the set of class thare are currently ''come'' toward this worker
				classes_moving_in.erase(it++);
				//look in the vector of task arrived during the coming_in phase: if some of them refer to this class add to the window
				//we don't erase them from the vector: this will be an efficient operation
				//on the other hand, use a structure such as a list if efficient in term of insertion operations (each of them will require a memory allocation)
				//for this reason, elements are kept in the vector and when the reconfiguration phase phinishes the whole vector is cleared
				for(int i=0;i<task_moving_in.size();i++)
				{ 
					if(task_moving_in[i]->type==moving_class)
					{
						ret=window->insert(task_moving_in[i],&res_buff[bi]);
						if (ret!=0) //we have a new result 
						{
							//set the timestamp to the timestamp of the one of the task that has triggered the computation
							res_buff[bi].timestamp=task_moving_in[i]->timestamp;
							//the same for the id
							res_buff[bi].id=task_moving_in[i]->internal_id; //for the moment just for ordering
							if((res_buff[bi].id+1)%window_slide!=0)
							{
								printf("-----------Worker %d error, computed on %Ld, class: %d\n",id, task_moving_in[i]->internal_id,task_moving_in[i]->type );
								exit(-1);
							}
							res_buff[bi].type=task_moving_in[i]->type;
							res_buff[bi].isEOS=false;
							res_buff[bi].wid=id;
							//send result to the collector
							send(&res_buff[bi],outqueue);
							//advance the buffer index
							bi=(bi+1)%buff_size;				
						}
					}
				}
			}
			else
				++it;
		}

		if(classes_moving_in.size()==0)
		{
			//we finished the reconfiguration phase
			reconfiguration_phase_in=false;
			//free the elements in the vector (we cannot free them previously otherwise new task will take their place in vector)
			#if !defined(TASK_BUFF)
			for(int i=0;i<task_moving_in.size();i++)
			{
						#if defined(DONT_USE_FFALLOC)
							free(task_moving_in[i]);
						#else
						ffalloc->free(task_moving_in[i]);
						#endif
			}
			#endif
			task_moving_in.clear(); 
			repo->reconfiguration_finished[id].store(1);
		}	
	}
	#endif
	#if defined(MONITORING) //BOHHHHHHHHHHHHHHHHHHHHHHHHHHHHH??????????????????????????
	//send the last monitoring, in order to avoid stall on the controller
	//bsend(&monitoring[mon_ptr],cn_outqueue);
	bsend((void *)(1),cn_outqueue);
	#endif
	//send EOS to collector
	
	res_buff[bi].isEOS=true;
	res_buff[bi].res_buff=res_buff;
	
	send(&res_buff[bi],outqueue);
	//printf("Worker %d ricevuto EOS and sended to collector, rcvd buff: %p\n",id,res_buff[bi].res_buff);
	//MEMORY FREEING
	//destroy queues from emitter and toward collector
	delete(inqueue);
	//delete(cn_outqueue); this has to be done by controller
	//there is no need to free tmp(it can be statically allocated on the emitter )
	//delete structure for monitoring, except the last one sent?????
	for(int i=0;i<NMONITORING;i++)
	{
		//init the internals
		free(monitoring[i].tcalc_per_class);
		monitoring[i].tcalc_per_class=NULL;
		free(monitoring[i].elements_per_class);
		monitoring[i].elements_per_class=NULL;
		free(monitoring[i].computations_per_class);
		monitoring[i].computations_per_class=NULL;
		
	}
	free(monitoring);
	monitoring=NULL;
	//since it is still used, res_buff will be destroyed by the collector
	
	//printf("Worker %d finito\n",id);

	//pthread_barrier_wait(barrier);
	//sleep(id+1);

	/*for(int i=0;i<240;i++)
		if(counts[i]>0)
		{
			comp_time[i]=((double)computation_time[i])/counts[i]/FREQ;
		}
*/
	/*for ( unsigned i = 0; i < map.bucket_count(); ++i) {
   		for ( auto local_it = map.begin(i); local_it!= map.end(i); ++local_it ){
    		//printf("Worker %d terminated. Class: %d Average computation time empy (usec): \n",id,(local_it->first));
    		comp_time[local_it->first]=(local_it->second->getAvgComputationTime())/freq;
    	}
  	}*/
   /* char fname[20];
    sprintf(fname,"worker_%d.dat",id);
    FILE *fout=fopen(fname,"w");
    for(int i=0;i<computed;i++)
    	fprintf(fout,"%Ld\n",triggers[i]);
    fclose(fout);*/
    return NULL;
}
