#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <vector>
#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>
#include <sched.h>
#include <unistd.h>
#include "../Includes/RandomGenerator.h"
#include "../Includes/cycle.h"
#include "../Includes/iaperf.h"
#include "../Includes/general.h"
#include "../Includes/zipf.h"
#include "../Includes/affinities.h"
#define NUM_SAMPLES 8388608 //number of sample classes used for the generation of tasks
#if defined(USE_ZMQ)
#include <zmq.h>
#endif

int num_task=1000000;
int numb_class;
//Funzione stupida per confrontare double:
int compare (const void *first, const void *second) {
	double *f = (double *) first;
	double *s = (double *) second;
	if(*f > *s) return -1;
	else if(*f == *s) return 0;
	else return 1;
}
//Optimization, otherwise it will compute any time
static double *sums=NULL;
inline void recomputeSum(double *probs, int numClass)  __attribute__((always_inline));
inline void recomputeSum(double *probs, int numClass)
{
	sums = (double *) malloc(sizeof(double) * numClass);
    bzero((void *) sums, numClass * sizeof(double));
    sums[0] = probs[0];
    for(int i = 1; i < numClass; i++) {
      sums[i] = sums[i-1] + probs[i];
      //printf("%d %1.6f\n",i,sums[i]);
    }




}
inline int generateRandomClass(double *probs, int numClass) __attribute__((always_inline));
int bsearch(double AR[], int N, double VAL) __attribute__((always_inline));
inline int bsearch(double AR[], int N, double VAL)
{
	int Mid,Lbound=0,Ubound=N-1;

	while(Lbound<=Ubound)
	{
		//printf("Cerco %f, valore di mezzo: %f (%d,%d)\n",VAL,AR[Mid],Lbound,Ubound);
		Mid=(Lbound+Ubound)/2;
		if(VAL>AR[Mid])
			Lbound=Mid+1;
		else if(VAL<AR[Mid])
			Ubound=Mid-1;
		else
			return Mid;
	}

	if(VAL<=AR[Lbound])
	{
		//printf("%6.4f < %6.4f, indice: %d\n",VAL,AR[Lbound],Lbound);
		return Lbound;
	}
	else
		if(Lbound<N-1)return Lbound+1;

}
inline int generateRandomClass(double *probs, int numClass) {
  //Creo un array dove la posizione i indica la somma delle prime i probabilità:
  if(sums==NULL)
  { 
    recomputeSum(probs,numClass);
  }

  //Genero un numero casuale uniformemente distribuito in [0,1]: 
  double random = rand()/((double)RAND_MAX + 1);

  //Controllo a quale classe appartiene:
  int index = bsearch(sums,numClass,random);
 // printf("Indice per %6.3f e' %d\n",random,index);
 /* bool trovato = false;
  while((!trovato) && (index < numClass)) {
    if(random <= sums[index]) trovato = true;
    else index++;
  }*/


  //if(!trovato) {
    //we suppose that this is a probability vector, hence exists the right class
    /*printf("cercavo: %1.6f\n",random);
    printf("Emitter: class not found! This will cause PROBLEMS!!\n");
    index = -1;*/
   // index=numClass-1;
 // }
  return index;
}

/*
	Reads the distribution from the file passed as argument
	File format:
		<num_class> <num_distributions>
		<time_0> <vect_0>
		...
	where times are expressed in second and represent the time from which the relative distribution (expressed with the vector
	of probability) is valid.
	The first time is assumed to be 0.

	Return the number of distribution.
	The pointer to the vector of times and the pointer to the matrix of distributions are reported in the 
	the parameters times and distributions
*/
int readDistributions(char *file, int **times, double ***distributions)
{
	FILE *fdistr=fopen(file,"r");
	
	int nclass;
	int ndistributions;
	int i,j;
	fscanf(fdistr,"%d %d\n",&nclass,&ndistributions);
	//allocate the space
	int *t=(int *)malloc(sizeof(int)*nclass);
	double **d=(double **)malloc(sizeof(double *)*ndistributions);
	for(i=0;i<ndistributions;i++)
		d[i]=(double *)malloc(sizeof(double)*nclass);

	//start reading
	for(i=0;i<ndistributions;i++)
	{
		fscanf(fdistr, "%d",&t[i]);
		for(j=0;j<nclass;j++)
		{
			fscanf(fdistr,"%lf",&d[i][j]);
			// printf("%3.5f ",d[i][j]);
		}
		
	}
	// printf("\n");
	*times=t;
	*distributions=d;
	return ndistributions;

}

/*
	Reads additional data rate (part from the one indicated via command line) from file
	FILE format: 
	 	- first row: number of rates
	 	- for each row there is the time at which the rate become valid (in second) and the rate expressed in tuples/s
	  It is assumed that the first time is  zero (the base rate)
	*/
int readRates(char *file, int **rate_times, double **rates)
{
	FILE *frate=fopen(file,"r");
	int nrates;
	fscanf(frate,"%d\n",&nrates);
	
	int *rt=(int *)malloc(sizeof(int)*nrates);
	double *r=(double *)malloc(sizeof(double)*nrates);
	for(int i=0;i<nrates;i++)
	{
		fscanf(frate,"%d %lf\n",&rt[i],&r[i]);
		if(i==0)
			assert(rt[i]==0);
	}
	fclose(frate);
	*rate_times=rt;
	*rates=r;
	return nrates;
}

int main(int argc, char *argv[])
{
	
	PERFINITMHZ(FREQ);
	cpu_set_t cpuset;
	struct timeval tmp_t;
	int ret;
	//rates from file
	int nrates=0;
	double *rates=NULL;
	int *rate_times=NULL;
	int next_rate=0;
	if(argc<7)
	{
		printf("Usage: %s hostname port numb_class distribution_file rate(msg/sec) num_elements [rate_file]\n", argv[0]);
		exit(-1);
	}
	numb_class=atoi(argv[3]);
	// double alpha=atof(argv[4]);
	double rate=atof(argv[5]);
	num_task=atoi(argv[6]);
	srand(10);
	#ifdef EXPONENTIAL
		printf("Generator: exponential distribution\n");
	#else
		#ifdef NORMAL
			printf("Generator: normal distribution\n");
		#else
			printf("Generator: discrete distribution\n");
		#endif
	#endif
	//set affinity
	CPU_ZERO(&cpuset);
	CPU_SET(generator_affinity, &cpuset);
	pthread_t tid = pthread_self();

	if (pthread_setaffinity_np(tid, sizeof(cpu_set_t), &cpuset)) {
		fprintf(stderr,"Cannot set thread to CPU %d\n", generator_affinity);
	}
		
	//for monitoring during testing phase
	vector<double> mon_times; //times at which we take the various metric
	vector<double> mon_rates;
	vector<int> mon_distribution; //probability distribution used
	//Reads the distributions from file
	int *times;
	double **distributions;
	int ndistributions=readDistributions(argv[4],&times, &distributions);
	/*
		now we generate the set of samples classes (distributed according the read distribution)
		in order to avoid to generate them on the fly
	*/
	/*int **classes_samples=(int **)malloc(sizeof(int *)*ndistributions);
	for(int i=0;i<ndistributions;i++)
	{
		classes_samples[i]=(int *)malloc(sizeof(int)*NUM_SAMPLES);
		recomputeSum(distributions[i],numb_class);
		for(int j=0;j<NUM_SAMPLES;j++)
		{
			classes_samples[i][j]=generateRandomClass(distributions[i],numb_class);
		}
	}*/
	int next_distr=1;
	int act_distr=0;
	double act_rate=rate;
	if(argc == 8)
	{
		
		nrates=readRates(argv[7],&rate_times,&rates);
		rate=rates[0];
		next_rate=1; //the next is the first read
	}

	printf("Generator, running on %d, read %d probability distributions\n",sched_getcpu(),ndistributions);
	printf("Rates: %d: ",nrates);
	for(int i=0;i<nrates;i++)
		printf("%7f ",rates[i]);
	printf("\n");

	//attenzione genero tutto prima
	/*printf("Generatore, attenzione genero tutto prima!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
	int* gen_classes=(int *)calloc(num_task,sizeof(int));
	for(int i=0;i<num_task;i++)
		gen_classes[i]=generateRandomClass(distributions[act_distr],numb_class);*/
	sleep(1);
	#if !defined(USE_ZMQ)
	int socket=connect_to(argv[1],atoi(argv[2]));
	//if(!SetSocketBlockingEnabled(socket, true))
	//{
	//	fprintf(stderr, "Cannot set Nonblocking socket\n");
	//	exit(-1);
	//}
	#else
	void *context = zmq_ctx_new ();
	void *socket=connect_to(context,argv[1],atoi(argv[2]));
	#endif


	// fcntl(socket, F_SETFL, O_NONBLOCK);
	// printf("Generator with non blocking socket\n");

	//sleep(1);
	task_t task;
	task.timestamp=0;
	//double *probs=getDistributionZipf(numb_class,alpha);
	//for(int i=0; i<numb_class;i++)
	//	printf("%d %1.4f\n",i,probs[i]);

	//the same seed in order to have the same executions
	RandomGenerator generator(1);
	double waitingTime=((double)1)/(rate);
	

	volatile ticks cicli_tot = 0;
	//ora facciamo la parte del sender, ci connettiamo al receiver
	gettimeofday(&tmp_t,NULL);
	
	
	//send the first dummy task for synchronizing the global start time...
	volatile ticks start_ticks=getticks();
	long int start_t=tmp_t.tv_sec*1000+(tmp_t.tv_usec)/1000;
	task_t t;
	t.type=-10;
	ticks no_more_init=(unsigned long long)FREQ*NO_MORE_INIT; //30 secondi
	t.timestamp=getticks(); //this has to be changed in case of distributed implementation

	if((ret=socket_send(socket,&t,sizeof(task_t)))!=sizeof(task_t))
	{
		
			fprintf(stderr,"The receiving program is a bottlenck\n");
			exit(BOTTLENECK_ERR);
		
	}
	//sleep(1);
	
	
	
	bool bsend=true;
	int sent=0;
	long int start_slot=start_t;
	for(int i=0;i<num_task;i++)
	{
		//periodically check if we have to change probability distribution
		if(i%50000==0 )
		{
			gettimeofday(&tmp_t,NULL);
			long int end_t=tmp_t.tv_sec*1000+(tmp_t.tv_usec)/1000-start_t;
			if(next_distr<ndistributions && end_t>times[next_distr]*1000)
			{
				/*printf("Time: %d ",times[next_distr]);
				for(int j=0;j<numb_class;j++)
					printf("%d %1.5f ",j,distributions[next_distr][j]);
				printf("\n");
				*/
				printf("Compute new probabilities\n");
					
				recomputeSum(distributions[next_distr],numb_class);
				act_distr=next_distr;
				next_distr++;

			}
			//printf("check rate: %Ld\n",i);
			if(next_rate<nrates && end_t>rate_times[next_rate]*1000)
			{
				//change rate
				printf("Changing rate to: %f\n",rates[next_rate]);
				//if(next_rate+1<nrates)
				//	printf("Next will be: %f - %d -%Ld -%Ld\n",rates[next_rate+1],((int)act_rate/8),i, (i%((int)act_rate/8)+1)*((int)act_rate/8) );
				waitingTime=((double)1)/(rates[next_rate]);
				act_rate=rates[next_rate];
				next_rate++;
			}
			//printf("Elapsed: %6.3f, actual rate (KT/s): %6.3f, distibution used\n",end_t/1000.0,(sent/(end_t-start_slot))*1.0);
			mon_times.push_back(end_t/1000.0);
			mon_rates.push_back((sent/(end_t-start_slot))*1.0);
			mon_distribution.push_back(act_distr);
			sent=0;
			start_slot=end_t;
		}
		//If it is specified a particular distribution, the waiting time is generated according to it
		#ifdef EXPONENTIAL
			waitingTime = generator.expntl(((double)1)/rate);
		#endif
		#ifdef NORMAL
			waitingTime = generator.normal(((double)1)/rate , ((double)1)/rate/10);
			
		#endif
		#ifdef UNIFORM
			waitingTime = generator.uniform(0, 2 * ((double)1)/rate);
		#endif
		
		//if(i>=27000000 && i%10000==0)
		//	printf("id: %Ld\n",i);
		task.id=i;
		//per farle triggerare tutte insieme
		//task.type=i%numb_class;
		//NEL CASO DI DISTRIBUZIONE EQUI POSSIAMO FARE UNA COSA PER INCREMENTARE IL THORUHGPUT DEL GENERATORE
		//NO< NON FUNGE
		 // task.type=rand()%numb_class;
		task.type=generateRandomClass(distributions[act_distr],numb_class);
		//task.type=classes_samples[act_distr][i%NUM_SAMPLES];
		//task.type=gen_classes[i];
		//task.type=i%numb_class;
		task.tag=generator.uniform(0, 1000);
		task.tag2=generator.random(0,500);
		volatile ticks end_cycles=start_ticks+task.timestamp; //the first task has timestamp=0
		volatile ticks cycles = getticks();
		while(cycles <= end_cycles) cycles = getticks();
			cicli_tot += cycles;
		
		// if (bsend)
		// 	ret=socket_bsend(socket,&task,sizeof(task_t));			
		// else
			ret=socket_send(socket,&task,sizeof(task_t));

		if(ret!=sizeof(task_t))
		{
			if(ret==BOTTLENECK_ERR)
			{
				//fprintf(stderr,"The receiving program is a bottleneck\n");
				#if defined(USE_ZMQ)
				//otherwise the peer does not recognize the disconection
				int err=BOTTLENECK_ERR;
				//we have to wait a little bit otherwise the message is discarded (the linger set to -1 does not work)
				sleep(2);
				socket_bsend(socket,&err,sizeof(int));
				fprintf(stderr,"The receiving program is a bottleneck\n");
				#endif
				exit(BOTTLENECK_ERR);
			}
		}
		//if(i>=27000000 && i%10000==0)
			//printf("id: %Ld\n",i);
		task.timestamp=task.timestamp+(waitingTime * FREQ*1000000); //compute the next timestamp
		if(bsend && getticks()-start_ticks>no_more_init)
			{
				bsend=false;
				printf("No more bsend\n");
				fcntl(socket, F_SETFL, O_NONBLOCK);
				printf("Generator with non blocking socket\n");

			}
		sent++;
	}
	task.type=-1;
	if((ret=socket_send(socket,&task,sizeof(task_t)))!=sizeof(task_t))
	{
		if(ret==BOTTLENECK_ERR)
		{
			fprintf(stderr,"The receiving program is a bottlenck at the end\n");
			exit(BOTTLENECK_ERR);
		}
	}
		gettimeofday(&tmp_t,NULL);
	long int end_t=tmp_t.tv_sec*1000+(tmp_t.tv_usec)/1000-start_t;
	printf("Gen: Tempo passato (msec): %Ld, msg per second: %6.3f\n",end_t,((double)num_task)/((double)end_t/1000));
	
	//print to file the various monitored metrics
	char fname[200];
	if(argc==7)
		sprintf(fname,"generator_%d_%d_%d_%s.dat",numb_class,(int)rate,num_task,argv[4]);
	else
		sprintf(fname,"generator_%d_%d_%d_%s_%s.dat",numb_class,(int)rate,num_task,argv[4],argv[7]);
	printf("%s\n",fname);
	FILE *fgenerator=fopen(fname,"w");
	fprintf(fgenerator, "TIME\tDISTR\tRATE(Kt\\s)\n");
	for(int i=0;i<mon_times.size();i++)
	{
		fprintf(fgenerator,"%3.3f\t%d\t%4.3f\n",mon_times[i],mon_distribution[i],mon_rates[i]);
	}

	fclose(fgenerator);

	#if defined (USE_ZMQ)
	close(socket);
	zmq_ctx_destroy (context);
	#endif
	return 0;
}



