/**
	Collector for Pane based Implementation
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <ff/buffer.hpp>
#include <iostream>
#include <unordered_map>
#include <sched.h>
#include <vector>
#include <algorithm> 

#include <ff/allocator.hpp>	
#include "../Includes/cycle.h"
#include "../Includes/iaperf.h"
#include "../Includes/general.h"
#include "../Includes/pane_cb.h"
#include "../Includes/cbpane.hpp"
#include "../Includes/pane_cb_reconf.h"
using namespace ff;
using namespace std;
#define PERCENTILE 0.95
/**
	Definition of the H function: it operates on the partial_results array and the result will be returned to the Worker
*/
HReturn_type H(GReturn_type *pane_results, int num_pr)
{
	//HReturn_type* ret;
	//possibile ottimizzazione: riciclo memoria allocata
	//posix_memalign((void **)&ret,CACHE_LINE_SIZE,CACHE_LINE_SIZE);
	//ret=(HReturn_type *)malloc(sizeof(HReturn_type));
	HReturn_type res=0.0;
	//it is very similar to the G but this is only for experimentation, it has not to be a meaningful computation
	for(int i = 0; i < num_pr; i++) 
	{
		//printf("H:Elemento su cui calcolo: %6.3f\n",pane_results[i]);
		for(int j = 0; j<H_GRAIN*10 ;j++) 
			res+=pow(pane_results[i],res);
			//res+=sin(pane_results[i]+res);
	} 
	return res;
}
/**
	The collector: it received the pane results (or partial results) produced by each Worker.
	A pane result, refer to a particular class (type) of the incoming task. The collector has to
	collect the pane results independently for each class and, for each class, compute the H function
	on the arrival of a new pane result. 

	In theory it keeps a count based sliding window of size window_size/pane_size, with slide 1, one per class. 
	In practice, these windows are implemented as simple circular buffers.

	The collector has to handle the situation in which results are received unordered from the workers.
	That is, for a particular class, the i-th pane result is received before (i-1)-th. In this situation it has
	to buffer it and compute the H function only when it received all the results.
*/
void * collector(void *args) {

	collector_data_t *data = (collector_data_t *) args;
	WorkerResult_t *rcvd=(WorkerResult_t*)malloc(sizeof(WorkerResult_t)); //variable for receiving from Workers
	HReturn_type res;
	int received_EOS=0; //number of workers from which the collector has received the EOS
	int index=0;
	double dlat=0.0; //latency in usec
	int64_t total_results=0; //the total number of H computation
	int64_t partial_results=0; // the number of H computations in the last print interval
	ticks last_print, start_ticks;
	char *suffix=data->suffix;
	long int freq=data->freq;
	ticks *start_global_ticks=data->start_global_ticks;
	ticks  print_rate=(unsigned long long)freq*1000000;
	ticks no_more_init=(unsigned long long)freq*NO_MORE_INIT; 

	int64_t *expected_iid; //expected internal id for the received results (it will be used for correctness)
	int *out_of_order; //number ofreceived out of order partial results for each class
	//FOR TESTING ONLY
	//ci salviamo tutte le latenze e le stampiamo sul file solo in caso di successo
	double latencies[500];
	int64_t lat_recv[500];
	double seconds[500];
	int num_latencies=0;
	//vector for computing response time percentile (maybe we can find something more efficient)
	vector<double> percentile_lat;
	double percentiles[800];
	pthread_barrier_t *barrier = data->barrier;
	SWSR_Ptr_Buffer **inqueue=data->inqueue;
	int num_workers=data->num_workers;
	int pane_size=data->pane_size;
	int num_classes=data->num_classes;
	int window_size=data->window_size;
	int window_slide=data->window_slide;
	int pane_per_slide=window_slide/pane_size;
	//we have to save the results and some other informations (for example the id of the last pane received for a particular class)
	int num_pr=window_size/pane_size; //number of pane results from wich is composed a window
	GReturn_type **pane_results=(GReturn_type **)malloc(sizeof(GReturn_type *)*num_classes);
	for(int i=0;i<num_classes;i++)
		pane_results[i]=(GReturn_type *)calloc(num_pr,sizeof(GReturn_type));
	//we have to save the actual insertion pointer for each classes
	int * ins_point_pr=(int *)calloc(num_classes,sizeof(int));
	//and the internal id of the last received partial result (for correctness)
	expected_iid=(int64_t *)malloc(num_classes*sizeof(int64_t));
	for(int i=0;i<num_classes;i++)
		expected_iid[i]=pane_size-1; //since the ids start from 0
	out_of_order=(int *)calloc(num_classes,sizeof(int));

	//vector that will contains the partial results received not in order 
	vector<WorkerResult_t> *buffer_disordered=(vector<WorkerResult_t> *)malloc(sizeof(vector<WorkerResult_t>)*num_classes);
	buffer_disordered=(vector<WorkerResult_t> *)malloc(sizeof(vector<WorkerResult_t>)*num_classes);
	for(int i=0;i<num_classes;i++)
		buffer_disordered[i]=*new vector<WorkerResult_t>(5); 
	
	#if defined(MONITORING)
		SWSR_Ptr_Buffer *cn_outqueue=data->cn_outqueue;
		//define the data structure for monitoring
		ticks start; //for computation times
		volatile ticks monitoring_timer;
		volatile ticks monitoring_step=MONITORING_STEP*1000*freq; //after how many cycles we have to monitor
		collector_monitoring_t *monitoring;
		char mon_ptr=0; //identifies the next monitoring structure in which write data
		posix_memalign((void **)&monitoring,CACHE_LINE_SIZE,NMONITORING*sizeof(collector_monitoring_t));
		for(int i=0;i<NMONITORING;i++)	
		{
			//init the internals
			monitoring[i].mon_id=i;
			monitoring[i].pane_results=0;
			posix_memalign((void **)&(monitoring[i].latency_per_class),CACHE_LINE_SIZE,num_classes*sizeof(double));
			memset(monitoring[i].latency_per_class,0,num_classes*sizeof(double));
			posix_memalign((void **)&(monitoring[i].pane_results_per_class),CACHE_LINE_SIZE,num_classes*sizeof(int));
			memset(monitoring[i].pane_results_per_class,0,num_classes*sizeof(int));
			posix_memalign((void**)&(monitoring[i].tcalch_per_class),CACHE_LINE_SIZE,num_classes*sizeof(double));
			memset(monitoring[i].tcalch_per_class,0,num_classes*sizeof(double));
		}
		#if defined(ADAPTIVITY)
			SWSR_Ptr_Buffer *cn_inqueue=data->cn_inqueue;
			//this particular monitoring data is used to notify that the reconfiguration is finished
			collector_monitoring_t reconf_finished;
			reconf_finished.tag=RECONF_FINISHED;
			reconf_data_c_t *reconf_data;
		#endif
	#endif
	
	//queues warmup
	for(int i=0;i<num_workers;i++)
		receive((void**)&rcvd,inqueue[i]);
	//synchronization barrier
	DEBUG(printf("Collector ready, running on %d\n",sched_getcpu()));
	pthread_barrier_wait(barrier);

	/**
		Start receiving from Workers
		NOTE: for the moment being workers are polled in round robin
	*/
	#if defined(MONITORING)
		monitoring_timer=getticks();
	#endif

	while(*start_global_ticks==0)
		REPEAT_25(asm volatile("PAUSE" ::: "memory");) //wait for the correct time

	last_print=getticks();
	start_ticks=last_print;
	
	printf("Collector, discrepancy: %Ld\n",getticks()-*start_global_ticks);
	while(received_EOS<num_workers) //the collector has to receive the EOS from all the workers
	{
		//non blocking receive
		if(inqueue[index]->pop((void **)&rcvd)) //we receive something from a worker
		{
			if(rcvd->isEOS) //it is an EOS
			{
				received_EOS++;
			}
			else
			{
				#if defined(MONITORING)
					asm volatile("":::"memory");
					start=getticks();
					asm volatile("":::"memory");
				#endif

				//DEBUG(printf("Collector, ricevuto risultato parziale per stream logico %d con id:%Ld, res: %6.3f \n",rcvd->type,rcvd->id,rcvd->res);)
				

				if((rcvd->id)>expected_iid[rcvd->type])
				{
					//we have received a partial result out of order
					//we have to buffer it and reuse when necessary
					//CURRENTLY IT IS IMPLEMENTED, BUT NOT INTENSIVELY TESTED 
					//insert it the buffer of disorder results
					buffer_disordered[rcvd->type].push_back(*rcvd);
					out_of_order[rcvd->type]++;
					//fprintf(stderr,"Received partial result out of order for class: %d expected: %Ld received: %Ld \n",rcvd->type,expected_iid[rcvd->type],rcvd->id);
				}
				else
				{
					//insert the result in the proper buffer
					pane_results[rcvd->type][ins_point_pr[rcvd->type]]=rcvd->res; //QUI in realta' andrebbe una memcpy perche' non sappiamo quanto sia grande il tipo G

					
					//compute H for the proper logical stream only if we have a window slide
					if((ins_point_pr[rcvd->type]+1)%pane_per_slide==0)
					{
						total_results++;
						partial_results++;
						res=H(pane_results[rcvd->type],num_pr);
						//IN REAL IMPLEMENTATIONS, the result will be sent afterward to the next module
						//...

						//compute latency statistics
						double lat=((double)(getticks()-rcvd->timestamp-*start_global_ticks))/freq;
						dlat+=lat;
						percentile_lat.push_back(lat);

						#if defined(MONITORING)
							asm volatile("":::"memory");
							monitoring[mon_ptr].tcalch_per_class[rcvd->type]+=(double)(getticks()-start)/freq; //usec
							asm volatile("":::"memory");
							monitoring[mon_ptr].pane_results_per_class[rcvd->type]++;
							monitoring[mon_ptr].pane_results++;
							monitoring[mon_ptr].latency_per_class[rcvd->type]+=lat;
						#endif
					}
					expected_iid[rcvd->type]+=pane_size;
					ins_point_pr[rcvd->type]=(ins_point_pr[rcvd->type]+1)%num_pr;
					

					

					//now check if previously we received disordered results: the idea is that
					//we will present in output all the results that should have been computed
					if(out_of_order[rcvd->type]>0) //we received out of order pane results
					{
						//brute force searching (maybe it can be optimized, but we will expect that only few elements arrive disordered)
						for(int i=0;i<buffer_disordered[rcvd->type].size();i++)
						{
							if(buffer_disordered[rcvd->type][i].id==expected_iid[rcvd->type])
							{
								//ok it is the right one
								#if defined(MONITORING)
									asm volatile("":::"memory");
									start=getticks();
									asm volatile("":::"memory");
								#endif
								pane_results[rcvd->type][ins_point_pr[rcvd->type]]=buffer_disordered[rcvd->type][i].res;
								//compute H for the proper logical stream only if we have a window slide
								if((ins_point_pr[rcvd->type]+1)%pane_per_slide==0)
								{
									total_results++;
									partial_results++;
									res=H(pane_results[rcvd->type],num_pr);
									double lat=((double)(getticks()-buffer_disordered[rcvd->type][i].timestamp-*start_global_ticks))/freq;
									dlat+=lat;
									#if defined(MONITORING)
										asm volatile("":::"memory");
										monitoring[mon_ptr].tcalch_per_class[rcvd->type]+=(double)(getticks()-start)/freq; //usec
										asm volatile("":::"memory");
										monitoring[mon_ptr].pane_results_per_class[rcvd->type]++;
										monitoring[mon_ptr].pane_results++;
										monitoring[mon_ptr].latency_per_class[rcvd->type]+=lat;
									#endif
								}
								ins_point_pr[rcvd->type]=(ins_point_pr[rcvd->type]+1)%num_pr;
								expected_iid[rcvd->type]+=pane_size;
								
																	
								//fprintf(stderr,"Received partial result out of order for class: %d, problem resolved!\n",rcvd->type);
								
								buffer_disordered[rcvd->type].erase(buffer_disordered[rcvd->type].begin()+i);
								out_of_order[rcvd->type]--;
								//reset i to zero, we have to start from the beginning
								i=0;
							}
						}
					}												
				}		
			}
		}
		
		//check if we have to print
		if(getticks()-last_print>print_rate)
		{

			double avglat=dlat/partial_results;
			int index;
			//order percentiles
			if(percentile_lat.size()>0)
			{
				sort(percentile_lat.begin(), percentile_lat.end()); 
				index=percentile_lat.size()*PERCENTILE;
				printf("Time: %6.3f, recvd results: %Ld, avg latency (usec): %6.3f  %d-percentile: %6.3f (top: %6.3f)\n",((double)(getticks()-start_ticks)/(freq*1000000)),partial_results,avglat,(int)(PERCENTILE*100),percentile_lat[index],percentile_lat[percentile_lat.size()-1]);
			}
			else
			printf("Time: %6.3f, recvd results: %Ld, avg latency (usec): %6.3f\n",((double)(getticks()-start_ticks)/(freq*1000000)),partial_results,avglat);
			if(getticks()-start_ticks>no_more_init)
			{
				//check latency for termination if bottleneck
				if(avglat>MAX_LAT)
				{	//non basta dobbiamo mettere qualcos'altro
					//si tengono le ultime 5 e si fa la media da 5 a 5?
					fprintf(stderr,"Collector found that the latency is too high\n");
					//print to file the latencies
					char out_file[100];
					sprintf(out_file,"%s_%s.dat",LAT_FILE,suffix);
					FILE* flat=fopen(out_file,"w");
					fprintf(flat,"#Second Num_results Latency\n");
					for(int i=0;i<num_latencies;i++)
					{
						fprintf(flat,"%3.3f\t%Ld\t%6.3f\n",seconds[i],lat_recv[i],latencies[i]);
					}
					fclose(flat);
					exit(BOTTLENECK_ERR);
				}
			}
			last_print=getticks();
			latencies[num_latencies]=avglat;
			lat_recv[num_latencies]=partial_results;
			seconds[num_latencies]=((double)(getticks()-start_ticks)/(freq*1000000));
			if(percentile_lat.size()>0)
				percentiles[num_latencies]=percentile_lat[index];
			else
				percentiles[num_latencies]=0.0;
			num_latencies++;
			partial_results=0;
			dlat=0.0;
			percentile_lat.clear();

		}
		#if defined(MONITORING)
			//chek if there are messages from controller
			#if defined(ADAPTIVITY)
				if(cn_inqueue->pop((void **)(&reconf_data)))
				{
					//if there are newly spanned threads, copy their queues
					printf("Collector: ricevuto messaggio, incremento par degree di %d worker\n",reconf_data->par_degree_changes);
					//take the new queues
					for(int i=0;i<reconf_data->par_degree_changes;i++)
						inqueue[num_workers+i]=reconf_data->wqueues[i];
					//increments the par degree
					num_workers+=reconf_data->par_degree_changes;
					//ok, notify to the collector that the reconfiguration started (we don't need a specific message, since it is a well defined req-reply protocol)
					bsend((void*)&reconf_finished,cn_outqueue);
					printf("Collector: finito\n");
				}
			#endif

			//check if we can send monitoring data to controller
			if(getticks()-monitoring_timer>monitoring_step)
			{ 	
				//EVENTUALMENTE POTREMMO MANDARGLI LA LATENZA TOTALE
				asm volatile("":::"memory");
				// printf("Monitoring data %d are valid. Elements received: %Ld\n",mon_ptr,monitoring[mon_ptr].elements);

				//compute average latencies
				for(int i=0;i<num_classes;i++)
					monitoring[mon_ptr].latency_per_class[i]=(monitoring[mon_ptr].latency_per_class[i]/monitoring[mon_ptr].pane_results_per_class[i]);
				//blocking send towards the collector
				bsend(&monitoring[mon_ptr],cn_outqueue);
				//printf("Collector sent step: %d, lqueue:%d\n",mon_step++,cn_outqueue->length());
				monitoring_timer=getticks();
				mon_ptr=(mon_ptr+1)%NMONITORING;
				//reset the monitoring data
				monitoring[mon_ptr].pane_results=0;
				memset(monitoring[mon_ptr].latency_per_class,0,num_classes*sizeof(double));
				memset(monitoring[mon_ptr].pane_results_per_class,0,num_classes*sizeof(int));
				memset(monitoring[mon_ptr].tcalch_per_class,0,num_classes*sizeof(double));

			}
		#endif
		

		index=(index+1)%num_workers;
	}
	printf("Collettore ricevuto EOS\n");
	//last print
	double avglat=dlat/partial_results;
	//order percentile
	sort (percentile_lat.begin(), percentile_lat.end()); 
	int idx=percentile_lat.size()*PERCENTILE;
	printf("Time: %6.3f, recvd results: %Ld, avg latency (usec): %6.3f %d-percentile: %6.3f\n",((double)(getticks()-start_ticks)/(freq*1000000)),partial_results,avglat,(int)(PERCENTILE*100),percentile_lat[idx]);
	
	last_print=getticks();
	latencies[num_latencies]=avglat;
	lat_recv[num_latencies]=partial_results;
	percentiles[num_latencies]=percentile_lat[idx];
	seconds[num_latencies]=((double)(getticks()-start_ticks)/(freq*1000000));
	num_latencies++;
	#if defined(MONITORING)
	//send the last monitoring, just for terminating the controller
	bsend(&monitoring[mon_ptr],cn_outqueue);
	#endif
	//print to file the latencies
	char out_file[100];
	sprintf(out_file,"%s_%s.dat",LAT_FILE,suffix);
	FILE* flat=fopen(out_file,"w");
	fprintf(flat,"#Second Num_results Latency Percentile\n");
	for(int i=0;i<num_latencies;i++)
	{
		fprintf(flat,"%3.3f\t%Ld\t%6.3f\t%6.3f\n",seconds[i],lat_recv[i],latencies[i],percentiles[i]);
	}
	fclose(flat);
	
	/*#ifdef MONITOR_COMP_TIME
	//print to file the computation times


	char out_file2[100];
	sprintf(out_file2,"%s_%s.dat",CLASS_FILE,suffix);
	FILE* fclass=fopen(out_file2,"w");
	fprintf(fclass,"#Number of classes: %d\n#",num_classes);
	fprintf(fclass, "#CLASS_ID FREQ GCOMP_TIME HCOMP_TIME\n");
	for(int i=0;i<num_classes;i++)
	{
		//compute the average computation time for function G
		double gctime=0;
		int64_t gcnumber=0;
		for(int j=0;j<num_workers;j++)
		{
			gctime+=gcomputation_times[j][i];
			gcnumber+=gcomputation_number[j][i];
		}

		fprintf(fclass,"%d\t%Ld\t%6.3f\t%6.3f\n",i,classes_freq[i],gctime/gcnumber,((double)hcomputation_times[i])/class_rcvd[i]/FREQ);
	}
	fclose(fclass);

	#endif*/
	DEBUG(printf("Collector terminated, received results: %Ld\n",total_results));
	//pthread_barrier_wait(barrier);

	return NULL;
}

