/**
	Controller for Functional Partitioning implementation

	NOTA: per ora l'implementazione dei meccanismi di riconfigurazione non e' best effort. Bisognerebbe finire di considerare la nuova scheduling
	table quando l'emettitore invia il nuovo monitoraggio
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <ff/buffer.hpp>
#include <ff/allocator.hpp>
#include <iostream>
#include <unordered_map>
#include <vector>
#include <sched.h>
#include "../Includes/RandomGenerator.h"
#include "../Includes/cycle.h"
#include "../Includes/iaperf.h"
#include "../Includes/general.h"
#include "../Includes/pane_cb.h"
#include "../Includes/pane_cb_reconf.h"

using namespace ff;
using namespace std;
/**
	Receive on FF queues. If there is more than one element, returns the last inserted
*/
inline void receiveLast ( void **t, ff::SWSR_Ptr_Buffer *inqueue) __attribute__((always_inline))/* __attribute__((noinline))*/; 
inline void receiveLast ( void **t, ff::SWSR_Ptr_Buffer *inqueue)
{
	do
	{
		while (!inqueue->pop( t)) {
			// Small sleep
			REPEAT_25(asm volatile("PAUSE" ::: "memory");)
		}
	}while(!inqueue->empty());
}

void *controller (void *args)
{
	//init
	controller_data_t *data = (controller_data_t *) args;
	int num_workers=data->num_workers;
	int num_classes=data->num_classes;
	SWSR_Ptr_Buffer *e_inqueue=data->e_inqueue;
	SWSR_Ptr_Buffer **w_inqueue=data->w_inqueue;
	SWSR_Ptr_Buffer *c_inqueue=data->c_inqueue;
	SWSR_Ptr_Buffer *e_outqueue=data->e_outqueue;
	SWSR_Ptr_Buffer *c_outqueue=data->c_outqueue;
	char *suffix=data->suffix;
	long int freq=data->freq;

		int max_workers=data->max_workers;
	int pane_size=data->pane_size;
	int *affinities=data->affinities;
	#if !defined(DONT_USE_FFALLOC) && !defined(TASK_BUFF)
		ff_allocator *ffalloc=data->ffalloc;
	#endif
	cpu_set_t cpuset;
	emitter_monitoring_t *em;
	worker_monitoring_t **wm=(worker_monitoring_t**)malloc(sizeof(worker_monitoring_t*)*num_workers);
	collector_monitoring_t *cm; 
	int64_t monitoring_step=0;
	//reconf_repository_t *repo=(reconf_repository_t *)data->repo;
	DEBUG(printf("Controller ready, running on: %d\n",sched_getcpu()))
	//The controller will be stopped by the emitter
	bool stop=false;
	//NOTA: da capire come gestire questi messaggi di configurazione (quanto ci basta le assunzioni sul ciclo chiuso tra emettitore  e controllore)

	reconf_data_em_t *reconf_data_em;
	posix_memalign((void **)&reconf_data_em,CACHE_LINE_SIZE,sizeof(reconf_data_em_t));
	posix_memalign((void **)&reconf_data_em->scheduling_table,CACHE_LINE_SIZE,sizeof(char)*num_classes);
	reconf_data_em->par_degree_changes=0;
	reconf_data_c_t *reconf_data_c;
	posix_memalign((void **)&reconf_data_c,CACHE_LINE_SIZE,sizeof(reconf_data_c_t));
	reconf_data_c->par_degree_changes=0;
	worker_data_t *worker_data; //for thread spawning
	pthread_t *wtids=(pthread_t *)malloc(sizeof(pthread_t)*(max_workers-num_workers));
	int threads_spawned=0; //number of thread spawned (that have to be checked on)
	//TMP: per il momento facciamo una serie di stampe su file
	int *classes_freq=(int *)calloc(num_classes,sizeof(int));
	double *gcomp_time=(double *)calloc(num_classes,sizeof(double));
	double *hcomp_time=(double *)calloc(num_classes,sizeof(double));
	int *results=(int *)calloc(num_classes,sizeof(int));
	int attempt;


	//prendiamo il tempo, utile per alcuni test
	struct timeval tmp_t;
	
	long int start_t;
	long int reconf_start_t;
	bool mess_sent_degree=false;
	bool mess_sent_scheduling=true;
	bool start_time_taken=false;
	while(!stop)
	{
		//PER IL MOMENTO FACCIAMO UNA LETTURA ROUND ROBIN
		//receive from emitter
		//NOTA PER ORA SONO BLOCCANTI: FORSE DOVREMO POI RENDENDERLE NON BLOCCANTI
		
		receiveLast((void **)&em,e_inqueue);
		if(!start_time_taken)
		{
			start_time_taken=true;
			gettimeofday(&tmp_t,NULL);
			start_t=tmp_t.tv_sec*1000+(tmp_t.tv_usec)/1000;
		}
		//printf("Contr: inizio ciclo %Ld\n",monitoring_step);
		if(em->stop)
		{
			stop=true;
		}
		else //facciamo attenzione (non vorrei che ci fosse un deadlock)
		{
			//receive form workers
			for(int i=0;i<num_workers;i++)
			{ 
				receiveLast((void **)&wm[i],w_inqueue[i]);	
			}
		
			receiveLast((void **)&cm,c_inqueue);
			//printf("Ricevuto dal collector\n");
			//TMP: contiamo le varie frequenze
			//NOTA QUESTO CONTEGGIO PUO' non essere preciso perche' saltiamo qualche messaggio
			for(int i=0;i<num_classes;i++)
			{
				classes_freq[i]+=em->elements_per_class[i];
				
				for(int j=0;j<num_workers;j++)
				{
					gcomp_time[i]+=wm[j]->tcalcg_per_class[i];
					results[i]+=wm[j]->computations_per_class[i]; //prendiamo questo come risultato casomai dal collettore si sfasa
					
				}				
				hcomp_time[i]=cm->tcalch_per_class[i]/cm->pane_results_per_class[i];
			}
			gettimeofday(&tmp_t,NULL);
			long int end_t=tmp_t.tv_sec*1000+(tmp_t.tv_usec)/1000-start_t;
			#if defined (ADAPTIVITY)
			//CASO INCREMENTO GRADO DI PARALLELISMO
			//note for the moment is currently handled only the increase of par degree
			if(!mess_sent_degree && end_t>66000)
			{
				mess_sent_degree=true;
				memcpy(reconf_data_em->scheduling_table,em->scheduling_table,sizeof(char)*num_classes);
				gettimeofday(&tmp_t,NULL);
				reconf_start_t=tmp_t.tv_sec*1000000+(tmp_t.tv_usec);
				int changes=1;
				printf("Controller spawning %d workers\n",changes);
				if(changes>0)
				{
					reconf_data_em->wqueues=(SWSR_Ptr_Buffer **)malloc(sizeof(SWSR_Ptr_Buffer *)*changes);
					reconf_data_c->wqueues=(SWSR_Ptr_Buffer **)malloc(sizeof(SWSR_Ptr_Buffer *)*changes);
					worker_data=(worker_data_t *)malloc(sizeof(worker_data_t)*changes);
					reconf_data_em->par_degree_changes=changes;
					reconf_data_c->par_degree_changes=changes;
					for(int i=0;i<changes;i++)	
					{
						reconf_data_em->wqueues[i]=new SWSR_Ptr_Buffer(QUEUE_SIZE);
						reconf_data_em->wqueues[i]->init();
						reconf_data_c->wqueues[i]=new SWSR_Ptr_Buffer(QUEUE_SIZE);
						reconf_data_c->wqueues[i]->init();
						//assuming that there is enough space, define queue worker->controller
						w_inqueue[num_workers+i]=new SWSR_Ptr_Buffer(QUEUE_SIZE_MON);
						w_inqueue[num_workers+i]->init();
					}
					//thread spawning
					for(int i=0;i<changes;i++)
					{
						pthread_t tid;
						worker_data[i].workerId = i+num_workers;
						worker_data[i].inqueue = reconf_data_em->wqueues[i];
						worker_data[i].outqueue = reconf_data_c->wqueues[i];
						worker_data[i].barrier=NULL; //in this way, newly spawned threads will not perform wait on barrier 
						worker_data[i].pane_size=pane_size;
						worker_data[i].freq=freq;
						worker_data[i].num_classes=num_classes;
						#if !defined(DONT_USE_FFALLOC) && !defined(TASK_BUFF)
							worker_data[i].ffalloc=ffalloc;
						#endif
						worker_data[i].cn_outqueue=w_inqueue[num_workers+i];
						
						pthread_create(&tid, NULL, worker, &(worker_data[i]));
						//set CPU affinity of worker threads
						CPU_ZERO(&cpuset);
						CPU_SET(affinities[i+num_workers], &cpuset);

						if (pthread_setaffinity_np(tid, sizeof(cpu_set_t), &cpuset)) {
							cerr << "Cannot set thread to CPU " << i << endl;
						}
						//worker_data[i].num_workers = num_workers;
						wtids[i] = tid;

					}
					threads_spawned+=changes;
					//communication toward the emitter
					//per il momento facciamo una scheduling table statica bilanciata sui worker
					//spostando il minimo
					//ci calcoliamo il numero di classi spettante ai nuovi worker
					int num_new_class=num_classes/(num_workers+changes);
					printf("Controller: ai nuovi worker spettano %d classi\n",num_new_class);
					int c=num_classes-1;
					for(int i=0;i<changes;i++)
					{
						for(int j=0;j<num_new_class;j++)
						{
							reconf_data_em->scheduling_table[c--]=num_workers+i+1; //already take into account +1
						}
					
					}

					num_workers+=changes;
					//send the data toward the emitter
					bsend(reconf_data_em,e_outqueue);

					//communication toward the collector
					bsend(reconf_data_c,c_outqueue);
					//wait for the reply from the emitter
					//ATTENZIONE QUESTA INTERAZIONE E' UN PO' TRABALLANTE
					do
					{
						receive((void **)&em,e_inqueue);
						// printf("Controller received: %p, with tag:%d\n",em,em->tag);
						//fprintf(stderr,"Reconfiguration error\n");
						//exit(RECONF_ERR);
					}while((em->tag!=RECONF_FINISHED));
					// printf("Controller: emitter ha finito\n");
					do
					{
						receive((void **)&cm,c_inqueue);
						// printf("Controller received: %p, with tag:%d\n",em,em->tag);
						//fprintf(stderr,"Reconfiguration error\n");
						//exit(RECONF_ERR);
					}while((cm->tag!=RECONF_FINISHED));
					gettimeofday(&tmp_t,NULL);
					printf("Controller: reconfiguration finished in %Ld usecs\n",tmp_t.tv_sec*1000000+(tmp_t.tv_usec)-reconf_start_t);
				}

			}
			if(!mess_sent_scheduling && end_t>63000)
			{
				mess_sent_scheduling=true;
				memcpy(reconf_data_em->scheduling_table,em->scheduling_table,sizeof(char)*num_classes);
				// RICONFIGURAZIONE PER SWITCH 8 WORKER
				int count=0;
				int indx;
				//se vogliamo farlo con 9 (in seguito a changes), count vale al piu' tre
				for(indx=1;indx<num_classes && count<3;indx++)
					if(reconf_data_em->scheduling_table[indx]==1)
						count++;
				//ora spostiamo le successive
				int wk=0;

				for(;indx<num_classes;indx++)
					if(reconf_data_em->scheduling_table[indx]==1)
					{
						reconf_data_em->scheduling_table[indx]=wk+2; //+1 per la scheduling table e +1 perche' non consideriamo il primo
						wk=(wk+1)%(num_workers-1);
					}
				
				//RICONFIGURAZIONE IN CUI SPOSTIAMO TUTTE LE CLASSI
				/*char w=1;
				for(int i=0;i<num_classes;i++)
				{
					reconf_data->scheduling_table[i]=w+1;
					w++;
					w%=num_workers;
				}*/
				//for(int i=0;i<num_classes;i++)
				//	printf("Class %d worker: %d\n",i,reconf_data->scheduling_table[i]-1);
			

				printf("Controller invio nuova scheduling table\n");
				gettimeofday(&tmp_t,NULL);
				reconf_start_t=tmp_t.tv_sec*1000000+(tmp_t.tv_usec);
				bsend(reconf_data_em,e_outqueue);
				

				//wait for the reply from the emitter
				//ATTENZIONE QUESTA INTERAZIONE E' UN PO' TRABALLANTE
				//PER ora, di fatto il controllore aspetta che arrivi la conferma dall'emitter di avvenuta riconfigurazione
				do
				{
					receive((void **)&em,e_inqueue);
					//fprintf(stderr,"Reconfiguration error\n");
					//exit(RECONF_ERR);
				}while(em->tag!=RECONF_FINISHED);
				
				gettimeofday(&tmp_t,NULL);
				printf("Controller: reconfiguration finished in %Ld usecs\n",tmp_t.tv_sec*1000000+(tmp_t.tv_usec)-reconf_start_t);
			}
			#endif



		}
		monitoring_step++;		
	}
	// Wait for the completion of the threads
	void *retval ;
	for (int i = 0; i < threads_spawned; i++) {
		pthread_join(wtids[i], &retval);
	}
	//print to file classes statistics
	char out_file[100];
	sprintf(out_file,"m%s_%s.dat",CLASS_FILE,suffix);
	FILE* fclass=fopen(out_file,"w");
	fprintf(fclass,"#Number of classes: %d\n#",num_classes);
	fprintf(fclass, "#CLASS_ID FREQ RESULTS GCOMP_TIME HCOMP_TIME\n");
	
	 for(int i=0;i<num_classes;i++)
		fprintf(fclass,"%d\t%d\t%d\t%6.3f\t%6.3f\n",i,classes_freq[i],results[i],gcomp_time[i]/results[i],hcomp_time[i]);
	 
	fclose(fclass);
	return NULL;
}
