/**
	Emitter for Pane Based implementation
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <ff/buffer.hpp>
#include <iostream>
#include <unordered_map>
#include <vector>
#include <sched.h>
#include <ff/allocator.hpp>	
#include "../Includes/cycle.h"
#include "../Includes/iaperf.h"
#include "../Includes/general.h"
#include "../Includes/pane_cb.h"
#include "../Includes/cbpane.hpp"
#include "../Includes/pane_cb_reconf.h"

using namespace ff;
using namespace std;

//scheduling table
char *scheduling_table;
//next assigned worker scheduling RR
char next_schedulingRR=0;
int64_t *classes_freq;

#if defined(LESS_ASSIGNED) || defined (EMPTIEST_QUEUE)
//number of classes currently assigned to the various workers
int *currently_assigned;
#else
#ifdef LESS_LOADED
//the classes assigned to each worker (currently_assigned[i][j]==1 if the class j-th is assigned to worker i; 0 otherwise)
bool **currently_assigned;
//approximately, the load of each worker (computed as the sum of the classes frequencies' asssigned to it)
double *load;
#endif
#endif
inline char schedulingLessAssigned (task_t *t, int num_workers, int pane_size) __attribute__((always_inline));
inline char schedulingLessLoaded(task_t *t, int num_workers, int num_classes, int pane_size) __attribute__((always_inline));
inline char schedulingEmptiestQueue(task_t *t, int num_workers, int num_classes, int pane_size, SWSR_Ptr_Buffer **outqueue) __attribute__((always_inline));
inline char schedulingAdaptive(task_t *t,int num_workers,int pane_size, char *new_scheduling_table) __attribute__((always_inline));
inline char schedulingRR (task_t *t, int num_workers, int pane_size) __attribute__((always_inline));

//number of elements of non-closed-panes, that are currently live in the program
int *elements_ncp;
int64_t msg;
int yet_to_be_reconfigured=0; //number of classes that have not been already allocated to the new worker

/**
	This was used if we have ADAPTIVITY (it will use the scheduling table passed by the controller)
*/
inline char schedulingAdaptive(task_t *t,int num_workers,int pane_size, char *new_scheduling_table)
{
	char ret=scheduling_table[t->type]-1;
	//scheduling table is never empty
	elements_ncp[t->type]++;
	//if the pane is closed, reset the association <class,worker> by using the one contained in passed scheduling table
	if(elements_ncp[t->type]==pane_size)
	{
		elements_ncp[t->type]=0;
		if(new_scheduling_table) //we are in a reconfiguration phase
		{
			//if(scheduling_table[t->type]!=new_scheduling_table[t->type])
			//printf("Class %d passed from worker %d to %d. Pane size:%d Actual task id: %Ld\n",t->type,scheduling_table[t->type],new_scheduling_table[t->type],pane_size,t->internal_id);
			
			scheduling_table[t->type]=new_scheduling_table[t->type];
			yet_to_be_reconfigured--;

		}	
		
		//DEBUG(printf("***Pane closed for logical stream: %d\n",t->type));
	}
	return ret;
}

//Every time that a new pane is opened, assign it to a worker following a RR
//strategy
inline char schedulingRR (task_t *t, int num_workers, int pane_size) 
{


	char ret;
	//The scheduling table is a simple array with numb_classes positions
	//if the i-th element is zero then for that logical stream we don't have 
	//an assigned Worker. Otherwise it contains the ID_Worker+1 to which send
	//the tasks of that class
	if(scheduling_table[t->type]==0)
	{
		//find the worker that has the minimum number of assigned classes

		scheduling_table[t->type]=next_schedulingRR	+1;
		next_schedulingRR=(next_schedulingRR+1)%num_workers;
		elements_ncp[t->type]++;
		ret= scheduling_table[t->type]-1;
		// DEBUG(printf("***New logical stream %d, assigned to Worker: %d\n",t->type,ret));

	}
	else
	{
		ret=scheduling_table[t->type]-1;
		//counts the element in the pane for that class
		elements_ncp[t->type]++;
		//if the pane is closed, reset the association <class,worker>
		if(elements_ncp[t->type]==pane_size)
		{
			scheduling_table[t->type]=0;
			elements_ncp[t->type]=0;
			//DEBUG(printf("***Pane closed for logical stream: %d\n",t->type));
		}
	}

	return ret;

}
#ifdef LESS_ASSIGNED

/**
	Determines the id of the Worker to which send the task
	according to the number of currently assigned classes.

	This function also count the number of elements, for each class, that are currently scheduled in the program.
	If the pane is completed, it resets the association <class,worker>

	It may works if the computation time is equal for all the classes
*/

inline char schedulingLessAssigned (task_t *t, int num_workers, int pane_size) 
{
	char ret;
	//The scheduling table is a simple array with numb_classes positions
	//if the i-th element is zero then for that logical stream we don't have 
	//an assigned Worker. Otherwise it contains the ID_Worker+1 to which send
	//the tasks of that class
	if(scheduling_table[t->type]==0)
	{
		//find the worker that has the minimum number of assigned classes
		char wmin=0;
		for(int i=1;i<num_workers;i++)
		{
			if(currently_assigned[i]<currently_assigned[wmin])
				wmin=i;
		}
		scheduling_table[t->type]=wmin+1;
		currently_assigned[wmin]++;
		elements_ncp[t->type]++;
		//DEBUG(printf("***New logical stream %d, assigned to Worker: %d\n",t->type,wmin));
		ret=wmin;
	}
	else
	{
		ret=scheduling_table[t->type]-1;
		//counts the element in the pane for that class
		elements_ncp[t->type]++;
		//if the pane is closed, reset the association <class,worker>
		if(elements_ncp[t->type]==pane_size)
		{
			scheduling_table[t->type]=0;
			elements_ncp[t->type]=0;
			currently_assigned[ret]--;
			//DEBUG(printf("***Pane closed for logical stream: %d\n",t->type));
		}
	}

	return ret;

}

#else
#ifdef LESS_LOADED
/**
	The task is sent to the worker that is currently less loaded.

	The current load of a worker is defined as the summation of the frequencies of the classes
	that are currently assigned on it
*/
// PER ORA RIMANE UN MENTODO NIUBBO CHE CONSIDERA LE FREQUNEZE DALL'INIZIO DELL'ESECUZIONE E NON MANO MANO
//static char wmin=0;
inline char schedulingLessLoaded(task_t *t, int num_workers, int num_classes, int pane_size)
{

	char ret;
	//The scheduling table is a simple array with numb_classes positions
	//if the i-th element is zero then for that logical stream we don't have 
	//an assigned Worker. Otherwise it contains the ID_Worker+1 to which send
	//the tasks of that class

	char wmin=0;
	if(scheduling_table[t->type]==0)
	{
		for(int j=0;j<num_workers;j++)
		{
			load[j]=0;
			for(int i=0;i<num_classes;i++)
				load[j]+=(currently_assigned[j][i]*(double)classes_freq[i])/msg;

			//load[j]/=msg;
			//find the worker that has the minimum load
			if(load[j]<load[wmin])
				wmin=j;
		}

		//find the worker that has the minimum load
		
		/*for(int i=0;i<num_workers;i++)
		{
			if(load[i]<load[wmin])
				wmin=i;
		}*/
		scheduling_table[t->type]=wmin+1;
		//the choosen worker has now a new associated class
		currently_assigned[wmin][t->type]=true;
		//recompute the loads
		//load[wmin]=0;
		//for(int i=0;i<num_classes;i++)
		///	load[wmin]+=(currently_assigned[wmin][i]*(double)classes_freq[i])/msg;
		
		elements_ncp[t->type]++;
		//DEBUG(printf("***New logical stream %d, assigned to Worker: %d, new load for the worker: %6.3f\n",t->type,wmin,load[wmin]));
		ret=wmin;
		// wmin=(wmin +1)%num_workers; //barbatrucco per non assegnare all'inizio tutto ad un worker
	}
	else
	{
		ret=scheduling_table[t->type]-1;
		//counts the element in the pane for that class
		elements_ncp[t->type]++;
		//if the pane is closed, reset the association <class,worker>
		if(elements_ncp[t->type]==pane_size)
		{
			scheduling_table[t->type]=0;
			elements_ncp[t->type]=0;
			currently_assigned[ret][t->type]=false;
			//recompute its load
			//load[ret]=0;
			//for(int i=0;i<num_classes;i++)
		//		load[ret]+=(currently_assigned[ret][i]*classes_freq[i])/msg;
			//DEBUG(printf("***Pane closed for logical stream: %d\n",t->type));
		}
	}

	return ret;

}
#else
#if defined(EMPTIEST_QUEUE)
static char wmin=0;
inline char schedulingEmptiestQueue(task_t *t, int num_workers, int num_classes, int pane_size, SWSR_Ptr_Buffer **outqueue)
{
	char ret;
	//The scheduling table is a simple array with numb_classes positions
	//if the i-th element is zero then for that logical stream we don't have 
	//an assigned Worker. Otherwise it contains the ID_Worker+1 to which send
	//the tasks of that class
	if(scheduling_table[t->type]==0)
	{
		//find the worker that has the minimum queue's length
		int min=outqueue[wmin]->length();
		for(int i=0;i<num_workers;i++)
		{
			if(outqueue[i]->length()<min)
			{
				wmin=i;
				min=outqueue[i]->length();
			}
		}
		scheduling_table[t->type]=wmin+1;
		elements_ncp[t->type]++;
		//DEBUG(printf("***New logical stream %d, assigned to Worker: %d\n",t->type,wmin));
		ret=wmin;
		wmin=(wmin +1)%num_workers; //barbatrucco per non assegnare all'inizio tutto ad un worker
	}
	else
	{
		ret=scheduling_table[t->type]-1;
		//counts the element in the pane for that class
		elements_ncp[t->type]++;
		//if the pane is closed, reset the association <class,worker>
		if(elements_ncp[t->type]==pane_size)
		{
			scheduling_table[t->type]=0;
			elements_ncp[t->type]=0;
			//DEBUG(printf("***Pane closed for logical stream: %d\n",t->type));
		}
	}

	return ret;

}
#endif
#endif
#endif

/*********************************






	ATTENZIONE: fare attenzione alle varie define e relative funzioni di scheduling





*////////////
void *emitter(void *args)
{
	struct timeval tmp_t;
	long start_t=0, end_t=0;
	task_t tmp;
	char to_send_to;
	int ret;
	 msg=0;
	//Init
	emitter_data_t *data = (emitter_data_t *) args;
	pthread_barrier_t *barrier = data->barrier;
	int num_workers=data->num_workers;
	int port=data->port;
	int pane_size=data->pane_size;
	SWSR_Ptr_Buffer **outqueue=data->outqueue;
	int num_classes=data->num_classes;
	int freq=data->freq;
	ticks *start_global_ticks=data->start_global_ticks;
	scheduling_table=(char *)calloc(num_classes,sizeof(char));
	//frequency counter for the various classes
	classes_freq=(int64_t*)calloc(num_classes,sizeof(int64_t));
	#if defined(MONITORING)
		SWSR_Ptr_Buffer *cn_outqueue=data->cn_outqueue;
		SWSR_Ptr_Buffer *cn_inqueue=data->cn_inqueue;
	#endif
	char *new_scheduling_table=NULL;
	
	#if defined (ROUND_ROBIN)
		printf("**********SCHEDULING: ROUND_ROBIN!!!**********\n");
	#endif
	#if defined (LESS_ASSIGNED)
		printf("**********SCHEDULING: LESS_ASSIGNED!!!**********\n");
	#endif
	#if defined (LESS_LOADED)
		printf("**********SCHEDULING: LESS_LOADED!!!**********\n");
	#endif

	#if defined (ROUND_ROBIN)
		elements_ncp=(int *)calloc(num_classes,sizeof(int));
	#endif
	#if defined(LESS_ASSIGNED) 
		//number of elements of non-closed-panes, that are currently live in the program
		elements_ncp=(int *)calloc(num_classes,sizeof(int));
		currently_assigned=(int *)calloc(num_workers,sizeof(int));
	#else
		#ifdef LESS_LOADED
			elements_ncp=(int *)calloc(num_classes,sizeof(int));
			currently_assigned=(bool **)malloc(num_workers*sizeof(bool *));
                for(int i=0; i< num_workers;i++)
                       	currently_assigned[i]=(bool *)calloc(num_classes,sizeof(bool));

               	load=(double *)calloc(num_workers,sizeof(double));

		#endif
	#endif
	
	#if !defined (ROUND_ROBIN)
	printf("---> ATTENTION: Emitter: static scheduling <---\n");
	char w=0;
	for(int i=0;i<num_classes;i++)
	{
		scheduling_table[i]=w+1;
		w++;
		w%=num_workers;
	}
	#endif

	#if defined(TASK_BUFF)
		//create the buffer for tasks
		task_t * task_buff;
		//NOTA: non ho ancora capito quanto debba essere la dimensione di questo buffer.
		//Per ora lasciamo cosi' in caso non torni il numero di risultati si deve guardare
		int buff_size=num_workers*(QUEUE_SIZE+2)*num_workers;
		posix_memalign((void **)&task_buff,CACHE_LINE_SIZE, buff_size*sizeof(task_t));
		int bi=0; //index pointer for the task buffer
	#else
		#if !defined(DONT_USE_FFALLOC)
			ff_allocator *ffalloc=data->ffalloc;
		#endif
	#endif
	task_t *tb; //receiving task
	
	#if defined(MONITORING)
		//define the data structure for monitoring
		volatile ticks monitoring_timer;
		volatile ticks monitoring_step=MONITORING_STEP*1000*freq; //after how many cycles we have to monitor
		//reconfiguration message that will be received from the controller
		reconf_data_em_t *reconf_data;
		emitter_monitoring_t *monitoring;
		char mon_ptr=0; //identifies the next monitoring structure in which write data
		posix_memalign((void **)&monitoring,CACHE_LINE_SIZE,NMONITORING*sizeof(emitter_monitoring_t));
		for(int i=0;i<NMONITORING;i++)	
		{
			//init the internals
			monitoring[i].mon_id=i;
			monitoring[i].tag=MON_TAG;
			monitoring[i].elements=0;
			monitoring[i].stop=false;
			monitoring[i].scheduling_table=NULL;
			posix_memalign((void **)&(monitoring[i].elements_per_class),CACHE_LINE_SIZE,num_classes*sizeof(int));
			memset(monitoring[i].elements_per_class,0,num_classes*sizeof(int));
		}
		//reconf_repository_t *repo=(reconf_repository_t *)data->repo;
		#if defined(ADAPTIVITY)
			//used for notification to controller
			emitter_monitoring_t reconf_finished;
			reconf_finished.tag=RECONF_FINISHED;
		#endif

	#endif
	//queues warmup
	for(int i=0;i<num_workers;i++)
		send((void *)1,outqueue[i]);
	DEBUG(printf("Emitter ready, running on: %d\n",sched_getcpu()))
	//synchronization barrier	
	pthread_barrier_wait(barrier);
	//accept connection from generator
	#if defined (USE_ZMQ)
		void *context = zmq_ctx_new ();
		void *socket=receive_connection(context,port);
	#else
		int socket;
		socket=*(int *)receive_connection(1,port);		
	#endif

	/**
		The first receive is for synchronizing the global start time
	*/
	if((ret=socket_receive(socket, &tmp, sizeof(task_t)))!=sizeof(task_t))
	{
		fprintf(stderr,"The program is a bottleneck\n");
		exit(BOTTLENECK_ERR);
	}
	*start_global_ticks=tmp.timestamp;
	gettimeofday(&tmp_t,NULL);

	#if defined(MONITORING)
		//start monitoring
        monitoring_timer=getticks();
    	volatile ticks start_ticks=getticks();
    #endif

	/**
		Start receiving real elements
	*/
	#if defined(TASK_BUFF)
		tb=&task_buff[bi];
	#else
		#if defined(DONT_USE_FFALLOC)
		//POSIX_MEMALIGN((void **)&tb,CACHE_LINE_SIZE,sizeof(task_t));
			posix_memalign((void **)&tb,CACHE_LINE_SIZE,sizeof(task_t));
			//tb=(task_t *)malloc(sizeof(task_t));
		#else
			tb=(task_t *)ffalloc->malloc(sizeof(task_t)); //posix memalign does not work for ff allocator (bugged)
		#endif
	#endif

	if((ret=socket_receive(socket, tb, sizeof(task_t)))!=sizeof(task_t))
	{
		fprintf(stderr,"The program is a bottleneck\n");
		exit(BOTTLENECK_ERR);
	}

	//NOTA: usiamo la seconda soluzione per il tempo di partenza: viene considerato il tempo di arrivo della prima tupla
	// DA COMMENTARE SE SI VOGLIONO I TEMPI "SINCRONIZZATI" TRA EMETTITORE E DISTRIBUTORE 	

	//start_global_ticks=getticks();
	asm volatile("":::"memory");
    start_t=tmp_t.tv_sec*1000+(tmp_t.tv_usec)/1000;
    monitoring_timer=getticks();

    bool reconfiguring=false;
	while(tb->type!=-1)
	{

		msg++;
		#if defined (MONITORING)
			monitoring[mon_ptr].elements++;
			monitoring[mon_ptr].elements_per_class[tb->type]++;
		#endif

		//set the internal_id of the task, incremental for each class
		tb->internal_id=classes_freq[tb->type]++;
		tb->punctuation=NO;

		//determine to which worker send the task
		/*#ifdef LESS_ASSIGNED
		to_send_to=schedulingLessAssigned(tb,num_workers,pane_size);
		#else
		#ifdef LESS_LOADED
		to_send_to=schedulingLessLoaded(tb,num_workers,num_classes, pane_size);
		#else
		#ifdef EMPTIEST_QUEUE
		to_send_to=schedulingEmptiestQueue(tb,num_workers,num_classes, pane_size, outqueue);
		#endif
		#endif
		#endif*/
		// to_send_to=scheduling_table[tb->type]-1;
		#if defined(ROUND_ROBIN)
		to_send_to=schedulingRR(tb,num_workers,pane_size);
		
		#endif
		//to_send_to=schedulingAdaptive(tb,num_workers,pane_size,new_scheduling_table);


		//to_send_to=task_buff[bi].type;
		tb->worker=to_send_to;
		//if(reconfiguring)
		//send towards the proper worker
		if(!send(tb,outqueue[to_send_to]))
		{
			fprintf(stderr, "Worker %d is a bottleneck\n",to_send_to);
			exit(BOTTLENECK_ERR);

		}	

		//take the pointer to the next task that will be received
		#if defined(TASK_BUFF)
			//advance buffer index
			bi=(bi+1)%(buff_size);
			tb=&task_buff[bi];
		#else
			#if defined(DONT_USE_FFALLOC)
				posix_memalign((void **)&tb,CACHE_LINE_SIZE,sizeof(task_t));
			#else
				tb=(task_t *)ffalloc->malloc(sizeof(task_t));
			#endif
		#endif

		#if defined(MONITORING)
			#if defined (ADAPTIVITY)

			//NOTA: per ora l'implementazione dei meccanismi di riconfigurazione non e' best effort. Bisognerebbe finire di considerare la nuova scheduling
			//table quando l'emettitore invia il nuovo monitoraggio
				
			//check if there are reconfiguration messages from the controller
			if(cn_inqueue->pop((void **)(&reconf_data)))
			{
				//check if there are newly spawned threads
				if(reconf_data->par_degree_changes>0)
				{
					printf("Emitter: ricevuto messaggio, incremento par degree di %d worker\n",reconf_data->par_degree_changes);
					//take the new queues
					for(int i=0;i<reconf_data->par_degree_changes;i++)
					{
						outqueue[num_workers+i]=reconf_data->wqueues[i];
					}
					//increments the par degree
					num_workers+=reconf_data->par_degree_changes;
				}
				new_scheduling_table=reconf_data->scheduling_table;
				yet_to_be_reconfigured=num_classes;
				//for(int i=0;i<num_classes;i++)
				//	printf("%d %d\n",i,new_scheduling_table[i]);
				printf("Emitter: reconfiguration started\n");
				reconfiguring=true;
			}
			//only when reconfiguration finishes, notify it to controller
			if(reconfiguring && yet_to_be_reconfigured==0)			
			{
				new_scheduling_table=NULL;
				bsend((void*)&reconf_finished,cn_outqueue);
				printf("Emitter: reconfiguration finished\n");
				reconfiguring=false;
			}

			#endif
			//check if we can send monitoring data to controller
			if(getticks()-monitoring_timer>monitoring_step)
			{ 	
				asm volatile("":::"memory");
				// printf("Monitoring data %d are valid. Elements received: %Ld\n",mon_ptr,monitoring[mon_ptr].elements);
				monitoring[mon_ptr].scheduling_table=scheduling_table; //NOTA: per il momento passiamo il puntatore (o fare la copia? dipende dal controllore che fa)
				
				//printf("Emitter sending...%d\n",cn_outqueue->length());
				bsend(&monitoring[mon_ptr],cn_outqueue); //pay attention, if queue is full it is blocked

				//printf("Emitter sent step: %d\n",mon_step++);
				monitoring_timer=getticks();
				mon_ptr=(mon_ptr+1)%NMONITORING;
				//reset the monitoring data
				monitoring[mon_ptr].elements=0;
				memset(monitoring[mon_ptr].elements_per_class,0,num_classes*sizeof(int));
				//qlength[mon_step++]=outqueue[0]->length();
			}
			
		#endif


		//receive the next element
		if((ret=socket_receive(socket, tb, sizeof(task_t)))!=sizeof(task_t))
		{
			fprintf(stderr,"The program is a bottleneck\n");
			exit(BOTTLENECK_ERR);
		}
	}

	//EOS: send terminating task to all the workers
	for(int i=0;i<num_workers;i++)
		send(tb,outqueue[i]);

	#if defined(MONITORING)
		//send termination to controller
		monitoring[mon_ptr].stop=true;
		bsend(&monitoring[mon_ptr],cn_outqueue);
	#endif
	//compute statistics
	gettimeofday(&tmp_t,NULL);
	end_t=tmp_t.tv_sec*1000+(tmp_t.tv_usec)/1000-start_t;
	printf("***EMITTER***\n");
	printf("Tempo passato (msec): %Ld, msg per second: %6.3f\n",end_t,((double)msg)/((double)end_t/1000));
	#if defined (USE_ZMQ)
	close(socket);
	zmq_ctx_destroy (context);
	#endif

	return NULL;
}