/**
	Worker for Pane Based Implementation
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <ff/buffer.hpp>
#include <iostream>
#include <unordered_map>
#include <vector>
#include <sched.h>
#include <ff/allocator.hpp>	
#include "../Includes/cycle.h"
#include "../Includes/iaperf.h"
#include "../Includes/general.h"
#include "../Includes/pane_cb.h"
#include "../Includes/cbpane.hpp"
using namespace ff;
using namespace std;

void * worker(void *args) {

	task_t *tmp=(task_t*)malloc(sizeof(task_t));
	CBPane *pane;
	WorkerResult_t *res_buff; //result buffer in order to reuse memory
	int buff_size;
	int bi=0;
	double res;

	

	//Init
	worker_data_t *data = (worker_data_t *) args;
	pthread_barrier_t *barrier = data->barrier;
	int id = data->workerId;
	SWSR_Ptr_Buffer *inqueue=data->inqueue;
	SWSR_Ptr_Buffer *outqueue=data->outqueue;
	int pane_size=data->pane_size;
	unordered_map<int,CBPane*> map; //map that contains the various windows
	long int freq=data->freq;
	int num_classes=data->num_classes;
	#if !defined(DONT_USE_FFALLOC) && !defined(TASK_BUFF)
			ff_allocator *ffalloc=data->ffalloc;
	#endif
	#if defined(MONITORING)
		SWSR_Ptr_Buffer *cn_outqueue=data->cn_outqueue;	
	#endif

	//EOS
	WorkerResult_t eos;
	eos.isEOS=true;
	//memory allocation
	buff_size=QUEUE_SIZE+10;//is it correct?
	posix_memalign((void **)&res_buff,CACHE_LINE_SIZE, buff_size*sizeof(WorkerResult_t));

	//MONITORING TO DO...
	#if defined(MONITORING)
		//define the data structures for monitoring
		
		ticks start; //for computation times

		volatile ticks monitoring_timer;
		volatile ticks monitoring_step=MONITORING_STEP*1000*freq; //after how many cycles we have to monitor
		worker_monitoring_t *monitoring;
		char mon_ptr=0; //the current monitoring structure in which write data
		posix_memalign((void **)&monitoring, CACHE_LINE_SIZE,NMONITORING*sizeof(worker_monitoring_t));
		for(int i=0;i<NMONITORING;i++)
		{
			//init the internals
			posix_memalign((void**)&(monitoring[i].tcalcg_per_class),CACHE_LINE_SIZE,num_classes*sizeof(double));
			memset(monitoring[i].tcalcg_per_class,0,num_classes*sizeof(double));
			posix_memalign((void **)&(monitoring[i].computations_per_class),CACHE_LINE_SIZE,num_classes*sizeof(int));
			memset(monitoring[i].computations_per_class,0,num_classes*sizeof(int));
		}
		//adaptivity data structures
		/*reconf_repository_t *repo=(reconf_repository_t *)data->repo;
		unordered_map<int,CBWindow*> moving_windows=*(repo->moving_windows);
		bool reconfiguration_phase_in=false; //indicates if the worker is currently in the reconfiguration phase (moving_in)
		bool reconfiguration_phase_out=false; //indicates if the worker is currently in the reconfiguration phase (moving_out)
		bool reconfiguration_phase=false; //indicate the whole process
		set<int> classes_moving_in; //it will contains all the id of the classes that are currently moving in
		vector<task_t *> task_moving_in(20000); //it will contain all the task belonging to moving in classes (not yet arrived to the worker) (to see why i choose vector look at the explnation below, when i use it)
		task_moving_in.clear();
		int64_t *prevv=(int64_t*)calloc(num_classes,sizeof(int64_t)); //debug*/
	#endif


	if(barrier) //the thread has been created at the starting time
	{
		//queues warmup
		receive((void **)&tmp,inqueue);
		send((void *)1,outqueue);
		//synchronization barrier
		DEBUG(printf("Worker %d ready, running on %d pane size:%d \n",id,sched_getcpu(),pane_size))
		pthread_barrier_wait(barrier);
	}
	
	#if defined(MONITORING)
	monitoring_timer=getticks();
	#endif
	/**
		Start receiving elements
	*/

	receive((void **)&tmp,inqueue);
	int count=0;
	while(tmp->type!=-1)
	{
		/*
			insert the task into the pane and if necessary, perform computation
		*/
		//if(tmp->worker!=id)
		//	printf("[%d] E' arrivato un elemento diretto a: %d\n",tmp->worker);
			/*count++;

				if(count==500)
				{
					volatile ticks st=getticks();
					while(getticks()-st<6000000);
					//usleep(3050);
					count=0;
					res_buff[bi].timestamp=tmp->timestamp;
					send(&res_buff[bi],outqueue);
					bi=(bi+1)%buff_size;
					res_buff[bi].type=tmp->type;
					res_buff[bi].isEOS=false;
					
				}*/
		#if defined(MONITORING)
			asm volatile("":::"memory");
			start=getticks();
			asm volatile("":::"memory");
		#endif
		pane=map[tmp->type];
		//monitoring to do...
		if(pane==NULL) //it's a new logical stream that has arrived
		{
			//DEBUG(printf("--------------------[%d] Ricevuto un elemento di un nuovo stream logico: %d\n",id,tmp->type));
			
			pane=new CBPane(pane_size);
			map[tmp->type]=pane;
		}

		//insert the element in window
		int ret=pane->insert(tmp,&res);

		if (ret!=0) //we have a new result 
		{
			//DEBUG(printf("--------------------[%d] Calcolo pane per stream logico: %d iid:%Ld\n",id,tmp->type,tmp->internal_id));
			//set the timestamp to the timestamp of the one of the task that has triggered the computation
			res_buff[bi].timestamp=tmp->timestamp;
			//set the result 
			res_buff[bi].res=res;
			//se the id to the internal id (incremental for class) of the task that has triggered the computation
			res_buff[bi].id=tmp->internal_id;
			res_buff[bi].type=tmp->type;
			res_buff[bi].isEOS=false;
			//send result to the collector
			
			send(&res_buff[bi],outqueue);
			//advance the buffer index
			bi=(bi+1)%buff_size;

			//if needed save the computation time
		

			//delete the entry in the map
			delete pane;
			map.erase(tmp->type); //richiama il distruttore dell'elemento
			//we have to destroy (EVENTUALLY RECYCLE) the pane that is now useless

			//qua si potrebbe riciclare il pane, oppure riciclare la entry nella map
			#if defined(MONITORING)
				asm volatile("":::"memory");
				monitoring[mon_ptr].tcalcg_per_class[tmp->type]+=(double)(getticks()-start)/freq; //usec
				monitoring[mon_ptr].computations_per_class[tmp->type]++;
				asm volatile("":::"memory");
			#endif

		}
		#if defined(MONITORING)
			//check if we can send monitoring data to controller

			if(getticks()-monitoring_timer>monitoring_step)
			{ 	
				asm volatile("":::"memory");
				// printf("Monitoring data %d are valid. Elements received: %Ld\n",mon_ptr,monitoring[mon_ptr].elements);
				
				//blocking send towards the collector
				bsend(&monitoring[mon_ptr],cn_outqueue);

				monitoring_timer=getticks();
				mon_ptr=(mon_ptr+1)%NMONITORING;
				//reset the monitoring data
				//memset(monitoring[mon_ptr].elements_per_class,0,num_classes*sizeof(int));
				memset(monitoring[mon_ptr].tcalcg_per_class,0,num_classes*sizeof(double));
				memset(monitoring[mon_ptr].computations_per_class,0,num_classes*sizeof(int));
			}
		#endif
		#if !defined(TASK_BUFF)
			#if defined(DONT_USE_FFALLOC)
				//if(tmp!=NULL)
				free(tmp);
			#else
			ffalloc->free(tmp);
			#endif
		#endif
		//receive the next element
		receive((void **)&tmp,inqueue);
	}

	#if defined(MONITORING)
		//send the last monitoring, in order to avoid stall on the controller
		bsend(&monitoring[mon_ptr],cn_outqueue);
	#endif
	//send EOS to collector
	send(&eos,outqueue);
	//printf("Worker %d finito\n",id);
  return NULL;
}